/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

// com.oracle.iot.client.device
/**
 * Implementation of protocol Equatable
 *
 * parameter lhs: one item to be compared
 *
 * parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
public func ==(lhs: AdvancedDirectlyConnectedDevice,
               rhs: AdvancedDirectlyConnectedDevice) -> Bool {
    return lhs.hashValue == rhs.hashValue
}

/**
 * A directly&ndash;connected device is able to send messages to, and receive 
 * messages from, the IoT server. A directly&ndash;connected device has a 
 * physical&ndash;device identifier that is registered with the server. When 
 * the directly&ndash;connected device is activated on the server, the server 
 * assigns a logical&ndash;endpoint identifier. This logical&ndash;endpoint 
 * identifier is required for sending messages to, and receiving messages from, 
 * the server.
 */
public class AdvancedDirectlyConnectedDevice : Hashable {
    
    internal var sendReceiveImpl: SendReceiveImpl?
    internal var secureConnection: HttpSecureConnection?
    internal var trustedAssetsManager: TrustedAssetsManager?
    
    // Checked by all methods. If a method is called and isClosed is true
    // call the callback with the error or throw `ClientError.state`
    // depending on the method signature.
    internal var isClosed: Bool = false
    
    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     */
    public var hashValue : Int {
        return ObjectIdentifier(self).hashValue
    }
    
    /**
     * This constructor is used internally and is not intended for general use. 
     *
     * - parameter trustedAssetsManager: the `TrustedAssetsManager` to store,
     *                            load and handle the device trust material.
     *
     */
    init(trustedAssetsManager: TrustedAssetsManager) throws {
        self.trustedAssetsManager = trustedAssetsManager
        self.secureConnection = nil
        self.secureConnection =
            HttpSecureConnection.createHttpSecureConnection(tam: trustedAssetsManager,
                                                    useOnlySharedSecret: false)
        self.sendReceiveImpl =
            SendReceiveImpl(secureConnection: secureConnection! )
    }
    
    /**
     * Constructs a new `AdvancedDirectlyConnectedDevice` instance that will
     * use a custom or default `TrustedAssetsManager` to store, load and handle
     * the device trust material.
     */
    convenience init() throws {
        try self.init(trustedAssetsManager: TrustedAssetsManagerFactory.getTrustedAssetsManager(context: AdvancedDirectlyConnectedDevice.self)!)
    }
    
    /**
     * Constructs a new `AdvancedDirectlyConnectedDevice` instance that will
     * use the given filename, fileextension, subdirectory and password to 
     * provision a `TrustedAssetManager` instance
     * - throws: 
     *      - `ClientError.security` if server policy cannot be set.
     *      - `ClientError.file` if the provision file is empty or has the
     *        wrong format version, or can't be decrypted or the server url is
     *        invalid.
     */
    convenience init(file: String, fileExtension: String? = nil,
                     subdirectory: String? = nil, password: String) throws {
        try self.init(trustedAssetsManager: TrustedAssetsManagerFactory.createTAMWithClassName(file: file,
                    fileExtension: fileExtension, subdirectory: subdirectory,
                    password: password)!)
    }
    
    /**
     * Constructs a new `AdvancedDirectlyConnectedDevice` instance that will 
     * use the given path and password to provision a `TrustedAssetManager` 
     * instance
     * 
     * - parameter path: the path to the provisioning file
     *
     * - parameter password: the password to protect the provisioning file 
     *                       during the creation of the file
     *
     * - throws: 
     *       - `ClientError.file` if the provisioning file is not present
     *       - `ClientError.security` if the provisioning file cannot be
     *           decrypted or there is invalid information in the file
     */
    public convenience init(path: String, password: String) throws {
        try self.init(trustedAssetsManager: TrustedAssetsManagerFactory.createTrustedAssetsManager(path: path, password: password)!)
    }

    /**
     * Create a new `StorageObject` that will have a name with the given
     * object name prefixed with the device's endpoint ID and a directory
     * separator. The prefix addition can be disabled by setting the
     * `Config.DISABLE_STORAGE_OBJECT_PREFIX` to `true`.
     *
     * If `contentType` is `nil`, the mime-type defaults to
     *  "application/octet&ndash;stream".
     *
     * - parameter name: the unique name to be used to reference the content
     *                    in storage
     * - parameter contentType: The mime-type of the content
     *
     * - returns: a `StorageObject`
     * - throws:
     *      - `ClientError.network` if there is an exception raised by the 
     *                             runtime, or an abnormal response from the
     *                             storage cloud
     *      - `ClientError.security` if there is an exception establishing a
     *                              secure connection to the storage cloud
     */
    public func createStorageObject(name:String, contentType:String? = OCTET_STREAM)
            throws -> AdvancedStorageObject {
        let storageConnection:StorageConnectionBase =
            StorageConnectionMap.getStorageConnection(secureConnection:secureConnection!)
            return try storageConnection.createStorageObject(clientId:getEndpointId(),
                                                             name:name, contentType:contentType!)
    }

    /**
     * Create a new `StorageObject` from the URI for a named object in storage.
     *
     * - parameter url: the URL of the object in the storage cloud
     *
     * - returns: a `StorageObject` for the object in storage
     *
     * - throws: 
     *      - `ClientError.network` if there is an exception raised by the
     *                              runtime, or a failure reported by the
     *                              storage cloud
     *      - `ClientError.security` if there is an exception establishing a
     *                               secure connection to the storage cloud
     */
    public func createStorageObject(uri:String,
                callback: @escaping (AdvancedStorageObject?, ClientError?)->()) {
        let storageConnection:StorageConnectionBase =
            StorageConnectionMap.getStorageConnection(secureConnection:secureConnection!)
        do {
            try storageConnection.createStorageObject(storageUrl:uri,
                                                      callback: callback)
        } catch {
            callback(nil, error as? ClientError)
        }
    }

    /**
     * Activate the device. The device will be activated on the server if
     * necessary. When the device is activated on the server, the server
     * assigns a logical endpoint identifier to the device.
     *
     * The user should call the `isActivated()` method prior to calling 
     * activate.
     *
     * - parameter deviceModels: should contain the device model type URNs of 
     *                           this directly connected device
     *
     * - throws: `ClientError.state` if the device is already activated or
     *           `close` has been called
     *
     * - see: `isActivated()`
     */
    public func activate(deviceModels: String...) throws {
        if (isClosed) {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment:""))
        }
        try self.activate(deviceModels: deviceModels)
    }
 
    /**
     * Activate the device. The device will be activated on the server if
     * necessary. When the device is activated on the server, the server
     * assigns a logical endpoint identifier to the device.
     *
     * The user should call the `isActivated()` method prior to calling 
     * activate.
     *
     * - parameter deviceModels: should contain the device model type URNs of 
     *                           this directly connected device
     *
     * - throws: `ClientError.state` if the device is already activated or 
     *           'close' has been called
     *
     * - see: isActivated()
     */
    public func activate(deviceModels: [String]) throws {
        if (isClosed) {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment:""))
        }
        try self.activate(deviceModels: deviceModels,
                          callback: { endpointId, error in })
    }
    
    /**
     * Activate the device. The device will be activated on the server if
     * necessary. When the device is activated on the server, the server
     * assigns a logical endpoint identifier to the device.
     *
     * The user should call the `isActivated()` method prior to calling 
     * activate.
     *
     * - parameter deviceModels: should contain the device model type URNs of
     *                           this directly connected device
     *
     * - parameter callback: The function to be executed after completion of
     *                      activation. If activation is not successful then
     *                      the object will be `nil` and an error object is 
     *                      passed in the form `callback(endpoint, error)` and 
     *                      the reason can be taken from error
     *
     * - throws: `ClientError.state` if the device is already activated or
     *           `close` has been called
     *
     * - see: isActivated()
     */
    public func activate(deviceModels: [String],
            callback: @escaping (String?,ClientError?) -> ()) throws {
        if (isClosed) {
            callback(nil, ClientError.state(NSLocalizedString(
                "Client is closed", comment:"")))
            return
        }
        
        if self.isActivated() {
            throw ClientError.state(NSLocalizedString("Already activated",
                                                      comment: ""))
        }
        
        var deviceModelSet: [String] = []
        if deviceModels.count > 0 {
            for deviceModel in deviceModels {
                deviceModelSet.append(deviceModel)
            }
        }
        deviceModelSet.append(ActivationManager.DIRECT_ACTIVATION_URN)
        if self is AdvancedGatewayDevice {
            deviceModelSet.append(ActivationManager.INDIRECT_ACTIVATION_URN)
        }
        //  --- Question this has to be also set for DCD or only for Gateway
        deviceModelSet.append(ActivationManager.MESSAGE_DISPATCHER_URN)
        deviceModelSet.append(ActivationManager.DIAGNOSTICS_URN)
        //
        
        let activationManager: ActivationManager = ActivationManager.getInstance()
        try activationManager.getActivationPolicy(secureConnection: secureConnection!,
                deviceId: trustedAssetsManager!.getClientId(),
                callback: {(activationPolicyResponse: ActivationPolicyResponse?, error: ClientError?) -> Void in
            if (error != nil) {
                callback(nil, error)
                return
            }
            
            if (activationPolicyResponse == nil) {
                callback(nil, ClientError.argument(NSLocalizedString("AdvancedDirectlyConnectedDevice activate() activationPolicyResponse should not be nil", comment: "")))
                return
            }
                    
            do {
                try self.trustedAssetsManager!.generateKeyPair(
                    algorithm: activationPolicyResponse!.getKeyType(),
                    keySize: activationPolicyResponse!.getKeySize())
            } catch {
                callback(nil, ClientError.security(String(format:
                    NSLocalizedString("Issue while generating Key Pair: %@",
                                      comment: ""), "\(error)")))
                return
            }
            
            var directActivationRequest: DirectActivationRequest?
            do {
                directActivationRequest =
                    try activationManager.createDirectActivationRequest(
                        trustedAssetsManager: self.trustedAssetsManager!,
                        hashAlgorithm: activationPolicyResponse!.getHashAlgorithm(),
                        deviceModels: deviceModelSet)
            } catch {
                callback(nil, ClientError.security(NSLocalizedString(
                    "Issue while generating ActivationRequest", comment:"")))
                return
            }
            do {
                try activationManager.postDirectActivationRequest(
                        secureConnection: self.secureConnection! ,
                        deviceId: self.trustedAssetsManager!.getClientId(),
                        directActivationRequest: directActivationRequest!,
                        callback: { (directActivationResponse: DirectActivationResponse?, error: ClientError?) -> Void in
                    if error != nil {
                        callback(nil, error)
                        return
                    }
                            
                    if (directActivationResponse == nil) {
                        callback(nil, ClientError.argument(NSLocalizedString("AdvancedDirectlyConnectedDevice activate() directActivationResponse should not be nil", comment: "")))
                        return
                    }
                            
                    let endpointId:String = directActivationResponse!.getEndpointId()
                    let certificate: Foundation.Data =
                        directActivationResponse!.getCertificate()
                    
                    do {
                        try self.trustedAssetsManager!.setEndPointCredentials(
                            endpointId: endpointId, certificate: certificate)
                    } catch let error as ClientError {
                        callback(nil, error)
                        return
                    } catch {
                        callback(nil, ClientError.security(NSLocalizedString(
                                "Issue while setting endpoint credentials",
                                comment: "")))
                        return
                    }
                    
                    // request to update accessToken
                    self.secureConnection?.disconnect()
                    
                    callback(endpointId, nil)
                })
            } catch {
                callback(nil, ClientError.network(NSLocalizedString(
                    "Issue while retrieving ActivationResponse", comment: "")))
                return
            }
        })
    }
    
    
    /**
     * Returns whether the device is activated.
     *
     * - returns: `true` if the device is activated, else `false`.
     */
    public func isActivated() -> Bool {
        // TODO verify 'isClosed' value in order to know if client has been 
        // closed
        return self.trustedAssetsManager!.isActivated()
    }
    
    /**
     * Send messages to the server. This call will block until the server
     * responds to the message delivery, or a network timeout occurs.
     *
     * - parameter messages: zero or more `DataMessage`, `AlertMessage`,
     *                      `ResponseMessage` and/or `ResourceMessage` objects 
     *                      authenticated with the server.
     * - parameter callback: The function to be executed after completion of
     *                      sending or queing a message. If send is not
     *                      successful, then an error object is passed
     *                      in the form `callback(error)` and the
     *                      reason can be taken from error
     * - throws: `ClientError.state` if the device is not activated or 'close'
     *           has been called
     */
    public func send(messages: Message...,
                     callback: @escaping (ClientError?) -> ()) throws {
        if (isClosed) {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment: ""))
        }
        try send(messages: messages, callback: callback)
    }
    
    /**
     * Send messages to the server. This call will block until the server
     * responds to the message delivery, or a network timeout occurs. 
     *
     * - parameter messages: zero or more `DataMessage`, `AlertMessage`,
     *                       `ResponseMessage` and/or `ResourceMessage` objects
     *                       authenticated with the server.
     * - parameter callback: The function to be executed after completion of
     *                      sending or queing a message. If send is not
     *                      successful, then an error object is passed
     *                      in the form `callback(error)` and the
     *                      reason can be taken from error
     * - throws: `ClientError.state` if the device is not activated or 'close'
     *           has been called
     */
    public func send(messages: [Message]?,
                     callback: @escaping (ClientError?) -> ()) throws {
        if (isClosed) {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment: ""))
        }
        
        if self.isActivated() {
            try self.sendReceiveImpl!.send(messages: messages,
                                           callback: callback)
        } else {
            throw ClientError.argument(NSLocalizedString(
                "Directly Connected Device not activated", comment: ""))
        }
    }
    
    /**
     * Receive a *RequestMessage* from the server, if any. This call may
     * block if the implementation sends a message to the server to poll
     * for requests (see *send(Message...)*). This call may return nil
     * if there are no requests from the server.
     *
     * - parameter timeout: maximum time, in milliseconds, to wait for a 
     *                      response from the server
     * - parameter callback: The function to be executed after receiving a
     *                      a message. If receive is not successful then
     *                      the `ReqeuestMessage` will be `nil` and an error
     *                      object is passed in the form `callback(nil, error)`
     *                      and the reason can be taken from error
     *
     * - throws: `ClientError.state` if the device is not activated or 'close' 
     *           has been called
     */
    public func receive(timeout: Int,
            callback: @escaping (RequestMessage?, ClientError?) -> ()) throws {
        if (isClosed) {
            throw ClientError.state(NSLocalizedString("Client is closed",
                                                      comment: ""))
        }
        if self.isActivated() {
            return try self.sendReceiveImpl!.receive(timeout: Int64(timeout),
                                                     callback: callback)
        } else {
            throw ClientError.argument(NSLocalizedString(
                "Directly Connected Device not activated", comment: ""))
        }
    }
    
    /**
     * Return the logical&ndash;endpoint identifier of this 
     * directly&ndash;connected device. The logical&ndash;endpoint identifier 
     * is assigned by the server as part of the activation process.
     *
     * - returns: the logical&ndash;endpoint identifier of this 
     *            directly&ndash;connected device.
     *
     * - see: `activate(String...)`
     */
    public func getEndpointId() -> String {
        // TODO verify 'isClosed' value in order to know if client has been 
        // closed
        return self.trustedAssetsManager!.getClientId()
    }
    
    /**
     * Get the `DeviceModel` for the device model urn. This method may
     * return `nil` if there is no device model for the URN. `nil` may also be
     * returned if the device model is a &quot;draft&quot; and the property
     * `Config.ALLOW_DRAFT_DEVICE_MODELS` is set to
     * `false`, which is the default.
     *
     * - parameter deviceModelUrn: The URN of the device model
     *
     * - parameter callback: This function is called with the retrieved device
     *                       model. If retrieving the device model is not 
     *                       successful then the object will be `nil` and an
     *                       error object is passed in the form callback(device,
     *                       error) and the reason can be taken from 
     *                       error
     *
     * On error `callback` receives a `ClientError.network` error if
     *    the object is not found (StatusCode.NOT_FOUND),
     *    the request fails (!StatusCode.OK), or
     *    the returned data is `nil` or JSON parsing fails
     */
    public func getDeviceModel(deviceModelUrn: String,
            callback: @escaping (DeviceModel?, ClientError?) -> Void) throws {
        // TODO: FIXME Does not really throw. Exceptions are thrown in 
        // secureConnection completion handler.
        if (isClosed) {
            callback(nil, ClientError.state(NSLocalizedString(
                "Client is closed", comment:"")))
            return
        }
        /*
         * The `DirectlyConnectedDevice` class has no trusted
         * assets manager and this class gives no access to the one it has,
         * so this method is here.
         */
        return try DeviceModelFactory.getDeviceModel(secureConnection: self.secureConnection!,
                                    urn: deviceModelUrn, callback: callback)
    }
    
    deinit {
        self.close()
    }
    
    /**
     * Close the AdvancedDirectlyConnectedDevice.
     */
    public func close() {
        if (!self.isClosed) {
            self.secureConnection?.close()
            
            self.sendReceiveImpl = nil
            self.secureConnection = nil
            self.trustedAssetsManager = nil
            
            self.isClosed = true
        }
    }

    /*
     * Private only
     */
    fileprivate class StorageConnectionMap  {
        // TODO: Is it ok to use endpoint id to uniquiely map to a 
        // secureConnection ? It is problematic to make SecureConnection
        // Hashable.

        private static
            var STORAGE_CONNECTION_MAP:Dictionary<String,StorageConnectionImpl> =
                Dictionary<String,StorageConnectionImpl>();

        fileprivate class func getStorageConnection(secureConnection:SecureConnection)
                -> StorageConnectionImpl {
            var storageConnection = STORAGE_CONNECTION_MAP[secureConnection.getEndpointId()!]
            if storageConnection == nil {
                storageConnection = StorageConnectionImpl(secureConnection:secureConnection)
                STORAGE_CONNECTION_MAP[secureConnection.getEndpointId()!] = storageConnection
            }
            return storageConnection!
        }
    }
}
