/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * An Alert to be sent to the server. The alert is sent by calling the
 * `raise()` method. The time of the alert is set when `raise()`
 * is called, allowing `raise()` to be called more than once.
 *
 * The `set` method returns the `Alert` instance to allow
 * the fields of the alert to be set in fluent style.
 
 * - see: `VirtualDevice.createAlert(format:)`
 */

public protocol Alert {

    /**
     * Set the value of a field in the `Alert`. The fields are
     * determined by the format given when the Alert is created.
     * The value is validated according to the constraints in the format.
     * If the value is not valid, a `ClientError.argument` is raised.
     *
     * All fields defined in the format that are *"optional"*
     * must be set before calling `raise()`.
     *
     * - parameter fieldName: the name of a field from the alert format
     *
     * - parameter value: the value to set
     *
     * - returns: Alert object that can be used for function chaining
     *
     * - throws: `ClientError.argument` if `fieldName` is not a valid field name
     *           or `value` is not the field type.
     */
    func set(fieldName: String, value: AnyObject) throws -> Alert

    /**
     * Send the alert. The event time is set when this method is called.
     *
     * The `onError` handler of the parent virtual device will be called
     * if there is error sending the alert.
     *
     * All fields defined in the format that are *"optional"*
     * must be set before calling `raise()`. If `raise()` is called
     * before setting all *"optional"* fields, the function will fail
     * with a specific ErrorType
     *
     * - throws: 
     *      - `ClientError.argument` if a required field is not set
     *      - `ClientError.network` if a `MessageDispatcher` cannot be
     *           obtained.
     */
    func raise() throws
}

class AlertImpl: Alert {
    
    private var  virtualDevice: VirtualDeviceImpl?
    private var  deviceModelFormat: DeviceModelFormat?
    private var fieldValues: [String : AnyObject]?
    
    /**
     * Creates instance of an `Alert` implementation.
     *
     * - parameter virtualDevice: the virtual device for which the `Alert` is 
     *                            created
     *
     * - parameter alertUrn: the URN for the alert
     *
     * - throws: `ClientError.argument` if a device model format cannot be 
     *           obtained from `alertUrn` or if the format is not
     *           `DeviceModelFormatType.ALERT`
     *
     */
    init(virtualDevice: VirtualDeviceImpl, alertUrn: String) throws {
        
        self.deviceModelFormat =
            AlertImpl.getDeviceModelFormat(virtualDevice: virtualDevice,
                                           alertUrn: alertUrn)
        
        if (self.deviceModelFormat == nil) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "%@ not found", comment: ""), alertUrn))
        }
        
        if (self.deviceModelFormat!.getType() != .ALERT) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "%@ is not an alert format", comment: ""), alertUrn))
        }
        
        self.virtualDevice = virtualDevice
        self.fieldValues = [ : ]
    }

    /**
     * - throws: `ClientError.argument` if `fieldName` is not a valid field name
     *           or `value` is not the field type.
     */
    func set(fieldName: String, value: AnyObject) throws -> Alert {
        let field: DeviceModelFormatField? = getField(fieldName: fieldName)
        if (field == nil) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "%@ is not in the model", comment: ""), fieldName))
        }
        
        switch (field!.getType()) {
        case .NUMBER:
            if (!(value is NSNumber)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not a number", comment: ""), fieldName))
            }
            break
        case .INTEGER:
            if (!(value is Int)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not an Int", comment: ""), fieldName))
            }
            break
        case .DATETIME:
            if (!(value is Int64)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not a DateTime", comment: ""), fieldName))
            }
            break
        case .BOOLEAN:
            if (!(value is Bool)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not a Bool", comment: ""), fieldName))
            }
            break
        case .STRING:
            if (!(value is String)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "%@ is not a String", comment: ""), fieldName))
            }
            break
        case .URI:
            if (!(value is ExternalObject)) {
                throw ClientError.argument(String(format: NSLocalizedString(
                    "value for '%@' is not an ExternalObject", comment: ""),
                                                  fieldName))
            }
            break
        }
        
        fieldValues![fieldName] = value
        
        return self
    }
    
   /**
    * - throws: 
    *       - `ClientError.argument` if a required field is not set
    *       - `ClientError.network` if a `MessageDispatcher` cannot be 
    *           obtained.
    */
   func raise() throws {
        let builder: AlertMessageBuilder = AlertMessageBuilder()
        let fieldList: [DeviceModelFormatField] = deviceModelFormat!.getFields()
        var storageObjects = [StorageObjectImpl]()
        for field in fieldList {
            let key: String = field.getName()
            let value: AnyObject? = fieldValues![key]
            fieldValues![key] = nil
            if (value != nil) {
                switch (field.getType()) {
                case .NUMBER:
                    try builder.addDataItem(key: key,
                                            value: Double(truncating: value as! NSNumber))
                    break
                case .INTEGER:
                    try builder.addDataItem(key: key, value: value as! Int)
                    break
                case .DATETIME:
                    try builder.addDataItem(key: key,
                                            value: Double(truncating: value as! NSNumber))
                    break
                case .BOOLEAN:
                    try builder.addDataItem(key: key, value: value as! Bool)
                    break
                case .STRING:
                    try builder.addDataItem(key: key, value: value as! String)
                    break
                case .URI:
                    if value is StorageObjectImpl {
                        let so = value as! StorageObjectImpl
                        if ((so.getSyncStatus() == SyncStatus.notInSync) ||
                                (so.getSyncStatus() == SyncStatus.syncPending)) {
                            storageObjects.append(so);
                        }
                        so.setSyncEventInfo(virtualDevice: virtualDevice!,
                                            nameForSyncEvent: key);
                        try so.sync();
                    }
                    let uri = (value as! ExternalObject).getURI()
                    try builder.addDataItem(key: key, value: uri)
                break
                }
                
            } else if (!field.isOptional()) {
                fieldValues = [ : ]
                throw ClientError.argument(String(format: NSLocalizedString(
                    "Non-optional field %@ is not set", comment: ""), key))
            }
        }
        
        fieldValues = [ : ]
        
        builder.setFormat(format: deviceModelFormat!.getURN())
        builder.setDescription(description: deviceModelFormat!.getDescription())
        builder.setSource(source: virtualDevice!.getEndpointId())
        
        let message: AlertMessage = try builder.build()
        let messageDispatcher: MessageDispatcherImpl =
            try MessageDispatcher.getMessageDispatcher(msgDirectlyConnectedDevice: virtualDevice!.msgDirectlyConnectedDevice!) as! MessageDispatcherImpl

        for so:StorageObjectImpl in storageObjects {
            messageDispatcher.addStorageObjectDependency(storageObject:so,
                                    clientMsgId: message.getClientId())
        }
        try messageDispatcher.queue(message: message)
    }
    
    private func getField(fieldName: String) -> DeviceModelFormatField? {
        if (deviceModelFormat == nil) {
            return nil
        }
        
        let fieldList: [DeviceModelFormatField] = deviceModelFormat!.getFields()
        for field in fieldList {
            if field.getName() == fieldName {
                return field
            }
        }
        return nil
    }
    
    private static func getDeviceModelFormat(virtualDevice: VirtualDeviceImpl,
                                             alertUrn: String) -> DeviceModelFormat? {
        
        let deviceModel: DeviceModel = virtualDevice.getDeviceModel()
        return (deviceModel as! DeviceModelImpl).getDeviceModelFormats()[alertUrn]
    }
}
