/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * The Severity of an Alert
 */
public enum AlertMessageSeverity: Int, CustomStringConvertible {
    case low = 4
    case normal = 3
    case significant = 2
    case critical = 1
    
    public var description: String {
        switch self {
        case .low: return "LOW";
        case .normal: return "NORMAL";
        case .significant: return "SIGNIFICANT";
        case .critical: return "CRITICAL";
        }
    }
}

/**
 * Implementation of protocol Equatable
 *
 * parameter lhs: one item to be compared
 *
 * parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
public func ==(lhs: AlertMessage, rhs: AlertMessage) -> Bool {
    if lhs === rhs {
        return true
    }
    
    if (lhs.description != rhs.description) {
        return false
    }
    
    if (lhs.severity != rhs.severity) {
        return false
    }
    
    if lhs.items.count != rhs.items.count {
        return false
    }
    for i in 0..<lhs.items.count {
        if (lhs.items[i] != rhs.items[i]) {
            return false
        }
    }
    
    if (lhs as Message != rhs as Message) {
        return false
    }
    
    return true
}

/**
 * This class defined the message type `AlertMessage`
 */
public class AlertMessage: Message {
    
    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     */
    override public var hashValue : Int {
        var result: Int = super.hashValue
        var itemsHash: Int = 0
        for item in items {
            switch item.getType() {
            case .STRING:
                do {
                    try itemsHash += DataItem(key: item.getKey(),value: (item.getValue() as! String)).hashValue
                } catch {
                }
                break
            case .BOOLEAN:
                do {
                    try itemsHash += DataItem(key: item.getKey(),value: (item.getValue() as! Bool)).hashValue
                } catch {
                }
                break
            case .DOUBLE:
                do {
                    try itemsHash += DataItem(key: item.getKey(),value: (item.getValue() as! Double)).hashValue
                } catch {
                }
                break
            }
        }
        result = 31 &* result &+ itemsHash
        result = 31 &* result &+ (description != nil ? (description?.hashValue)! : 0)
        result = 31 &* result &+ (severity?.hashValue)!
        return result
    }
    
    /**
     * optional description for this alert
     */
    fileprivate var description: String?
    
    /** 
     * severity of this alert 
     */
    fileprivate var severity: AlertMessageSeverity?
    
    /** 
     * format for the alert message 
     */
    fileprivate var format: String?
    
    /** 
     * Actual data for the alert message 
     */
    fileprivate var items: Array<DataItem> = []
    
    /**
     * The `AlertMessage` constructor takes AlertMessage.Builder and set values 
     * to each field. In case the value is `nil`, this method will set a default
     * value.
     * 
     * - `format` cannot be `nil` or empty.
     * - `description` cannot be `nil` or empty.
     *
     * - parameter builder: `AlertMessageBuilder`
     *
     * - throws: `ClientError.state` if `build.format` is `nil` or empty
     */
    fileprivate init(builder: AlertMessageBuilder) throws {
        try super.init(builder: builder)
        
        guard let format = builder.format, !format.isEmpty else {
            throw ClientError.state(NSLocalizedString(
                "Format cannot be nil or empty", comment: ""))

        }
        self.format = format
        
        // TODO : see if description should be removed from this object.
        self.description = builder.description
        
        for di in builder.items {
            items.append(di)
        }
        
        if builder.severity == nil {
            self.severity = AlertMessageSeverity.significant;
        }
        else {
            self.severity = builder.severity!
        }
    }
    
    
    /**
     * Gets the severity of this alert
     *
     * - returns: the severity of this alert
     */
    public func getSeverity() -> AlertMessageSeverity {
        if severity == nil {
            return .low
        } else {
            return self.severity!
        }
    }
    
    /**
     * Gets the description for this alert.
     *
     * - returns: the description for this alert.
     */
    public func getDescription() -> String {
        if description == nil {
            return ""
        } else {
            return self.description!
        }
    }
    
    /**
     * Get the name of the format for data.
     *
     * - returns: data format, never `nil`.
     */
    public func getFormat() -> String {
        if format == nil {
            return ""
        } else {
            return self.format!
        }
    }
    
    /**
     * Get message payload as array of *DataItems*
     *
     * - returns: Payload that is carried by `AlertMessage`, never `nil`
     */
    public func getDataItems() -> [DataItem] {
        return self.items
    }
    
    /**
     * Get type message type.
     *
     * - returns: type, never `nil`
     */
    override public func getType() -> MsgType {
        return .ALERT
    }
    
    /**
     * Exports data from `AlertMessage` to string using JSON interpretation of the message.
     
     * - returns: JSON interpretation of the message as string
     */
    override public func toString() -> String {
        return toJsonString()
    }
    
    /**
     * Export `AlertMessage` as JSON string
     
     * - returns: `AlertMessage` as JSON string
     */
    override public func toJsonString() -> String {
        var result: String
        var addings: String = ""
        
        addings = addings + "\"payload\":{"
        addings = addings + "\"format\":\"" + format! + "\","
        
        if description != nil {
            addings = addings + "\"description\":\"" + description! + "\","
        }
        
        addings = addings + "\"severity\":\"" + severity!.description + "\","
        addings = addings + "\"data\":{"
        var z: Int = 0
        for item in items {
            addings = addings + "\"" + item.getKey() + "\":"
            switch item.getType() {
            case .STRING:
                addings = addings + "\"" + (item.getValue() as! String) + "\""
                break
            case .BOOLEAN:
                addings = addings + (item.getValue() as! Bool).description
                break
            case .DOUBLE:
                addings = addings + Double(truncating: item.getValue() as! NSNumber).description
                break
            }
            z += 1
            if z < items.count {
                addings = addings + ","
            }
        }
        addings = addings + "}}"
        
        result = toJsonString(expand: false, additionalItems: addings)
        
        return result
    }

}

/**
 * *AlertMessageBuilder* extends *MessageBuilder* class. `AlertMessage` class is immutable.
 * A *Builder* is required when creating `AlertMessage`. `AlertMessage` uses Builder
 * design pattern.
 */
public class AlertMessageBuilder: MessageBuilder {
    
    
    /** message for the alert */
    fileprivate var description: String?
    
    /** severity for the alert */
    fileprivate var severity: AlertMessageSeverity?
    
    public override init() {
        super.init()
        // Alert messages are highest priority
        self.priority = .highest;
        
        // default severity is SIGNIFICANT
        self.severity = .significant;
    }
    
    /** format for the alert message */
    fileprivate var format: String?
    
    /** List of data items */
    fileprivate var items: [DataItem] = []
    
    private static func parseJsonStr(str: String) throws -> [String : AnyObject]? {
        if let data = str.data(using: .utf8) {
            return try JSONSerialization.jsonObject(with: data,
                                                        options: []) as? [String : AnyObject]
        }
        return nil
    }
    
    /**
     * Creates JSON structure of this `AlertMessage`.
     *
     * - parameter data: data to be processed
     *
     * - returns: JSON dictionary
     *
     * - throws: `ClientError.state` if `payload` is `nil`
     */
    override public func fromJSON(data: Foundation.Data?) throws -> Dictionary<String, Any> {
        
        let jsonDict: Dictionary<String, Any> = try super.fromJSON(data: data)
        for (key, value) in jsonDict {
            if key == "payload" {
                let payload: Dictionary? = value as? Dictionary<String, Any>
                if payload == nil {
                    throw ClientError.state(NSLocalizedString(
                        "Request Message payload must not be nil", comment: ""))
                }
                for (key1, value1) in payload! {
                    if key1 == "description" {
                        self.description = value1 as? String
                    }
                    if key1 == "severity" {
                        let ver = value1 as! String
                        let verint = (ver == "LOW") ? 4 :
                            (ver == "NORMAL") ? 3 :
                            (ver == "SIGNIFICANT") ? 2 : 1
                        self.severity = AlertMessageSeverity(rawValue: verint)
                    }
                    if key1 == "format" {
                        self.format =  value1 as? String
                    }
                    if key1 == "data" {
                        let data: Dictionary = value1 as! Dictionary<String,Any>
                        for (key2, value2) in data {
                            if value2 is String {
                                try self.items.append(DataItem(key: key2, value: value2 as? String))
                            } else {
                                try self.items.append(DataItem(key: key2, value: Double(truncating: value2 as! NSNumber)))
                            }
                        }
                    }
                }
            }
        }
        return jsonDict
    }
    
    /**
     * Set alert severity
     *
     * - parameter severity: the severity for the `AlertMessage`
     
     */
    public func setSeverity(severity: AlertMessageSeverity) {
        self.severity = severity
    }
    
    /**
     * Set the description for the alert
     *
     * - parameter description: description for the `AlertMessage`
     */
    public func setDescription(description: String) {
        self.description = description
    }
    
    /**
     * Set message format.
     *
     * - parameter format: format for the `AlertMessage`
     */
    public func setFormat(format: String) {
        self.format = format
    }
    
    /**
     * Add `Double` `DataItem`. `Key` cannot be `nil`, empty or long `String`.
     *
     * - parameter key: `String` item key.
     *
     * - parameter value: `Double` item value.
     *
     * - throws: `ClientError.argument` if `key` is `nil` or is `Double.NaN` or
     *           `Double.infinite`
     */
    public func addDataItem(key: String, value: Double) throws {
        try self.items.append(DataItem(key: key, value: value))
    }
    
    /**
     * Add `Int` `DataItem`. `Key` cannot be `nil`, empty or long `String`.
     *
     * - parameter key: `String` item key.
     *
     * - parameter value: `Int` item value.
     *
     * - Throws: IllegalDataItemKey, IllegalDataItemValue
     */
    public func addDataItem(key: String, value: Int) throws {
        try self.items.append(DataItem(key: key, value: value))
    }
    
    /**
     * Add *Boolean* `DataItem`. Key cannot be `nil`, empty or long `String`.
     *
     * - parameter key: `String` item key.
     *
     * - parameter value: *Bool* item value.
     *
     * - Throws: IllegalDataItemKey, IllegalDataItemValue
     *
     */
    public func addDataItem(key: String, value: Bool) throws {
        try self.items.append(DataItem(key: key, value: value))
    }
    
    /**
     * Add `String` `DataItem`. Key cannot be `nil`, empty or long `String`.
     *
     * - parameter key: `String` item key.
     *
     * - parameter value: `String` item value.
     *
     * - Throws: IllegalDataItemKey, IllegalDataItemValue
     */
    public func addDataItem(key: String, value: String) throws {
        try self.items.append(DataItem(key: key, value: value))
    }
    
    /**
     * Add all *DataItems* to existing ones.
     *
     * - parameter dataItems: array of `DataItem`
     */
    public func addDataItems(dataItems: [DataItem]) {
        for di in dataItems {
            self.items.append(di)
        }
    }
    
    /**
     * Creates new instance of `AlertMessage` using values from *AlertMessageBuilder*
     *
     * - returns: Instance of `AlertMessage`
     *
     * - Throws: WrongSource, WrongFormat
     */
    override public func build() throws -> AlertMessage {
        return try AlertMessage(builder: self);
        
    }
    
}


