/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * `DataMessageBuilder` extends `MessageBuilder` class. `DataMessage` class is
 * immutable. A
 * builder is required when creating `DataMessage`. `DataMessage` uses Builder
 * design pattern.
 */
public class DataMessageBuilder: MessageBuilder {
    
    /** Format for the data message */
    fileprivate var format: String?
    
    /** List of data items */
    fileprivate var items: [DataItem] = []
    
    public override init() {
        // TODO: should this not call super.init()
    }
    
    /**
     * Method to read a JSON object to a `RequestMessageBuilder`.
     *
     * - parameter data: the JSON object as `Foundation.Data`.
     *
     * - returns: the extracted item from MessageBuilder as dictionary
     *
     * - throws: `ClientError.state` if `payload` is `nil`
     */
    override public func fromJSON(data: Foundation.Data?) throws -> Dictionary<String, Any> {
        
        let jsonDict: Dictionary = try super.fromJSON(data: data)
        for (key, value) in jsonDict {
            if key == "payload" {
                let payload: Dictionary<String, Any>? = value as? Dictionary<String, Any>
                if payload == nil {
                    throw ClientError.state(NSLocalizedString(
                        "Request Message payload must not be nil", comment: ""))
                }                
                for (key1, value1) in payload! {
                    if key1 == "format" {
                        self.format = (value1 as! String)
                    }
                    if key1 == "data" {
                        let data: Dictionary<String, Any> = value1 as! Dictionary<String, Any>
                        for (key2, value2) in data {
                            switch value2 {
                            case is Bool:
                                try self.items.append(DataItem(key: key2,
                                                    value: (value2 as! Bool)))
                                break
                            case is Double:
                                try self.items.append(DataItem(key: key2,
                                                    value: (value2 as! Double)))
                                break
                            case is Int:
                                try self.items.append(DataItem(key: key2,
                                                    value: (value2 as! Int)))
                                break
                            case is String:
                                try self.items.append(DataItem(key: key2,
                                                    value: (value2 as! String)))
                                break
                            default:
                                print("something we don't know")
                                try self.items.append(DataItem(key: key2,
                                                    value: (value2 as! String)))
                            }
                        }
                    }
                }
            }
        }
        return jsonDict
        
    }

    /**
     * Set message format.
     *
     * - parameter format: Format for the `DataMessage`
     */
    public func setFormat(format: String) {
        self.format = format;
    }
    
    /**
     * Add `Double` `DataItem`. `Key` cannot be `nil`, empty or long `String`.
     *
     * - parameter key: `String` item key.
     *
     * - parameter value: `Double` item value.
     *
     * - throws: `ClientError.argument` if `key` is `nil` or is the empty 
     *           string.
     */
    public func addDataItem(key: String, value: Double) throws {
        try self.items.append(DataItem(key: key, value: value))
    }
    
    /**
     * Add `Int` `DataItem`. Key cannot be `nil`, empty or long `String`.
     *
     * - parameter key: `String` item key.
     *
     * - parameter value: `Int` item value.
     *
     * - throws: `ClientError.argument` if `key` is `nil` or is the empty
     *           string.
     */
    public func addDataItem(key: String, value: Int) throws {
        try self.items.append(DataItem(key: key, value: value))
    }
    
    /**
     * Add `Boolean` `DataItem`. `Key` cannot be `nil`, empty or long `String`.
     *
     * - parameter key: `String` item key.
     *
     * - parameter value: `Bool` item value.
     *
     * - throws: `ClientError.argument` if `key` is `nil` or is the empty 
     *           string.
     */
    public func addDataItem(key: String, value: Bool) throws {
        try self.items.append(DataItem(key: key, value: value))
    }
    
    /**
     * Add `String` `DataItem`. Key cannot be `nil`, empty or long `String`.
     *
     * - parameter key: `String` item key.
     *
     * - parameter value: `String` item value.
     *
     * - throws: `ClientError.argument` if `value` or `key` is `nil` or is the
     *                                empty string
     */
    public func addDataItem(key: String, value: String) throws {
        try self.items.append(DataItem(key: key, value: value))
    }
    
    /**
     * Add all `DataItems` to existing ones.
     *
     * - parameter dataItems: array of `DataItem`
     */
    public func addDataItems(dataItems: [DataItem]) {
        for di in dataItems {
            self.items.append(di)
        }
    }
    
    /**
     * Creates new instance of `DataMessage` using values from 
     * `DataMessageBuilder`
     *
     * - returns: Instance of `DataMessage`
     *
     * - throws: `ClientError.argument` if `builder.source` or `builder.format`
     *           is `nil` or is the empty string.
     */
    override public func build() throws -> DataMessage {
        return try DataMessage(builder: self);
        
    }
    
}

/**
 * Implementation of protocol `Equatable`
 *
 * parameter lhs: one item to be compared
 *
 * parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
public func ==(lhs: DataMessage, rhs: DataMessage) -> Bool {
    if lhs === rhs {
        return true
    }
    
    if (lhs.format != rhs.format) {
        return false
    }
    
    if (lhs.items.count != rhs.items.count) {
        return false
    }
    for i in 0..<lhs.items.count {
        if (lhs.items[i] != rhs.items[i]) {
            return false
        }
    }

    if (lhs as Message != rhs as Message) {
        return false
    }
    
    return true
}

/**
 * DataMessage extends Message class. It can be used for sending sensor data
 * to IoT Server. This class is immutable
 */
public class DataMessage: Message {

    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     *
     */
    override public var hashValue : Int {
        var result: Int = super.hashValue
        var itemsHash: Int = 0
        for item in items {
            switch item.getType() {
            case .STRING:
                do {
                    try itemsHash = itemsHash &+ DataItem(key: item.getKey(),
                                value: (item.getValue() as! String)).hashValue
                } catch {
                }
                break
            case .BOOLEAN:
                do {
                    try itemsHash = itemsHash &+ DataItem(key: item.getKey(),
                                value: (item.getValue() as! Bool)).hashValue
                } catch {
                }
                break
            case .DOUBLE:
                do {
                    try itemsHash = itemsHash &+ DataItem(key: item.getKey(),
                        value: Double(truncating: item.getValue() as! NSNumber)).hashValue
                } catch {
                }
                break
            }
        }
        result = 31 &* result &+ itemsHash
        return result
    }
    
    /**
     * Format for the data message
     */
    fileprivate var format: String?
    
    /** Actual data for the data message */
    fileprivate var items: [DataItem] = []
    
    /**
     * `DataMessage` constructor takes `DataMessage.Builder` and set values to 
     * each field. If the value is `nil`, set a default value. Format cannot be
     * `nil` or empty.
     *
     * - parameter builder: DataMessageBuilder
     *
     * - throws: `ClientError.state` if `builder.source` or `builder.format`
     *           is `nil` or is the empty string.
     */
    fileprivate init(builder: DataMessageBuilder) throws {
        try super.init(builder: builder)
        
        guard let format = builder.format, !format.isEmpty else {
            throw ClientError.state(NSLocalizedString(
                "Format cannot be nil or empty", comment: ""))

        }
        // TODO: Should this be modified if items is empty ?
        self.format = format
        
        if builder.items.isEmpty {
            throw ClientError.state(NSLocalizedString("Data field is empty",
                                                      comment: ""))
        }
        
        for di in builder.items {
            items.append(di)
        }
    }
    
    /**
     * Get the name of the format for data.
     *
     * - returns: data format, never `nil`
     */
    public func getFormat() -> String {
        return self.format!
    }
    
    /**
     * Get message payload as array of `DataItems`
     *
     * - returns: Payload that is carried by `AlertMessage`, never `nil`
     */
    public func getDataItems() -> [DataItem] {
        return self.items
    }
    
    /**
     * Get type message type.
     *
     * - returns: type, never `nil`
     */
    override public func getType() -> MsgType {
        return MsgType.DATA
    }
    
    /**
     * Exports data from `DataMessage` to string using JSON interpretation of 
     * the message.
     *
     * - returns: JSON interpretation of the message as string
     */
    override public func toString() -> String {
        return toJsonString()
    }
    
    /**
     * Export `DataMessage` as JSON string
     *
     * - returns: `DataMessage` as JSON string
     */
    override public func toJsonString() -> String {
        var result: String
        var addings: String = ""
        
        addings = addings + "\"payload\":{"
        addings = addings + "\"format\":\"" + format! + "\","
        addings = addings + "\"data\":{"
        var z: Int = 0
        for item in items {
            addings = addings + "\"" + item.getKey() + "\":"
            switch item.getType() {
            case .STRING:
                addings = addings + "\"" + (item.getValue() as! String) + "\""
                break
            case .BOOLEAN:
                addings = addings + (item.getValue() as! Bool).description
                break
            case .DOUBLE:
                addings = addings + Double(truncating: item.getValue() as! NSNumber).description
                break
            }
            z += 1
            if z < items.count {
                addings = addings + ","
            }
        }
        addings = addings + "}}"
        
        result = toJsonString(expand: true, additionalItems: addings)
        
        return result
    }
    
}
