/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * IndirectActivationRequest
 */
class IndirectActivationRequest {
    
    /**
     * The hardware id of the device
     */
    private static let HARDWARE_ID: String = "hardwareId"
    
    private static let FIELD_DEVICE_MODELS: String  = "deviceModels"
    
    private static let SIGNATURE = "signature"
    
    private var hardwareId: String?
    private var metadata: [String : String]?
    private var deviceModels: [String]?
    private var signature:String?
    
    init() {
    }
    
    init(hardwareId: String, metadata: [String : String],
         deviceModels: [String], signature:String? = nil ) {
        self.hardwareId = hardwareId
        self.metadata = metadata
        self.deviceModels = deviceModels
        self.signature = signature
    }
    
    func getHardwareId() -> String {
        return self.hardwareId!
    }
    
    func setHardwareId(hardwareId: String) {
        self.hardwareId = hardwareId;
    }
    
    func getDeviceModels() -> [String] {
        return self.deviceModels!
    }
    
    func setDeviceModels(deviceModels: [String]) {
        self.deviceModels = deviceModels
    }
    
    func getMetadata() -> [String : String] {
        return metadata!
    }
    
    private static func parseJsonStr(str: String) throws -> [String : AnyObject]? {
        if let data = str.data(using: .utf8) {
            return try JSONSerialization.jsonObject(with: data, options: []) as? [String : AnyObject]
        }
        return nil
    }
    
    static func fromJson(jsonStr: String) throws -> IndirectActivationRequest? {
        let jsonDict: [String: AnyObject]? = try parseJsonStr(str: jsonStr)
        let result: IndirectActivationRequest = IndirectActivationRequest()
        return try IndirectActivationRequest.fromJson(jsonDict: jsonDict!, iar: result)
    }
    
    static func fromJson(jsonDict: [String : AnyObject], iar: IndirectActivationRequest) throws -> IndirectActivationRequest? {
        var metadata = Dictionary<String, String>()
        for (key, val) in jsonDict {
            if key == IndirectActivationRequest.HARDWARE_ID {
                iar.hardwareId = val.string
            } else {
                metadata[key] = val.string
            }
        }
        
        let deviceModelArray: [String]? = jsonDict[IndirectActivationRequest.FIELD_DEVICE_MODELS]! as? Array<String>
        if deviceModelArray != nil {
            var deviceModels : [String] = []
            for deviceModel in deviceModelArray! {
                deviceModels.append(deviceModel)
            }
            iar.setDeviceModels(deviceModels: deviceModels)
        }
        iar.signature = jsonDict[IndirectActivationRequest.SIGNATURE] as! String?
        return iar
    }
    
    func fromJson(data: Foundation.Data?)  throws {
        if data == nil {
            return
        }
        let jsonStr: String = String(data: data!, encoding: .utf8)!
        let jsonDict: [String: AnyObject]? = try IndirectActivationRequest.parseJsonStr(str: jsonStr)
        let _ = try IndirectActivationRequest.fromJson(jsonDict: jsonDict!, iar: self)
    }
    
    func toJsonString() -> String? {
        var resultStr: String = "{"
        resultStr = resultStr + "\"" + IndirectActivationRequest.HARDWARE_ID + "\":\"" + self.hardwareId! + "\""
        if metadata != nil {
            for (key, val) in metadata! {
                // TODO what if val is no string??
                resultStr = resultStr + ",\"" + key + "\":\"" + val + "\""
            }
        }
        resultStr = resultStr + ",\"" + IndirectActivationRequest.FIELD_DEVICE_MODELS + "\":["
        var z: Int = 0
        for dm in deviceModels! {
            resultStr = resultStr + "\"" + dm + "\""
            z += 1
            if z < deviceModels!.count {
                resultStr = resultStr + ","
            }
        }
        resultStr = resultStr + "]"
        if (signature != nil) {
            resultStr = resultStr + ",\"" + IndirectActivationRequest.SIGNATURE +
            "\":\"" + signature! + "\""
        }
        return resultStr + "}"
    }
    
    func toString() -> String? {
        var result: String = "IndirectActivationRequest"
        result = result + "{" + "hardwareId=" + self.hardwareId!
        result = result + ", deviceModels=["
        var i = 0
        for dm in deviceModels! {
            result = result + dm
            i += 1
            if i < deviceModels!.count {
                result = result + ", "
            }
        }
        result = result + "]"
        if (metadata != nil) {
            for (key, value) in metadata! {
                result = result + ", " + key + " = " + value
            }
        }
        if (signature != nil) {
            result = result + ", signature=" + self.signature!
        }
        result = result + "}"
        return result
    }
}
