/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * The MessageDispatcher queues messages for automatic dispatching to the
 * for the Oracle IoT Cloud Service.
 *
 * The MessageDispatcher prioritizes message dispatching to ensure
 * high priority messages are dispatched ahead of lower priority messages.
 * There is one MessageDispatcher instance per client.
 */
public class MessageDispatcher {
    
    internal static var dispatcherMap: [AdvancedDirectlyConnectedDevice : MessageDispatcher] = [:]
    
    /**
     * Get the instance of a `MessageDispatcher` for the Client.
     *
     * - parameter msgDirectlyConnectedDevice: is a gateway or directly 
     *                                         connected device implementation
     *
     * - returns: A MessageDispatcher for the Client.
     * 
     * - throws: `ClientError.network` if `MessageDispatcherImpl.init()` cannot 
     *            register resources
     */
    public static func getMessageDispatcher(msgDirectlyConnectedDevice: AdvancedDirectlyConnectedDevice) throws -> MessageDispatcher {
        var messageDispatcher:MessageDispatcher? = dispatcherMap[msgDirectlyConnectedDevice]
        if messageDispatcher == nil {
            messageDispatcher = try MessageDispatcherImpl(deviceClient: msgDirectlyConnectedDevice)
            dispatcherMap[msgDirectlyConnectedDevice] = messageDispatcher
        }
        return messageDispatcher!
    }

    /**
     * A callback interface for successful delivery of messages.
     *
     * - (Abstract class)
     *
     */
    open class DeliveryCallback {
        public init() {
        }
        
        /**
         * Notify that messages have been successfully dispatched. This callback
         * indicates that the messages have been delivered to, and accepted by,
         * the server and are no longer queued.
         *
         * - (Abstract method)
         *
         * - parameter messages: the messages that were delivered
         */
        open func delivered(messages: [Message]) {
            fatalError("Subclass need to implement the 'delivered()' method.")
        }
    }
    
    /**
     * A callback interface for errors in delivery of messages.
     *
     * - (Abstract class)
     *
     */
    open class ErrorCallback {
        public init() {
        }
        
        /**
         * Notify that an error occurred when attempting to deliver messages
         * to the server. This callback indicates that the messages have
         * not been delivered to the server and are no longer queued.
         *
         * - (Abstract method)
         *
         * - parameter messages: the messages that were not delivered
         */
        open func failed(messages: [Message], error:ClientError? = nil) throws {
            fatalError("Subclass need to implement the 'failed()' method.")
        }
    }
    
    /**
     * Get the RequestDispatcher that is used by this MessageDispatcher for
     * handling RequestMessages.
     *
     * - (Abstract method)
     *
     * - returns: a RequestDispatcher
     *
     * - see: RequestDispatcher
     */
    public func getRequestDispatcher() -> RequestDispatcher {
        fatalError("Subclass need to implement the 'getRequestDispatcher()' method.")
    }
    
    /**
     * Add the message to the outgoing message queue if it is possible to
     * do so without violating capacity restrictions.
     *
     * - (Abstract method)
     *
     * - parameter message: The message to be queued
     */
    public func queue(message: Message) throws {
        fatalError("Subclass need to implement the 'queue(Message)' method.")
    }
    
    /**
     * Set a callback to be notified if there is an error in dispatching.
     *
     * - (Abstract method)
     *
     * - parameter callback: callback to invoke
     */
    public func setOnError(callback: ErrorCallback) {
        fatalError("Subclass need to implement the 'setOnError(ErrorCallback)' method.")
    }
    
    /**
     * Set a callback to be notified if message is successfully delivered.
     *
     * - (Abstract method)
     *
     * - parameter callback: callback to invoke
     */
    public func setOnDelivery(callback: DeliveryCallback) {
        fatalError("Subclass need to implement the 'setOnDelivery(DeliveryCallback)' method.")
    }
    
    /**
     * Closes the MessageDispatcher.
     *
     * - (Abstract method)
     *
     * - parameter callback: callback to invoke
     */
    public func close(callback: @escaping () -> Void) {
        fatalError("Subclass need to implement the 'close()' method.")
    }
}
