/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}


/*************************************************************************
 *
 * Handling of the blocking queue, servicing the queue,
 * and dispatching dequeued entries.
 *
 *************************************************************************/
class Transmitter : NSObject {
    
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
        super.init()
    }
    
    @objc func run() {
        self.impl.group.enter()
        self.impl.transmitQueue.async(execute: {
            while (true) {
                self.impl.messageQueued.lock()
                while (!self.impl.requestClose && self.impl.outgoingMessageQueue.isEmpty) {
                    // transmit dispatch queue blocks waiting for an outgoing message
                    self.impl.messageQueued.wait(until: Date().addingTimeInterval(30.0))
                }
                self.impl.messageQueued.unlock()
                // TODO: Shouldn't this be in the lock ?
                if (self.impl.requestClose && self.impl.outgoingMessageQueue.isEmpty) {
                    self.impl.group.leave()
                    break
                }
                
                var messageList: [Message]? = nil
                var errorList:[Message] = [Message]()
                self.impl.messageQueue.sync(execute: {
                    let size: Int = min(
                        self.impl.outgoingMessageQueue.count,
                        self.impl.maximumMessagesPerConnection!)
                    
                    if (size > 0) {
                        messageList = []
                    } else {
                        return
                    }
                    
                    var inProgressSources:Set<String> = Set<String>()
                    var waitList:[Message] = [Message]()
                    var sendList:[Message] = [Message]()

                    for _ in 0..<size {
                        let m:Message = self.impl.outgoingMessageQueue.pop()!
                        let clientId:String = m.getClientId();
                        let source:String = m.getSource();
                        
                        if self.impl.failedContentIds.contains(clientId) {
                            errorList.append(m);
                            continue;
                        }
                        
                        if m.getType() == Message.MsgType.RESPONSE ||
                                !(inProgressSources.contains(source) ||
                                    self.impl.isContentDependent(clientId: clientId)) {
                            sendList.append(m);
                            if sendList.count == size {
                                break;
                            }
                        } else {
                            inProgressSources.update(with: source)
                            waitList.append(m)
                        }
                    }
                    if sendList.count > 0 {
                        messageList = sendList
                    }
                    for m:Message in waitList {
                        let _ = self.impl.outgoingMessageQueue.push(new: m)
                    }
                })
                
                if (messageList != nil) {
                    self.dispatch(messages: messageList)
                }
                if !errorList.isEmpty && self.impl.errorCallback != nil {
                    do {
                        try self.impl.errorCallback!.failed(messages: errorList,
                            error: ClientError.network(NSLocalizedString(
                                "Content sync failed", comment: "")))
                    } catch {
                        print("errorcallback for " + NSLocalizedString(
                            "Content sync failed", comment: "") +
                            " threw exception")
                    }
                }
            }
        })
    }
    
    private func handleDispatchError(msgErr: String, messages: [Message]?) {
        if (impl.errorCallback != nil) {
            do {
                try impl.errorCallback!.failed(messages: messages!)
            } catch {
                print(msgErr + ": errorCallback threw exception " +
                                                    error.localizedDescription)
            }
        }
        impl.counterQueue.sync {
            impl.totalProtocolErrors = impl.totalProtocolErrors! + 1
        }
    }

    private func dispatch(messages: [Message]?) {
        do {
            self.impl.group.enter()
            try impl.deviceClient!.send(messages: messages, callback: {error in
                if (error != nil) {
                    self.handleDispatchError(
                        msgErr: "Message transmit dispatcher 1",
                        messages: messages!)
                } else {
                    do {
                        if self.impl.deliveryCallback != nil {
                            self.impl.deliveryCallback!.delivered(messages: messages!)
                        }
                        
                        try self.impl.counterQueue.sync(execute: {
                            self.impl.totalMessagesSent =
                                self.impl.totalMessagesSent! + messages!.count
                            
                            var jsonStr: String = ""
                            for message in messages! {
                                try jsonStr = jsonStr + message.toJsonString()
                            }
                            self.impl.totalBytesSent = self.impl.totalBytesSent! +
                                Int64((jsonStr.data(using: .utf8)?.count)!)
                        })
                    } catch {
                        self.handleDispatchError(
                                        msgErr: "Message transmit dispatcher 2",
                                        messages: messages!)
                    }
                }
                self.impl.group.leave()
            })
        } catch {
            handleDispatchError(msgErr: "Message transmit dispatcher 3",
                                messages: messages!)
            self.impl.group.leave()
        }
    }
}

class Dispatcher {
    var requestMessage: RequestMessage?
    var mdi: MessageDispatcherImpl
    
    init(requestMessage: RequestMessage, mdi: MessageDispatcherImpl) {
        self.requestMessage = requestMessage
        self.mdi = mdi
    }

    func run() {
        do {
            try mdi.requestDispatcher!.dispatch(requestMessage: requestMessage,
                        callback: { (responseMessage, error) in
                do {
                    if error == nil {
                        try self.mdi.queue(message: responseMessage!)
                    } else {
                        print(error!)
                    }
                } catch {
                    print("Issue while dispatching: \(error)")
                }
            })
        } catch {
            print("Issue while dispatching: \(error)")
        }
    }
}

class Receiver : NSObject {
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
        super.init()
    }
    
    @objc func run() {
        self.impl.group.enter()
        self.impl.receiveQueue.async(execute: {
            
            while (!self.impl.requestClose) {
                if (Config.DISABLE_LONG_POLLING) {
                    sleep(UInt32(Config.MESSAGE_POLLING_INTERVAL/1000))
                }
                do {
                    try self.impl.deviceClient!.receive(timeout: -1, callback: { requestMessage, error in
                        defer {
                            self.impl.receiveSem.signal()
                        }
                        if (error != nil) {
                            // TODO: Can cause infinite loop if network receive
                            // error does not get corrected.
                            print("Message receive returned error: \(error!)")
                        } else if (requestMessage != nil) {
                            self.impl.counterQueue.sync(execute: {
                                self.impl.totalMessagesReceived = self.impl.totalMessagesReceived! + 1
                                self.impl.totalBytesReceived = self.impl.totalBytesReceived! +
                                    Int64(((requestMessage!.toJsonString()).data(using: String.Encoding(rawValue: String.Encoding.utf8.rawValue))?.count)!)
                            })
                            if (!self.impl.requestClose) {
                                //new reference for later use when processed from the queue as requestMessage changes in the loop
                                let msg = requestMessage
                                Dispatcher(requestMessage: msg!, mdi: self.impl).run()
                            }
                        }
                    })
                } catch {
                    print("Message receive dispatcher threw exception : \(error)")
                    self.impl.receiveSem.signal()
                }
                self.impl.receiveSem.wait()
            }
            self.impl.group.leave()
        })
    }
}

class CounterHandler: RequestHandler {
    
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
    }
    
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        if (requestMessage == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Request must not be nil", comment: ""))
        }
        
        var status: StatusCode
        if (requestMessage!.getMethod() != nil) {
            if ("GET" == requestMessage!.getMethod()?.uppercased()) {
                do {
                    var jobDict: [String : AnyObject] = [ : ]
                    var jsonBody: String = "{"
                    var size: Float = Float(0)
                    impl.messageQueue.sync(execute: {
                        size = Float(impl.outgoingMessageQueue.count)
                    })
                    let max: Float = Float(impl.maximumQueueSize!)
                    // we only want 1 decimal point precision
                    let load: Int = Int(size / max * 1000)
                    self.impl.counterQueue.sync(execute: {
                        jobDict["load"] = Float(load / 10) as AnyObject?
                        jsonBody = jsonBody + "\"load\":" + (Float(load / 10)).description
                        jobDict["totalBytesSent"] = impl.totalBytesSent! as AnyObject?
                        jsonBody = jsonBody + ",\"totalBytesSent\":" + impl.totalBytesSent!.description
                        jobDict["totalBytesReceived"] = impl.totalBytesReceived! as AnyObject?
                        jsonBody = jsonBody + ",\"totalBytesReceived\":" + impl.totalBytesReceived!.description
                        jobDict["totalMessagesReceived"] = impl.totalMessagesReceived as AnyObject?
                        jsonBody = jsonBody + ",\"totalMessagesReceived\":" + impl.totalMessagesReceived!.description
                        jobDict["totalMessagesRetried"] = impl.totalMessagesRetried as AnyObject?
                        jsonBody = jsonBody + ",\"totalMessagesRetried\":" + impl.totalMessagesRetried!.description
                        jobDict["totalMessagesSent"] = impl.totalMessagesSent as AnyObject?
                        jsonBody = jsonBody + ",\"totalMessagesSent\":" + impl.totalMessagesSent!.description
                        jobDict["totalProtocolErrors"] = impl.totalProtocolErrors! as AnyObject?
                        jsonBody = jsonBody + ",\"totalProtocolErrors\":" + impl.totalProtocolErrors!.description
                        jsonBody = jsonBody + "}"
                    })
                    let rmb: ResponseMessageBuilder =
                        ResponseMessageBuilder(requestMessage: requestMessage!)
                    rmb.setStatusCode(statusCode: .OK)
                    rmb.setBody(body: jsonBody)
                    return try rmb.build()
                } catch {
                    status = .BAD_REQUEST;
                }
            } // GET
            else  {
                //Unsupported request method
                status = .METHOD_NOT_ALLOWED;
            }
        }
        else {
            //Unsupported request method
            status = .METHOD_NOT_ALLOWED;
        }
        let rmb: ResponseMessageBuilder =
            ResponseMessageBuilder(requestMessage: requestMessage!)
        rmb.setStatusCode(statusCode: status)
        return try rmb.build()
    }
}


class ResetHandler: RequestHandler {
    
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
    }
    
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        if (requestMessage == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Request must not be nil", comment: ""))
        }
        
        var status: StatusCode
        if (requestMessage!.getMethod() != nil) {
            if ("PUT" == requestMessage!.getMethod()?.uppercased()) {
                impl.totalBytesSent = 0
                impl.totalBytesReceived = 0
                impl.totalProtocolErrors = 0
                impl.totalMessagesSent = 0
                impl.totalMessagesReceived = 0
                impl.totalMessagesRetried = 0
                status = .OK
            }
            else  {
                //Unsupported request method
                status = .METHOD_NOT_ALLOWED
            }
        }
        else {
            status = .METHOD_NOT_ALLOWED
        }
        let rmb: ResponseMessageBuilder =
            ResponseMessageBuilder(requestMessage: requestMessage!)
        rmb.setStatusCode(statusCode: status)
        return try rmb.build()
    }
}

class PollingIntervalHandler: RequestHandler {
    
    private var impl: MessageDispatcherImpl
    
    init(impl: MessageDispatcherImpl) {
        self.impl = impl
    }
    
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        
        if (requestMessage == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Request must not be nil", comment: ""))
        }
        
        var status: StatusCode
        if (requestMessage!.getMethod() != nil) {
            if ("GET" == requestMessage!.getMethod()?.uppercased()) {
                do {
                    let jobStr: String = "{\"pollingInterval\":" +
                        (impl.pollingInterval! / 1000).description + "}"
                    
                    let rmb: ResponseMessageBuilder =
                        ResponseMessageBuilder(requestMessage: requestMessage!)
                    rmb.setStatusCode(statusCode: .OK)
                    rmb.setBody(body: jobStr)
                    let messRet = try rmb.build()
                    return messRet
                } catch {
                    status = .BAD_REQUEST;
                }
            }
            else if ("PUT" == requestMessage!.getMethod()?.uppercased()) {
                do {
                    let errors: String = ""
                    // Use Long objects here in case there was a problem 
                    // getting the parameters.
                    impl.pollingInterval = impl.getParam(data: requestMessage!.getBody() as Foundation.Data,
                                    paramName: "value", errors: errors) * 1000
                    
                    // TODO: What's this for errors is "" ?
                    // Is errors supposed to be "inout" ?
                    // Also not that getParam always returns 0
                    if (errors.count != 0) {
                        return try impl.getBadRequestResponse(requestMessage: requestMessage!,
                                        message: impl.endpointId, src: errors)
                    }
                    status = .OK;
                } catch {
                    status = .BAD_REQUEST
                }
            }
            else  {
                //Unsupported request method
                status = .METHOD_NOT_ALLOWED
            }
        }
        else {
            status = .METHOD_NOT_ALLOWED
        }
        let rmb: ResponseMessageBuilder = ResponseMessageBuilder(requestMessage: requestMessage!)
        rmb.setStatusCode(statusCode: status)
        return try rmb.build()
    }
}

/**
 * MessageDispatcherImpl
 */
class MessageDispatcherImpl: MessageDispatcher {
    
    private static var DEFAULT_MAXIMUM_MESSAGES_TO_QUEUE: Int = 1000
    private static var DEFAULT_MAXIMUM_MESSAGES_PER_CONNECTION: Int = 100
    private static var DEFAULT_POLLING_INTERVAL: Int64 = 1000 // milliseconds
    
    private static let MAXIMUM_MESSAGES_TO_QUEUE_PROPERTY: String =
        "oracle.iot.client.device.dispatcher_max_queue_size"
    private static let MAXIMUM_MESSAGES_PER_CONNECTION_PROPERTY: String  =
        "oracle.iot.client.device.dispatcher_max_messages_per_connection"
    private static let POLLING_INTERVAL_PROPERTY: String  =
        "oracle.iot.client.device.dispatcher_polling_interval"
    
    
    static let backslash = "\\"
    
    static let COUNTERS_URL: String? = "deviceModels" + "/" + ActivationManager.MESSAGE_DISPATCHER_URN + "/counters"
    static let RESET_URL: String? = "deviceModels" + "/" + ActivationManager.MESSAGE_DISPATCHER_URN + "/reset"
    static let POLLING_INTERVAL_URL: String? = "deviceModels" + "/" + ActivationManager.MESSAGE_DISPATCHER_URN + "/pollingInterval"
    static let DIAGNOSTICS_URL: String? = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/info"
    static let TEST_CONNECTIVITY_URL: String? = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/testConnectivity"
    
    static let DIAGNOSTICS_URL_STATUS: String? = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/testConnectivity/status"
    static let DIAGNOSTICS_URL_START: String? = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/testConnectivity/start"
    static let DIAGNOSTICS_URL_STOP: String? = "deviceModels" + "/" + ActivationManager.DIAGNOSTICS_URN + "/testConnectivity/stop"
    

    
    /**
     * Serial dispatch queues for message dispatching activities
     */
    let messageQueue  = DispatchQueue(label: "com.oracle.client.impl.device.messageQueue",
        qos: .utility)
    let counterQueue  = DispatchQueue(label: "com.oracle.client.impl.device.counterQueue",
        qos: .utility)
    let transmitQueue = DispatchQueue(label: "com.oracle.client.impl.device.transmitQueue",
        qos: .utility)
    let receiveQueue  = DispatchQueue(label: "com.oracle.client.impl.device.receiveQueue",
        qos: .utility)
    let group = DispatchGroup()
    
    private var counterHandler: RequestHandler?
    private var resetHandler: RequestHandler?
    
    private var pollingIntervalHandler: RequestHandler?
    private var diagnosticsHandler: RequestHandler?
    private var testConnectivityHandler: RequestHandler?
    
    fileprivate var contentMap:Dictionary<StorageObject, Set<String>>!
    fileprivate var failedContentIds:Set<String>!

    var transmitter: Transmitter?
    var receiver: Receiver?
    let receiveSem = DispatchSemaphore(value:0)
    
    /*
     * Queue of outgoing messages to be dispatched to server
     * (with thread for servicing the queue).
     * Note: it has package access for unit test access.
     */
    var outgoingMessageQueue: PriorityQueue<Message>
    
    private var contentLock:NSRecursiveLock = NSRecursiveLock()
    private var closeLock: AnyObject?
    fileprivate var messageQueued:NSCondition = NSCondition()
    
    /*
     * maximum number of messages allowed in the outgoing message queue
     */
    fileprivate var maximumQueueSize: Int?
    
    /*
     * maximum number of messages to send in one connection
     */
    fileprivate var maximumMessagesPerConnection: Int?
    
    // Counter indicating total number of messages that have been delivered
    fileprivate var totalMessagesSent: Int?
    
    // Counter indicating total number of messages that have been delivered
    fileprivate var totalMessagesReceived: Int?
    
    // Counter indicating total number of messages that were retried for delivery
    fileprivate var totalMessagesRetried: Int?
    
    // Counter indicating total number of bytes that have been delivered
    fileprivate var totalBytesSent: Int64?
    
    // Counter indicating total number of bytes that have been delivered
    fileprivate var totalBytesReceived: Int64?
    
    // Counter indicating total number of protocol errors.
    fileprivate var totalProtocolErrors: Int64?
    
    fileprivate var endpointId: String
    
    /*
     * Maximum time in milliseconds to wait for a message to be queued
     * before the send thread will poll the server
     */
    fileprivate var pollingInterval: Int64?
    
    fileprivate var deviceClient: AdvancedDirectlyConnectedDevice?
    
    var requestDispatcher: RequestDispatcher?
    
    /*
     * Set to true if close method has been called.
     */
    fileprivate var closed: Bool = false
    fileprivate var requestClose: Bool = false
    
    fileprivate var deliveryCallback: MessageDispatcher.DeliveryCallback?
    fileprivate var errorCallback: MessageDispatcher.ErrorCallback?
    
    private let mesDispTime:Double = 1.0
    
    private var resourceMessage:ResourceMessage?
    
    override func setOnDelivery(callback: MessageDispatcher.DeliveryCallback) {
        self.deliveryCallback = callback
    }
    
    override func setOnError(callback: MessageDispatcher.ErrorCallback) {
        self.errorCallback = callback
    }
    
    override func getRequestDispatcher() -> RequestDispatcher{
        return requestDispatcher!
    }
// ------------------------------------------
    init(deviceClient: AdvancedDirectlyConnectedDevice) throws {
        
        self.closed = false
        self.requestClose = false
        
        self.deviceClient = deviceClient
        self.requestDispatcher = RequestDispatcher()
        
        self.pollingInterval = MessageDispatcherImpl.getPollingInterval()
        self.maximumQueueSize = MessageDispatcherImpl.getQueueSize()
        self.maximumMessagesPerConnection =
            MessageDispatcherImpl.getMaximumMessagesPerConnection();
        
        self.endpointId = deviceClient.getEndpointId()
        
        self.contentMap = Dictionary<StorageObject, Set<String>>()
        self.failedContentIds = Set<String>()
        
        
        self.outgoingMessageQueue = PriorityQueue<Message>(
            compare: {
                // Note this implementation is not consistent with equals. It 
                // is possible that a.compareTo(b) == 0 is not that same 
                // boolean value as a.equals(b)
                
                // The natural order of enum is the enum's ordinal, i.e.,
                // x.getPriority().compareTo(y.getPriority() will give {x,y}
                // if x is a lower priority. What we want is to sort by the
                // higher priority.
                return ($0.getPriority().rawValue < $1.getPriority().rawValue)
                    // If they are the same priority, take the one that was 
                    // created first
                    || ($0.getEventTime() < $1.getEventTime())
                    
                    // If they are still the same, take the one with higher 
                    //reliability.
                    || ($0.getReliability().rawValue < $1.getReliability().rawValue)
        })
        
        self.totalMessagesSent = 0
        self.totalMessagesReceived = 0
        self.totalMessagesRetried = 0
        self.totalBytesSent = 0
        self.totalBytesReceived = 0
        self.totalProtocolErrors = 0
        
        transmitter = nil
        receiver = nil
        
        super.init()
        
        let counterHandler: RequestHandler = CounterHandler(impl: self)
        
        let resetHandler: RequestHandler = ResetHandler(impl: self)
        
        /*
         * resource definition for pollingInterval: {
         *   "get": {
         *     "schema": {
         *       "properties": {
         *         "value": {
         *           "type": "number",
         *           "description": "The incoming message polling interval
         *              in seconds on the directly connected device."
         *         }
         *       }
         *     },
         *   "put" : {
         *     "parameters": [
         *       {"name": "value",
         *        "type": "integer",
         *        "in": "body",
         *        "description": "The incoming message polling interval in
         *          seconds.",
         *        "required": "true"},
         *       ]
         *   }
         */
        self.pollingIntervalHandler = PollingIntervalHandler(impl: self)
        
        do {
            diagnosticsHandler = Diagnostics()
            let testConnectivity: TestConnectivity =
                TestConnectivity(endpointId: endpointId, messageDispatcher: self)
            testConnectivityHandler = testConnectivity.getTestConnectivityHandler()
            let rmb: ResourceMessageBuilder = ResourceMessageBuilder()
            rmb.setEndpointName(endpointName: endpointId)
            rmb.setSource(source: endpointId)
            rmb.setSender(sender: endpointId)
            // build a list String that can be hashed
            var stringList:[String] = Array()
            stringList.append(MessageDispatcherImpl.COUNTERS_URL!)
            stringList.append(MessageDispatcherImpl.RESET_URL!)
            stringList.append(MessageDispatcherImpl.POLLING_INTERVAL_URL!)
            stringList.append(MessageDispatcherImpl.DIAGNOSTICS_URL!)
            stringList.append(MessageDispatcherImpl.DIAGNOSTICS_URL_STATUS!)
            stringList.append(MessageDispatcherImpl.DIAGNOSTICS_URL_START!)
            stringList.append(MessageDispatcherImpl.DIAGNOSTICS_URL_STOP!)
            let reconMark = ResourceMessage.getMD5ofList(resStrings: stringList)
            // calculate the reconciliation mark as MD5 hash over all path information (in the same order)
            rmb.setReconciliationMark(rM: reconMark!)
            //
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "counters",
                                             path: MessageDispatcherImpl.COUNTERS_URL!,
                                             requestHandler: counterHandler,
                                             methods: Array(arrayLiteral: Method.GET)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "reset",
                                             path: MessageDispatcherImpl.RESET_URL!,
                                             requestHandler: resetHandler,
                                             methods: Array(arrayLiteral: Method.PUT)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "polling_interval",
                                             path: MessageDispatcherImpl.POLLING_INTERVAL_URL!,
                                             requestHandler: pollingIntervalHandler!,
                                             methods: Array(arrayLiteral: Method.GET, Method.PUT)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "dignostic_info",
                                             path:MessageDispatcherImpl.DIAGNOSTICS_URL!,
                                             requestHandler: diagnosticsHandler!,
                                             methods: Array(arrayLiteral: Method.GET)))
            // additional resources like in the Posix API
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "diagnostic_connectivity_status",
                                             path: MessageDispatcherImpl.DIAGNOSTICS_URL_STATUS!,
                                             requestHandler: testConnectivityHandler!,
                                             methods: Array(arrayLiteral: Method.GET, Method.PUT)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "diagnostic_connectivity_start",
                                             path: MessageDispatcherImpl.DIAGNOSTICS_URL_START!,
                                             requestHandler: testConnectivityHandler!,
                                             methods: Array(arrayLiteral: Method.GET, Method.PUT)))
            try rmb.register(ri: getResource(endpointId: endpointId,
                                             name: "diagnostic_connectivity_stop",
                                             path: MessageDispatcherImpl.DIAGNOSTICS_URL_STOP!,
                                             requestHandler: testConnectivityHandler!,
                                             methods: Array(arrayLiteral: Method.GET, Method.PUT)))
            self.resourceMessage = try rmb.build()
            // TODO: handle error
        } catch {
            throw ClientError.network(NSLocalizedString(
                "Illegal Resource Message", comment: ""))
        }
        
        transmitter = Transmitter(impl: self)
        receiver = Receiver(impl: self)
        
        self.transmitter!.run()
        self.receiver!.run()
        // TODO: IOT-37051
        // Fails if device not active, Does this message need to be sent ?
        // It is "diagnostic_connectivity_stop"
        try self.queue(message: self.resourceMessage!)
    }
    
    private static func getQueueSize() -> Int {
        let size: Int = MessageDispatcherImpl.DEFAULT_MAXIMUM_MESSAGES_TO_QUEUE
        // size must be at least 1 or PriorityQueue constructor will throw 
        // exception
        return (size > 0 ? size : MessageDispatcherImpl.DEFAULT_MAXIMUM_MESSAGES_TO_QUEUE)
    }
    
    private static func getMaximumMessagesPerConnection() -> Int {
        let max: Int = MessageDispatcherImpl.DEFAULT_MAXIMUM_MESSAGES_PER_CONNECTION
        // size must be at least 1
        return (max > 0 ? max : MessageDispatcherImpl.DEFAULT_MAXIMUM_MESSAGES_PER_CONNECTION)
    }
    
    private static func getPollingInterval() -> Int64 {
        let interval: Int64 = MessageDispatcherImpl.DEFAULT_POLLING_INTERVAL
        // polling interval may be zero, which means wait forever
        return (interval >= 0 ? interval : MessageDispatcherImpl.DEFAULT_POLLING_INTERVAL)
    }
    
    /**
     * Returns the Int64 parameter specified by `paramName` in the request if 
     * it's available.
     *
     * - parameter data: a `Foundation.Data` object containing the request data.
     *
     * - parameter paramName: the name of the parameter to get.
     *
     * - parameter errors: a `String` of errors.  Any errors produced from
     *                     retrieving the parameter will be appendeto this.
     *
     * - returns: the parameter if it can be retrieved, otherwise `nil`
     */
    fileprivate func getParam(data: Foundation.Data, paramName: String,
                              errors: String) -> Int64 {
       return 0
    }
    
    /**
     * Appends spaces (error separators) to the `errors` list if it's not empty.
     *
     * - parameter errors: a list of errors in a String
     */
    private func appendSpacesToErrors(errors: String) -> String {
        if !errors.isEmpty {
            return errors + "  "
        } else {
            return errors
        }
    }
    
    /**
     * Returns an appropriate response if the request is bad.
     *
     * - parameter requestMessage: the request for this capability.
     *
     * - returns an appropriate response if the request is bad.
     */
    fileprivate func getBadRequestResponse(requestMessage: RequestMessage,
                    message: String, src: String) throws -> ResponseMessage {
        return try getResponseMessage(requestMessage: requestMessage,
                                      body: message, statusCode: .BAD_REQUEST)
    }
    
    /**
     * Returns a *RequestMessage* with the supplied parameters.
     *
     * - parameter requestMessage: the request for this capability.
     *
     * - parameter body:           the body for the response.
     *
     * - parameter statusCode:     the status code of the response.
     *
     * - returns: an appropriate response if the request is *nil*
     */
    private func getResponseMessage(requestMessage: RequestMessage, body: String, statusCode: StatusCode) throws -> ResponseMessage {
        let rmb: ResponseMessageBuilder = ResponseMessageBuilder(requestMessage: requestMessage)
        rmb.setBody(body: body)
        rmb.setStatusCode(statusCode: statusCode)
        return try rmb.build()
    }
    
    // Get current software version
    private func getSoftwareVersion() -> String {
        // TODO: Need better mechanism for obtaining the version
        return Config.VERSION!
    }
    // TODO: Use Dispatch queues instead of locks.
    public func addStorageObjectDependency(storageObject:StorageObject, clientMsgId:String) {
        contentLock.lock();
        defer {
            contentLock.unlock()
        }
        var msgIds:Set<String>? = contentMap[storageObject]
        if msgIds == nil {
            msgIds = Set<String>()
            contentMap[storageObject] = msgIds
        }
        msgIds!.update(with: clientMsgId)
    }
    
    public func removeStorageObjectDependency(storageObject:StorageObject) {
        let completed:Bool = storageObject.getSyncStatus() == SyncStatus.inSync
        contentLock.lock()
        defer {
            contentLock.unlock()
        }
        if let ids:Set<String> = contentMap.removeValue(forKey: storageObject) {
            if !completed {
                failedContentIds.formUnion(ids)
            }
        }
    }
    
    func isContentDependent(clientId:String) -> Bool {
        contentLock.lock();
        defer {
            contentLock.unlock()
        }
        for set:Set<String> in contentMap.values {
            if set.contains(clientId) {
                return true
            }
        }
        return false
    }

    override func queue(message: Message) throws {
        self.messageQueue.async(execute: {
            // serial dispatch queue to read/write "outgoingMessageQueue"
            var queued: Bool = false
            self.messageQueued.lock()
            // TODO: No indication that the queue has exceeeded max queue size
            queued = (self.outgoingMessageQueue.count < self.maximumQueueSize) &&
                     (self.outgoingMessageQueue.push(new: message))
            
            if (!queued && self.errorCallback != nil) {
                do {
                    try self.errorCallback!.failed(messages: Array<Message>(arrayLiteral: message))
                } catch {
                    print("Issue when calling error callback")
                }
            } else {
                self.messageQueued.signal()
            }
            defer {
                self.messageQueued.unlock()
            }
        }) 
    }
    
    // TODO: Shouldn't there be synchronization around "closed" and 
    // "requestClose"
    /** @see oracle.iot.client.Client#close() */
    override func close(callback: @escaping () -> Void) {
        if (!closed) {
            self.messageQueued.lock()
            requestClose = true
            self.receiveSem.signal()
            self.messageQueued.signal()
            group.notify(queue: DispatchQueue.main) {
                self.closed = true
                callback()
            }
            self.messageQueued.unlock()
        } else {
            callback()
        }
    }    
    
    /* package */

    /**
     * - throws: `ClientError.argument` if `requestHandler` is `nil`
     */
    func getResource(endpointId: String, name: String, path: String,
                     requestHandler: RequestHandler,
                     methods: [Method]) throws -> Resource {
        
        let messageDispatcher: MessageDispatcher = self
        let requestDispatcher: RequestDispatcher = messageDispatcher.getRequestDispatcher()
        try requestDispatcher.registerRequestHandler(endpointId: endpointId,
                                                     path: path,
                                                     handler: requestHandler)
        
        let builder: ResourceBuilder = ResourceBuilder()
        builder.setEndpointName(endpointName: endpointId)
        for m in methods {
            builder.addMethod(method: m)
        }
        builder.setPath(path: path)
        builder.setName(name: name)
        return try builder.build()
    }
}
