/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import CommonCrypto

/**
 * Enumeration for choosing type of the resource directory message. Allowed
 * statuses are `UPDATE`, `DELETE`, `RECONCILIATION`.
 */
public enum ResourceMessageType: String {
    case UPDATE
    case DELETE
    case RECONCILIATION
}

/**
 * 1ResourceMessageBuilder1 extends `MessageBuilder` class. `ResourceMessage`
 * class is immutable.
 * A builder is required when creating a new `ResourceMessage.
 */
public class ResourceMessageBuilder: MessageBuilder {
    
    /**
     * Array of resources. Each Resource object is represented by a path
     * which is the path to the resource.
     */
    fileprivate var resources: [Resource] = []
    fileprivate var type: ResourceMessageType = .UPDATE
    fileprivate var endpointName: String? = nil
    fileprivate var reconciliationMark: String? = ""
    
    public override init() {
    }
    
    /**
     * Set the type of the `ResourceMessage` to`DELETE`.
     *
     */
    public func delete() {
        self.type = .DELETE
        if self.resources.count > 0 {
            self.resources = []
        }
    }
    
    /**
     * Sets the type of the `Resource` to `UPDATE` and
     * adds the resource to the list
     *
     * - parameter ri: the resource to register.
     */
    public func register(ri: Resource) {
        let ri1 = ri
        self.type = .UPDATE
        ri1.setStatus(st: .ADDED);
        self.resources.append(ri1)
    }
    
    /**
     * Set the type of the `Resource` to `REMOVE` and
     * adds the resource in the list
     *
     * - parameter ri: the resource to remove.
     */
    public func remove(ri: Resource) {
        let ri1 = ri
        self.type = .DELETE
        ri1.setStatus(st: .REMOVED);
        self.resources.append(ri1)
    }
    
    /**
     * Add a resource in the list
     *
     * - parameter ri: the resource to add.
     */
    public func add(ri: Resource) {
        self.type = .UPDATE
        self.resources.append(ri)
    }
    
    /**
     * Adds all resources.
     *
     * - parameter resources: Resources to be added.
     */
    public func addResources(resources: [Resource]) {
        self.type = .UPDATE
        for r in resources {
            self.resources.append(r)
        }
    }
    
    /**
     * Set the reconciliation mark
     *
     * - parameter rM: reconciliation mark
     */
    public func setReconciliationMark(rM: String) {
        self.reconciliationMark = rM
    }
    
    /**
     * Set the endpointName this message is for.
     *
     * - parameter endpointName: the endpoint Name for this message.
     */
    public func setEndpointName(endpointName: String) {
        self.endpointName = endpointName
    }
    
    /**
     * Sets the message type to `RECONCILIATION`.
     *
     */
    public func reconcile() {
        self.type = .RECONCILIATION
    }
    
    /**
     * Creates new instance of a `ResourceMessage` using values from this
     * *ResourceMessageBuilder*.
     *
     * - returns: Instance of *ResourceMessage*
     *
     * - Throws: WrongSource, IllegalResources
     */
    override public func build() throws -> ResourceMessage {
        return try ResourceMessage(builder: self)
    }
    
    /**
     * Initialize this builder from a JSON object
     *
     * - parameter jsonObject: the jsonObject to initialize from
     
     * - returns: the extracted item as AnyObject
     *
     */
    override public func fromJSON(data: Foundation.Data?) throws -> Dictionary<String,Any> {
        
        let jsonDict: Dictionary = try super.fromJSON(data: data)
        var resourceDictArray: Array<Any>? = nil
        for (key, value) in jsonDict {
            if key != "payload" {
                continue
            }
            guard let payload:Dictionary = (value as? Dictionary<String,Any>) else {
                continue
            }
            for (key1, value1) in payload {
                if key1 != "value" {
                    continue
                }
                // TODO: Need to harden this code, should not be forcing
                // unwrap on an Any value.
                let payloadValue = value1 as! Dictionary<String,Any>
                for (key2, value2) in payloadValue {
                    if key2 == "reconciliationMark" {
                        self.reconciliationMark = (value2 as! String)
                    } else
                    if key2 == "endpointName" {
                        self.endpointName = (value2 as! String)
                    } else
                    if key2 == "resources" {
                        resourceDictArray = (value2 as! Array)
                    }
                }
            }
        }
        
        if let resourceDictArray = resourceDictArray {
            for dict in resourceDictArray {
                let rb: ResourceBuilder = ResourceBuilder()
                rb.fromData(jsonDict: dict as! Dictionary)
                try resources.append(rb.build())
            }
        }
        
        return jsonDict
    }
}

/**
 * Implementation of protocol `Equatable`
 *
 * parameter lhs: one item to be compared
 *
 * parameter rhs: the other item to be compared
 *
 * - returns: `true` if equal, `false` otherwise
 */
public func ==(lhs: ResourceMessage, rhs: ResourceMessage) -> Bool {
    if lhs === rhs {
        return true
    }
    if !((lhs as Message) == (rhs as Message)) {
        return false
    }
    if lhs.type != rhs.type {
        return false
    }
    if lhs.resources!.count != rhs.resources!.count {
        return false
    }
    for i in 0..<lhs.resources!.count {
        if lhs.resources![i] != rhs.resources![i] {
            return false
        }
    }
    return lhs.reconciliationMark == rhs.reconciliationMark
}

/**
 * `ResourceMessage` extends `Message` class. It is used for synchronizing
 * information between Resource Directories on the client and server side.
 * This class is immutable.
 */
public class ResourceMessage: Message {

    /**
     * List of resources described by the message
     */
    fileprivate var resources: [Resource]?
    
    /**
     * Type of the message
     */
    fileprivate var type: ResourceMessageType?
    
    /**
     * hash for resource available  on client at moment of message generation
     */
    fileprivate var reconciliationMark: String?
    
    /**
     * Endpoint ID, for that the resources shall be updated
     */
    fileprivate var endpointName: String?
    
    /**
     * Message constructor takes message builder and set values to each field.
     * In case a value is `nil`, it sets a default value.
     *
     * - parameter builder: `ResourceMessageBuilder` containing all information
     *                      for Resource Directory message.
     *
     * - throws: `ClientError.state` if `builder.endpointName` is `nil` or empty
     *           or resources count is 0 or of if type is `.UPDATE`
     */
    fileprivate init(builder: ResourceMessageBuilder) throws {
        try super.init(builder: builder)
        guard let endpointName = builder.endpointName,
                !endpointName.isEmpty &&
                    !(builder.resources.isEmpty && builder.type == .UPDATE) else {
            throw ClientError.state(NSLocalizedString(
                "ResourceMessage resources cannot be empty.", comment: ""))
        }
        self.endpointName = endpointName
        self.resources = builder.resources
        self.type = builder.type
        self.reconciliationMark = builder.reconciliationMark ?? ""
    }
    
    /**
     * Calculates the MD5 hash value for a list of Strings.
     * The String list is being alphabetical ordered before the calculation
     *
     * - parameter resStrings: List of Strings
     * 
     * - returns: hash value
     */
    public static func getMD5ofList(resStrings: [String]) -> String? {
        // TODO: this method is no where used shall we delete it?
        
        // NOTE: This implementation must match the implementation on the server!
        let rs: [String] = Array(resStrings)
        
        var allString: String = ""
        for s in rs {
            allString = allString + s
        }
        var digest = [UInt8](repeating: 0, count: Int(CC_MD5_DIGEST_LENGTH))
        if let data = allString.data(using: .utf8) {
            CC_MD5((data as NSData).bytes, CC_LONG(data.count), &digest)
        }
        var digestHex = ""
        for index in 0..<Int(CC_MD5_DIGEST_LENGTH) {
            digestHex += String(format: "%02x", digest[index])
        }
        return digestHex
    }


    /**
     * Exports data from *ResourceMessage* to *String* using JSON interpretation 
     * of the message.
     *
     * - returns: JSON interpretation of the message as String.
     */
    override public func toString()-> String {
        var result: String = ""
        result = toJsonString()
        return result
    }
    
    /**
     * Method to export the Resource Directory message as a JSON string.
     *
     * - returns: JSON string for this `ResourceMessage`
     */
    override public func toJsonString() -> String {
        var result: String
        var addings: String = ""
        
        addings = addings + "\"payload\":{"
        addings = addings + "\"value\":{"
        addings = addings + "\"reportType\":\"" + self.type!.rawValue + "\","
        addings = addings + "\"endpointName\":\"" + self.endpointName! + "\"," // "\""
        // TODO: in Cposix the reconciliation mark is the hash value
        if self.reconciliationMark != nil && !self.reconciliationMark!.isEmpty {
            addings = addings + "\"reconciliationMark\":\"" + self.reconciliationMark! + "\","

        }
        if let resources = resources {
            addings = addings + "\"resources\":["
            var z: Int = 0
            for r in resources {
                addings = addings + r.toJsonString()
                z += 1
                if z < resources.count {
                    addings = addings + ","
                }
            }
            addings = addings + "]"
        }
        addings = addings + "}}"
        
        result = toJsonString(expand: false, additionalItems: addings)
        
        return result
    }
    
    /**
     * Gets the message type.
     *
     * - returns: message type `RESOURCE`
     */
    override public func getType() -> MsgType {
        return .RESOURCE
    }
    
    /**
     * Returns an array of resources described in the message.
     *
     * - returns: array of resources, never `nil`
     */
    public func getResources() -> [Resource] {
        return self.resources!
    }
    
    /**
     * Returns type of the message (Could be `UPDATE` or `DELETE`)
     *
     * - returns: RDMessageType
     */
    public func getMessageType() -> ResourceMessageType {
        return self.type!
    }
    
    /**
     * Gets the message's reconciliation mark.
     *
     * - returns: the reconciliation mark
     */
    public func getReconciliationMark() -> String {
        return self.reconciliationMark!
    }
    
    /**
     * Returns endpoint id for the message
     
     * - returns: endpoint id
     */
    public func getEndpointName() -> String {
        return self.endpointName!
    }
    
    /**
     * Implementation of protocol `Hashable`
     *
     * - returns: the hash value
     */
    override public var hashValue : Int {
        var result: Int = super.hashValue
        result = 31 &* result &+ (type?.hashValue)!
        var resourcesHash: Int = 0
        for r in resources! {
            resourcesHash = resourcesHash &+ r.hashValue
        }
        result = 31 &* result &+ resourcesHash
        return result;
    }

}

/**
 * Response status enum
 */
public enum ResponseStatus: String {
    case OK
    case BAD_REPORT
    case RECONCILIATION
}

