/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import QuartzCore
// FIXME: comparison operators with optionals were removed from the Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}

// FIXME: comparison operators with optionals were removed from the Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func >= <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l >= r
  default:
    return !(lhs < rhs)
  }
}


/**
 * SendReceiveImpl is an implementation of the send(Message...) method of
 * *AdvancedDirectlyConnectedDevice*. The send is synchronous. Receive should be
 * considered as a synchronous call. The implementation has a buffer for 
 * receiving request messages. If there are no messages in the buffer, the 
 * receive implementation will send a message to the server to receive any 
 * pending requests the server may have.
 *
 * The size of the buffer can be configured by setting the property *TBD*.
 */
class SendReceiveImpl {
    
    /*
     * Messages from server to be handled. The buffer is a circular buffer of
     * bytes. The first two bytes at the head of the buffer always contains
     * the number of bytes in the message that follows. The tail points to the
     * next available chunk of buffer.  When bytes are written to the buffer,
     * the bytes are written from (tail+2) to the end of the buffer, and then
     * from buffer[0] if necessary. In this way, the buffer wraps from tail
     * to head. Head keeps moving as messages are read. Tail keeps moving as
     * requests are received. Bytes in the buffer never have to be moved.
     *
     * The number of bytes available are sent to the server in a query parameter
     * to the messages REST API:
     * *POST /iot/api/v2/messages?acceptBytes=<number-of-bytes>*
     */
    private static let DISABLE_LONG_POLLING_PROPERTY: Bool = Config.DISABLE_LONG_POLLING
    
    private static let REQUEST_BUFFER_SIZE_PROPERTY: Int = Config.REQUEST_BUFFER_SIZE
    private static let SEND_RECEIVE_TIMEOUT_PROPERTY: Int64 = Config.SEND_RECEIVE_TIMEOUT
    
    private static let MIN_ACCEPT_BYTES_HEADER: String = "X-min-acceptBytes"
    
    private var requestBuffer: [UInt8]?
    private var head: Int?
    private var tail: Int?
    private let reqBuffQueue: DispatchQueue = DispatchQueue(label: "com.oracle.client.impl.device.reqBuffQueue")
    
    var secureConnection: HttpSecureConnection?
    
    private var sendReceiveTimeLimit: Int64?
    
    private var sendCallTime: Int64?
    
    private var requestBufferSize: Int?
    
    /**
     * - parameter secureConnection: the secure connection over which bytes are sent
     *                         and received
     */
    init(secureConnection: HttpSecureConnection) {
    
        self.secureConnection = secureConnection;
    
        requestBufferSize =
            SendReceiveImpl.REQUEST_BUFFER_SIZE_PROPERTY
    
        self.sendReceiveTimeLimit =
            SendReceiveImpl.SEND_RECEIVE_TIMEOUT_PROPERTY
    
    // TODO: configurable.
        requestBuffer = Array<UInt8>(repeating: 0, count: requestBufferSize!)
        head = 0
        tail = 0
        self.sendCallTime = -1;
    }
    
    deinit {
        self.secureConnection = nil
    }

    
    /*
     * Called by *AdvancedDirectlyConnectedDevice#send(Message..., callback:)*
     *
     * - parameter messages: zero or more *DataMessage*, *AlertMessage*,
     *                       *ResponseMessage* and/or *ResourceMessage* objects
     *                       authenticated with the server.
     * - parameter callback: The function to be executed after completion of
     *                      sending or queing a message. if send is not
     *                      successful, then an error object is passed
     *                      in the form `callback(error)` and the
     */
    func send(messages: Message...,
              callback: @escaping (ClientError?) -> ()) throws {
        try send(messages: messages, callback: callback)
    }
    /*
     * Callbed by *AdvancedDirectlyConnectedDevice#send(Message...., callback:)*
     *
     * - parameter messages: zero or more *DataMessage*, *AlertMessage*,
     *                       *ResponseMessage* and/or *ResourceMessage* objects
     *                       authenticated with the server.
     * - parameter callback: The function to be executed after completion of
     *                      sending or queing a message. if send is not
     *                      successful, then an error object is passed
     *                      in the form `callback(error)` and the
     *                      reason can be taken from error.message
     */
    func send(messages: [Message]?,
              callback: @escaping (ClientError?) -> ()) throws {
        
        var payload: Foundation.Data? = nil
        do {
            if (messages != nil) && (messages!.count > 0) {
                self.sendCallTime = TimeManager.currentTimeMillis()
                
                let jsonStr: String = try Message.toJsonString(messages: messages!, expand: true)
                
                payload = jsonStr.data(using: .utf8)
                try post(payload: payload, timeout: 0, callback: callback)
            } else {
                callback(nil)
            }
        } catch {
            callback(ClientError.network(error.localizedDescription))
            throw error
        }
    }
    
    /*
     * Called by *AdvancedDirectlyConnectedDevice#receive(long, callback)* 
     * - parameter timeout: maximum time, in milliseconds, to wait for a
     *                      response from the server
     * - parameter callback: The function to be executed after receiving a
     *                      a message. If receive is not successful then
     *                      the `ReqeuestMessage` will be `nil` and an error
     *                      object is passed in the form `callback(nil, error)`
     *                      and the reason can be taken from error.message
     */
    func receive(timeout: Int64,
                 callback: @escaping (RequestMessage?, ClientError?) -> ()) throws  {
        
        var req : RequestMessage? = try self.getRequest()
        if (req != nil) {
            callback(req, nil)
            return
        }

        let receiveCallTime: Int64 = TimeManager.currentTimeMillis()
        if ((SendReceiveImpl.DISABLE_LONG_POLLING_PROPERTY) &&
            ((receiveCallTime - self.sendCallTime!) < self.sendReceiveTimeLimit!)) {
            // time delta between last send and this receive is too small
            // do not make a call to the network to get request messages
            callback(nil, nil)
        } else {
            do {
                let callbackPost: ((ClientError?) -> ()) = {error in
                    if (error != nil) {
                        callback(nil, error)
                    } else {
                        do {
                            req = try self.getRequest()
                            // The (nil, nil) case is handled
                            callback(req, nil)
                        } catch {
                            callback(nil, ClientError.network(error.localizedDescription))
                        }
                    }
                }
                try post(payload: nil, timeout: timeout, callback: callbackPost)

            } catch {
                callback(nil, ClientError.network(error.localizedDescription))
                throw error
            }
            self.sendCallTime = TimeManager.currentTimeMillis()
        }
    }
    
    /*************************************************************************
     *
     * Methods and classes for handling outgoing message dispatch
     * Implementation and API in this section should be private
     *
     *************************************************************************/
    
    func post(payload: Foundation.Data?, timeout:Int64, callback: @escaping (ClientError?) -> ()) throws -> Void {
        var headers: [String : String] = [ : ]
        headers["Content-Type"] = "application/json"
        headers["Accept"] = "application/json"
        headers["Accept-Encoding"] = ""
        headers["X-EndpointId"] = self.secureConnection?.getTrustedAssetsManager().getClientId()
    
        // head and tail are kept so that they are less than requestBuffer.length.
        // This avoids the (unlikely) situation that they would overflow.
        // But this also means that tail might be less than head, since the buffer
        // is circular. This has to be taken into account when calculating the
        // available bytes. Remember that tail is the pointer to the next
        // available chunk and does not count toward used bytes (no need to
        // adjust used byte count by 1). If tail == head, then used bytes is
        // zero - perfect!
        let usedBytes: Int = getUsedBytes()
    
        // subtract 2 for the length bytes.
        // TODO: subtracting 2 may fail if there is more than one message
        // in the received buffer. Say we had 100 bytes available and we two
        // messages of 49 bytes. We'd write 2 bytes for the 49, then the
        // 49 bytes. Now we only have 49 bytes of buffer left, which is not
        // enough to hold 2 bytes of length plus 49 bytes of message.
        let availableBytes: Int = getRequestBufferSize()
        var status: Int = 0
        var map: [String : [String]]?
        var localError : ClientError? = nil
        var url = RestApi.V2.getReqRoot() + "/messages?acceptBytes=" + availableBytes.description
        
        if (!SendReceiveImpl.DISABLE_LONG_POLLING_PROPERTY && (timeout != 0)) {
            url = url + "&iot.sync"
            
            if (timeout > 0) {
                // iot.timeout is in seconds
                var iotTimeout = timeout / 1000
                if (iotTimeout == 0) {
                    iotTimeout = 1
                }
                url = url + "&iot.timeout=" + String(iotTimeout)
            }
        }
        
        var ipayload:Foundation.Data? = payload
        if(payload == nil) {
            ipayload = String("[]").data(using: .utf8)!
        }
        
        try secureConnection!.post(
            restApi: url,
            payload: ipayload, headers: headers, response: { httpResponse, error in
                if (error != nil) {
                    localError = error
                } else {
                    if let httpObject = httpResponse {
                        status = httpObject.getStatus()
                        if status == StatusCode.ACCEPTED.getCode() {
                            map = httpObject.getHeaders()
                            var header: [String]? = map![SendReceiveImpl.MIN_ACCEPT_BYTES_HEADER]

                            self.reqBuffQueue.sync(execute :{
                                if (header != nil) && (header!.count > 0) {
                                    let minAcceptBytes: Int = Int(header![0])!
                                    
                                    // The next request the server has for this client is larger
                                    // than the bytes available in the request buffer.
                                    //
                                    // A properly configured and responsive system should never
                                    // get in this state.
                                    if (minAcceptBytes > (self.requestBufferSize! - 2)) {
                                        // The request the server has is larger than buffer.
                                        localError = ClientError.network(NSLocalizedString("The server has a request of " + minAcceptBytes.description +
                                            " bytes for this client, which is too large for the " +
                                            self.requestBufferSize!.description + " byte request buffer. Please restart the client with larger value for " +
                                            SendReceiveImpl.REQUEST_BUFFER_SIZE_PROPERTY.description, comment: ""))
                                    } else {
                                        // The message(s) from the last time have not been processed.
                                        localError = ClientError.network(NSLocalizedString("The server has a request of " + minAcceptBytes.description +
                                            " bytes for this client, which cannot be sent because the " +
                                            self.requestBufferSize!.description + " byte request buffer is filled with " + usedBytes.description +
                                            " of unprocessed requests", comment: ""))
                                    }
                                } else {
                                    self.bufferRequest(data: httpObject.getData())
                                    // adjust tail to keep it < requestBuffer.length to avoid overflow.
                                    self.tail = self.tail! % self.requestBuffer!.count
                                }
                            })
                        }
                    } else {
                        localError = ClientError.network(NSLocalizedString("SendReceiveImpl post() httpResponse should not be nil", comment: ""))
                    }
                }
                
                callback(localError)
            }
        )
    }
    
    func getUsedBytes() -> Int {
        return self.reqBuffQueue.sync { tail >= head ? tail! - head! : (tail! + requestBuffer!.count) - head! }
    }
    
    func getRequestBufferSize() -> Int {
        return self.reqBuffQueue.sync { requestBuffer!.count }
    }
    
    private func clearText(bytes: [UInt8], p: Int) -> String {
        return String(bytes: Array(bytes[p..<(p+1)]), encoding: .utf8)!
    }
    
    private func getRequest() throws -> RequestMessage? {
        var reqMsg : RequestMessage? = nil
        
        try self.reqBuffQueue.sync(execute :{
            if (head != tail) {
                var nBytes: Int = 0
                var b1: UInt8 = requestBuffer![head! % requestBuffer!.count]
                head = head! + 1
                var i1: Int = (Int(b1) & 0xFF) << 8
                nBytes = nBytes + i1
                b1 = requestBuffer![head! % requestBuffer!.count]
                head = head! + 1
                i1 = Int(b1) & 0xFF
                nBytes = nBytes + i1
                
                let offset: Int = head!
                
                // keep head < requestBuffer.length to avoid overflow of head.
                head = (head! + nBytes) % requestBuffer!.count
                
                var dataBytes: [UInt8] = [UInt8](repeating: 0, count: nBytes)
                for i in 0..<nBytes {
                    dataBytes[i] = requestBuffer![(offset + i) % requestBuffer!.count]
                }
                let data: Foundation.Data = Foundation.Data(bytes: UnsafePointer<UInt8>(dataBytes), count: nBytes)
                let builder: RequestMessageBuilder = RequestMessageBuilder()
                let _ = try builder.fromJSON(data: data)
                
                reqMsg = try builder.build()
            }
        })
            
        return reqMsg
    }
    
    func bufferRequest(data: Foundation.Data?) {
        if data == nil {
            return
        }
        var dataBytes: [UInt8] = [UInt8](repeating: 0, count: data!.count)
        (data! as NSData).getBytes(&dataBytes, length: data!.count)
        // if data.length == 2, then it is an empty json array and there are
        // no values in the message.
        if dataBytes.count > 2 {
            var pos: Int = 1
            while (clearText(bytes: dataBytes, p: pos) == "{") {
                let start: Int = pos
                var lengthIndex0: Int = tail!
                // first two bytes is for length, start buffering json bytes at tail + 2
                tail = tail! + 2
                // json objects start with '{', end with '}'
                // count braces to pair '{' with matching '}'
                var braceCount: Int = 0
                while pos < dataBytes.count {
                    let b: UInt8 = dataBytes[pos]
                    pos = pos + 1
                    requestBuffer![tail! % requestBuffer!.count] = b
                    tail = tail! + 1
                    if (clearText(bytes: dataBytes, p: pos - 1) == "{") {
                        braceCount += 1
                    } else {
                        if (clearText(bytes: dataBytes, p: pos - 1) == "}") {
                            braceCount -= 1
                            if braceCount == 0 {
                                break
                            }
                        }
                    }
                }
                
                let nbytes: Int = pos - start
                requestBuffer![lengthIndex0 % requestBuffer!.count] =
                    UInt8(((0xff00 & nbytes) >> 8))
                lengthIndex0 += 1
                requestBuffer![lengthIndex0 % requestBuffer!.count] =
                    UInt8(((0x00ff & nbytes)))
                
                // get past the end of the json object
                // json objects end with '}' and json objects are separated by ','
                if(pos < dataBytes.count) && (clearText(bytes: dataBytes, p: pos) == "}") {
                    pos += 1
                }
                if(pos < dataBytes.count) && (clearText(bytes: dataBytes, p: pos) == ",") {
                    pos += 1
                }
            }
        }
        // adjust tail to keep it < requestBuffer.count to avoid overflow
        tail = tail! % requestBuffer!.count
    }
}
