/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
// FIXME: comparison operators with optionals were removed from the Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}

// FIXME: comparison operators with optionals were removed from the Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func > <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l > r
  default:
    return rhs < lhs
  }
}


/**
 * Class which handles HTTP requests for test connectivity.
 */
class TestConnectivityHandler: RequestHandler {
    
    private var testConnectivity: TestConnectivity
    var asyncThreadQueue: OperationQueue?
    var testConnectivityThread: TestConnectivityThread?
    private let syncQueue:DispatchQueue = DispatchQueue.global(qos:.utility)
    
    init(tc: TestConnectivity) {
        self.testConnectivity = tc
    }
    
    override func handleRequest(requestMessage: RequestMessage?) throws -> ResponseMessage {
        if ((requestMessage != nil) && (requestMessage!.getMethod() != nil)) {
            let method: String = (requestMessage!.getMethod()?.uppercased())!
            
            if (method == "GET") {
                return try getTestConnectivity(request: requestMessage!)
            } else if (method == "PUT") {
                return try putTestConnectivity(request: requestMessage!)
            }
            
            return try testConnectivity.getMethodNotAllowedResponse(requestMessage: requestMessage!)
        }
        
        return try testConnectivity.getNilRequestResponse(requestMessage: requestMessage!)
    }
    
    private func getTestConnectivity(request: RequestMessage) throws -> ResponseMessage {
        var jsonDict: [String : AnyObject] = [ : ]
        var jsonStr: String = "{"
        syncQueue.sync(execute: {
            if (self.testConnectivityThread != nil) || (self.asyncThreadQueue?.operationCount > 0) {
                jsonDict["active"] = true as AnyObject?
                jsonStr = jsonStr + "\"active\":true"
            } else {
                jsonDict["active"] = false as AnyObject?
                jsonStr = jsonStr + "\"active\":false"
            }
        })
        
        jsonDict["count"] = testConnectivity.count as AnyObject?
        jsonStr = jsonStr + ",\"count\":" + (testConnectivity.count?.description)!
        jsonDict["currentCount"] = testConnectivity.currentCount as AnyObject?
        jsonStr = jsonStr + ",\"currentCount\":" + (testConnectivity.currentCount?.description)!
        jsonDict["interval"] = testConnectivity.interval as AnyObject?
        jsonStr = jsonStr + ",\"interval\":" + (testConnectivity.interval?.description)!
        jsonDict["size"] = testConnectivity.size as AnyObject?
        jsonStr = jsonStr + ",\"size\":" + (testConnectivity.size?.description)!
        jsonStr = jsonStr + "}"
        
        return try testConnectivity.getResponseMessage(requestMessage: request, body: jsonStr, statusCode: .OK)
    }
    
    private func putTestConnectivity(request: RequestMessage) throws -> ResponseMessage {

        var active: Bool? = nil
        var errors: String = ""
        var json: [String : AnyObject]?
        let data: Foundation.Data = request.getBody() as Foundation.Data
        do {
            json = try JSONSerialization.jsonObject(with: data, options: JSONSerialization.ReadingOptions()) as? [String : AnyObject]
        } catch {
            return try testConnectivity.getBadRequestResponse(requestMessage: request,
                                                              message: "Error parsing JSON request, check parameters.")
            
        }
        if (json == nil) {
            return try testConnectivity.getBadRequestResponse(requestMessage: request,
                                             message: "Error parsing JSON request, check parameters.")
        }
            
        active = json!["active"] as? Bool
            
        if (active == nil) {
            // Invalid request, active is a required field.
            return try testConnectivity.getBadRequestResponse(requestMessage: request,
                                                          message: "Required parameter 'active' is missing.")
        }
            
        if (active == false) {
            // Stop the test connectivity run.
            syncQueue.sync(execute: {
                if (self.testConnectivityThread != nil) || (self.asyncThreadQueue!.operationCount > 0) {
                    self.asyncThreadQueue!.cancelAllOperations()
                    self.testConnectivityThread = nil
                }
            })
            
            // Always return success
            return try testConnectivity.getResponseMessage(requestMessage: request, body: "", statusCode: .OK)
        }
            
        // active is TRUE, attempt to start a test
        // connectivity run.
        var alreadyRunning:Bool = false
        var responseMessage:ResponseMessage!
        try syncQueue.sync(execute: {
            if (self.testConnectivityThread != nil) || (self.asyncThreadQueue!.operationCount > 0) {
                // A test connectivity run is already running.
                alreadyRunning = true
                responseMessage = try testConnectivity.getAlreadyRunningResponse(requestMessage: request);
            }
        })
        if alreadyRunning {
            return responseMessage
        }
        
        // Start the test connectivity run.
        testConnectivity.count = testConnectivity.getParam(jsonDict: json!, paramName: "count", errors: &errors)
        testConnectivity.interval = testConnectivity.getParam(jsonDict: json!, paramName: "interval", errors: &errors);
        testConnectivity.size = testConnectivity.getParam(jsonDict: json!, paramName: "size", errors: &errors);
            
        if (testConnectivity.count < 1) {
            errors = errors + "count must be greater then 0."
        }
            
        if (testConnectivity.interval < 0) {
            errors = errors + "interval must be greater then or equal " + "to 0."
        }
            
        if (testConnectivity.size > 4096) {
            errors = errors + "size is greater than 4096."
        }
            
        if !(errors == "") {
            testConnectivity.resetParameters()
            return try testConnectivity.getBadRequestResponse(requestMessage: request, message: errors)
        }
            
        self.testConnectivityThread = try TestConnectivityThread(tc: testConnectivity)
        
        asyncThreadQueue!.addOperation({
            do {
                try self.testConnectivityThread!.run()
            } catch {
            }
        })
        
        return try testConnectivity.getResponseMessage(requestMessage: request, body: "", statusCode: .OK);
    }
}

/**
 * A Thread which sends a DataMessage to the server using the parameters
 * of the test connectivity request.
 */
class TestConnectivityThread {
    private var payload: Foundation.Data?
    private var testConnectivityMessageBuilder: DataMessageBuilder?
    private var tc: TestConnectivity
    
    /**
     * Constructs a TestConnectivity "Thread"
     */
    init(tc: TestConnectivity) throws {
        self.tc = tc
        let sizeInt: Int = Int(tc.size!) > Int.max ? Int.max : Int(tc.size!)
        self.payload = Foundation.Data(bytes: UnsafePointer<UInt8>([UInt8](repeating: 0, count: sizeInt)), count: sizeInt)
        
        self.testConnectivityMessageBuilder = DataMessageBuilder()
        try self.testConnectivityMessageBuilder!.addDataItem(key: "count", value: Double(tc.count!))
        try self.testConnectivityMessageBuilder!.addDataItem(key: "payload",
                value: String(data: payload!, encoding: .utf8)!)
        self.testConnectivityMessageBuilder!.setFormat(format: "urn:oracle:iot:dcd:capability:diagnostics:test_message")
        self.testConnectivityMessageBuilder!.setSource(source: tc.endpointId!)
    }
    
    func run() throws {
        for c in 0..<Int(tc.count!) {
            do {
                try self.testConnectivityMessageBuilder!.addDataItem(key: "current", value: Double(c))
                try tc.messageDispatcher!.queue(message: self.testConnectivityMessageBuilder!.build())
                tc.currentCount = Int64(c + 1)
                sleep(UInt32(tc.interval!))
            } catch {
                tc.resetParameters()
                break
            }
        }
    }
}

/**
 * This class is connected to classes for handling diagnostics test connectivity messages.
 * It is assumed there will only be one instance of this class constructed.
 */
class TestConnectivity {
    // ----- these value should be initialized 
    // The number of Messages to send.
    fileprivate  var count: Int64?
    // The number of Messages already sent.
    fileprivate  var currentCount: Int64?
    // The interval between sending each Message.
    fileprivate  var interval: Int64?
    // The size of the payload of the Message.
    fileprivate var size: Int64?
    
    /** The ID of the current or last test connectivity invocation so we can get the status and cancel. */
    fileprivate var messageDispatcher: MessageDispatcher?
    fileprivate var endpointId: String?
    /** The status of the current or last test connectivity invocation.  We only keep one */
    fileprivate var testConnectivityThread: TestConnectivityThread?
    private var testConnectivityHandler: TestConnectivityHandler?
    fileprivate var testConnectivityThreadLock: AnyObject? = Int(1) as AnyObject?
    
    /**
     * Constructs a TestConnectivity
     *
     * - parameter endpointId: the endpoint ID of the Gateway.
     * - parameter messageDispatcher:  a reference to a MessageDispatcher.
     */
    init(endpointId: String, messageDispatcher: MessageDispatcher) {
        self.endpointId = endpointId
        self.messageDispatcher = messageDispatcher
        //  -- test to see if this makes the response fo back
        // where should this data be incremented ??? to deliver meaningful values
        self.count = 1
        self.currentCount = 1
        self.interval = 1
        self.size = 4096
    }
    
    /**
     * Appends spaces (error separators) to the errors list if it's not empty.
     *
     * - parameter errors: a list of errors in a String
     *
     * - returns: String with appended spaces
     */
    private func appendSpacesToErrors(errors: inout String) {
        if errors.count != 0 {
            errors = errors + "  "
        }
    }
    
    /**
     * Returns an appropriate response if the connectivity test is currently running.
     *
     * - parameter requestMessage: the request for this capability.
     
     * - returns: an appropriate response if the connectivity test is currently running.
     */
    fileprivate func getAlreadyRunningResponse(requestMessage: RequestMessage) throws -> ResponseMessage {
        return try getResponseMessage(requestMessage: requestMessage, body: "Test connectivity is already running.", statusCode: .CONFLICT)
    }
    
    /**
     * Returns an appropriate response if the request is bad.
     *
     * - parameter requestMessage: the request for this capability.
     
     * - returns: an appropriate response if the request is bad.
     */
    fileprivate func getBadRequestResponse(requestMessage: RequestMessage, message: String) throws -> ResponseMessage {
        return try getResponseMessage(requestMessage: requestMessage, body: message, statusCode: .BAD_REQUEST)
    }
    
    /**
     * Returns a *RequestMessage* with the supplied parameters.
     *
     * - parameter requestMessage: the request for this capability.
     
     * - parameter body:           the body for the response.
     
     * - parameter  statusCode:    the status code of the response.
     
     * - returns: an appropriate response if the request is *nil*
     */
    fileprivate func getResponseMessage(requestMessage: RequestMessage, body: String, statusCode: StatusCode) throws -> ResponseMessage {
        let rmb: ResponseMessageBuilder = ResponseMessageBuilder(requestMessage: requestMessage)
        rmb.setBody(body: body)
        rmb.setStatusCode(statusCode: statusCode)
        return try rmb.build()
    }
    
    /**
     * Returns an appropriate response if a method call is not supported by this handler.
     *
     * - parameter requestMessage: the request for this capability.
     
     * - returns: an appropriate response if a method call is not supported by this handler.
     */
    fileprivate func getMethodNotAllowedResponse(requestMessage:RequestMessage) throws -> ResponseMessage {
        return try getResponseMessage(requestMessage: requestMessage, body: "Unsupported request: " + requestMessage.toString(), statusCode: .METHOD_NOT_ALLOWED);
    }
    
    /**
     * Returns an appropriate response if the request is *nil*
     *
     * - parameter requestMessage: the request for this capability.
     
     * - returns: an appropriate response if the request is *nil*
     */
    fileprivate func getNilRequestResponse(requestMessage: RequestMessage) throws -> ResponseMessage {
        let rmb: ResponseMessageBuilder = ResponseMessageBuilder(requestMessage: requestMessage)
        rmb.setBody(body: "Unsupported request.")
        rmb.setStatusCode(statusCode: .METHOD_NOT_ALLOWED)
        return try rmb.build()
    }
    
    /**
     * Returns the Long parameter specified by 'paramName' in the request if it's available.
     *
     * - parameter jsonObject: a JSON dictionary containing the JSON request.
     
     * - parameter paramName:   the name of the parameter to get.
     
     * - parameter errors:  a String of errors.  Any errors produced from retrieving the parameter will be appended to this.
     
     * - returns: the parameter if it can be retrieved, otherwise *nil*
     */
    fileprivate func getParam(jsonDict: [String : AnyObject], paramName: String, errors: inout String) -> Int64 {
        var value: Int64? = nil
    
        let jsonValue: AnyObject? = jsonDict[paramName]
    
        if (jsonValue != nil) {
            value = jsonValue as? Int64
            if (value < 1) {
                errors = errors + paramName + " must be a numeric value greater then 0."
            }
    
        } else {
            appendSpacesToErrors(errors: &errors)
            errors = errors + "The " + paramName + " value must be supplied."
        }
        return value!
    }
    
    func getTestConnectivityHandler() -> RequestHandler {
        if (testConnectivityHandler == nil) {
            testConnectivityHandler = TestConnectivityHandler(tc: self);
        }
    
        return testConnectivityHandler!
    }
    
    /**
     * Reset the test connectivity run parameters.
     */
    fileprivate func resetParameters() {
        count = 0
        currentCount = 0
        interval = 0
        size = 0
    }

}
