/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * VirtualDevice for a device-client adds methods to handle write-only and
 * executable actions.
 *
 */
public protocol VirtualDevice: AbstractVirtualDevice {
    
    /**
     * Set callback for handling an action
     *
     * - parameter actionName: The action name
     *
     * - parameter callback: The function to invoke
     *
     * - throws: `ClientError.argument` if `actionName` is not an action on 
     *           the model.
     */
    func setOnCall(actionName: String,
                callback: @escaping (VirtualDevice, AnyObject?) -> Void) throws
    
    /**
     * Create a custom data object for this VirtualDevice. The custom data
     * object will be created with the given format URN.
     *
     * - parameter format: the custom data message format URN.
     *
     * - returns: a `Data` object
     *
     * - throws: `ClientError.argument` if a device model format cannot be 
     *           obtained from `format` or if the format type is not
     *           `DeviceModelFormatType.DATA`
     */
    func createData(format: String) throws -> Data
    
    /**
     * Create an `Alert` for this VirtualDevice. The alert will be created with
     * the given format URN.
     *
     * - parameter format: the alert format URN.
     *
     * - returns: an `Alert` object
     *
     * - throws: `ClientError.argument` if a device model format cannot be 
     *            obtained from `alertUrn` or if the format is not
     *            `DeviceModelFormatType.ALERT`
     */
    func createAlert(format: String) throws -> Alert
}

class VirtualDeviceImpl: VirtualDevice {
 

    var base: VirtualDeviceBase?
    
    // used from AlertImpl
    var msgDirectlyConnectedDevice: AdvancedDirectlyConnectedDevice?
    
    var attributeMap: [String : VirtualDeviceAttribute] = [ : ]
    
    private var ERROR_CALLBACK_BRIDGE: ErrorCallbackBridge = ErrorCallbackBridge()
    
    var messsageDispatcher: MessageDispatcher?
    var requestDispatcher: RequestDispatcher?
    
    var adapter: VDIAdapter?
    var requestHandler: VDIRequestHandler?
    
    var endpointId: String?
    
    /**
     * - throws: `ClientError.network` if a `MessageDispatcher` cannot be 
     *           obtained.
     */
    init(msgDirectlyConnectedDevice: AdvancedDirectlyConnectedDevice,
                endpointId: String, deviceModel: DeviceModel) throws {
        self.endpointId = endpointId
        self.msgDirectlyConnectedDevice = msgDirectlyConnectedDevice
        self.adapter = VDIAdapter(impl: self)
        self.base = VirtualDeviceBase(adapter: self.adapter!,
                                      endpointId: endpointId,
                                      deviceModel: deviceModel)
       
        self.requestHandler = VDIRequestHandler(impl: self)
        self.attributeMap = VirtualDeviceImpl.createAttributeMap(
                                virtualDevice: self, deviceModel: deviceModel)
        
        messsageDispatcher = try MessageDispatcher.getMessageDispatcher(
                        msgDirectlyConnectedDevice: msgDirectlyConnectedDevice)
        
        requestDispatcher = self.messsageDispatcher!.getRequestDispatcher()
        
        // handler for all attribute and action requests
        try requestDispatcher!.registerRequestHandler(endpointId: endpointId,
            path: "deviceModels/"+deviceModel.getURN(), handler: requestHandler)
        
        try ERROR_CALLBACK_BRIDGE.add(virtualDevice: self)
        self.messsageDispatcher!.setOnError(callback: ERROR_CALLBACK_BRIDGE)
    }
    
    private static func createAttributeMap(virtualDevice: VirtualDeviceImpl,
            deviceModel: DeviceModel) -> [String : VirtualDeviceAttribute] {
        
        var map: [String : VirtualDeviceAttribute] = [ : ]
        
        for attribute in (deviceModel as! DeviceModelImpl).getDeviceModelAttributes().values {
            let vda: VirtualDeviceAttribute = VirtualDeviceAttribute(
                                virtualDevice: virtualDevice, model: attribute)
            map[attribute.getName()!] = vda
            if let alias: String = attribute.getAlias(), !alias.isEmpty {
                map[alias] = vda
            }
        }
        return map;
    }
    
    /**
     * - throws: `ClientError.argument` if a device model format cannot be 
     *           obtained from `format` or if the format type is not
     *           `DeviceModelFormatType.DATA`
     */
    func createData(format: String) throws -> Data {
        return try DataImpl(virtualDevice: self, formatUrn: format)
    }
    
    func getDeviceModel() -> DeviceModel {
        return base!.getDeviceModel()
    }
    
    func getEndpointId() -> String {
        return base!.getEndpointId()
    }
    
    /**
     * - throws: `ClientError.argument` if `attributeName` is not a device 
     *           attribute.
     */
    func get(attributeName: String) throws -> AnyObject? {
        let attribute: VirtualDeviceAttribute =
                                try adapter!.getAttribute(name: attributeName)
        let val:AnyObject? = attribute.get()
        return val
    }
    /**
     * - throws: `ClientError.argument` if `attributeName` is not a device 
     *           attribute.
     */
    func getLastKnown(attributeName: String) throws -> AnyObject? {
        let attribute: VirtualDeviceAttribute =
                                try adapter!.getAttribute(name: attributeName)
        let val: AnyObject? = attribute.getLastKnown()
        return val
    }
    /**
     * - throws: `ClientError.argument` if attempting to set a read-only
     *           attribute, or setting an invalid value
     */
    @discardableResult
    func set(attributeName: String,
             attributeValue: AnyObject) throws -> VirtualDevice {
        try base!.set(attributeName: attributeName, value: attributeValue)
        return self
    }
    
    func update() -> VirtualDevice {
        base!.update()
        return self
    }
    /*
     * - throws: `ClientError.argument` if an attributes is nil or is the empty
     *            string or if an attribute type is string, if `newValue` is 
     *            nil or the empty string.
     */
    func finish() throws {
        try base!.finish()
    }
    
    /**
     * - throws: `ClientError.argument` if `attributeName` is not a device 
     *           attribute.
     */
    func setOnChange(attributeName: String,
                     callback: @escaping (ChangeEvent)->Void) throws {
        let attribute: VirtualDeviceAttribute =
                                try adapter!.getAttribute(name: attributeName)
        attribute.setOnChange(callback: callback)
    }
    
    func setOnChange(callback: @escaping (ChangeEvent)->Void) {
        base!.setOnChange(callback: callback)
    }
    
    func setOnError(callback: @escaping (ErrorEvent)->Void) {
        base!.setOnError(callback: callback)
    }
    /**
     * - throws: `ClientError.argument` if `attributeName` is not a device 
     *           attribute.
     */
    func setOnError(attributeName: String,
                    callback: @escaping (ErrorEvent)->Void) throws {
        let attribute: VirtualDeviceAttribute =
                                try adapter!.getAttribute(name: attributeName)
        attribute.setOnError(callback: callback)
    }
    
    var actionMap: [String : ((VirtualDevice, AnyObject?) -> Void)?]?
    
    static func getDeviceModelAction(deviceModel: DeviceModel?,
                                     actionName: String) -> DeviceModelAction? {
        if (deviceModel != nil) {
            var actions: [String : DeviceModelAction] =
                    (deviceModel as! DeviceModelImpl).getDeviceModelActions()
            if actions.keys.count > 0 {
                let act: DeviceModelAction? = actions[actionName]!
                if act != nil {
                    return act!
                }
                for action in actions.values {
                    if actionName == action.getAlias() {
                        return action
                    }
                }
            }
        }
        return nil
    }
    
    /**
     * Set callback for handling an action
     *
     * - parameter actionName: The action name
     *
     * - parameter callback: The function to invoke
     *
     * - throws: `ClientError.argument` if `actionName` is not an action on the
     *            model.
     */
    func setOnCall(actionName: String,
            callback: @escaping (VirtualDevice, AnyObject?) -> Void) throws {
        
        // TODO: NOT THREAD SAFE
        if (actionMap == nil) {
            actionMap = [ : ]
        }
        
        var actions: [String : DeviceModelAction] =
            (base!.getDeviceModel() as! DeviceModelImpl).getDeviceModelActions()
        var deviceModelAction: DeviceModelAction? = actions[actionName]
        if (deviceModelAction == nil) {
            for action in actions.values {
                if actionName == action.getAlias() {
                    deviceModelAction = action
                    break
                }
            }
        }
        if deviceModelAction == nil {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Action %@ not found in model", comment: ""), actionName))
        }
        
        actionMap![deviceModelAction!.getName()] = callback
        if deviceModelAction!.getAlias() != nil {
            actionMap![deviceModelAction!.getAlias()!] = callback
        }
    }
    
    /**
     * - throws: `ClientError.argument` if a device model format cannot be
     *           obtained from `format` or if the format is not
     *           `DeviceModelFormatType.ALERT`
     */
    func createAlert(format: String) throws -> Alert {
        return try AlertImpl(virtualDevice: self, alertUrn: format)
    }
    
    // called from VirtualDeviceAttribute
    /**
     * Call the `onChange` callback for `virtualDeviceAttribute`
     *
     * - parameter virtualDeviceAttribute: the changed attribute
     *
     * - parameter newValue: the new attribute value
     *
     * - throws: 
     *    - `ClientError.network` if a `MessageDispatcher` cannot be
     *      obtained.
     *    - `ClientError.argument` if `DataMessageBuilder.builder.source`
     *      or `DataMessageBuilder.builder.format` is `nil` or is the
     *      empty string.
     */
    func processOnChange(virtualDeviceAttribute: VirtualDeviceAttribute,
                         newValue: AnyObject) throws {
        
        let builder: DataMessageBuilder = DataMessageBuilder()
        var storageObjects:[StorageObjectImpl] = [StorageObjectImpl]()
        do {
            try processOnChange(builder: builder,
                                attribute: virtualDeviceAttribute,
                                newValue: newValue,
                                storageObjectList: &storageObjects)
        } catch {
            return
        }

        let dataMessage:DataMessage = try builder.build()
        let messageDispatcher:MessageDispatcherImpl =
            try MessageDispatcher.getMessageDispatcher(
                msgDirectlyConnectedDevice: msgDirectlyConnectedDevice!) as! MessageDispatcherImpl
        for so:StorageObjectImpl in storageObjects {
            messageDispatcher.addStorageObjectDependency(storageObject: so,
                                            clientMsgId: dataMessage.getClientId())
        }
        try messageDispatcher.queue(message: dataMessage)
    }
    
    /**
     * - throws: 
     *     - `ClientError.network` if a `MessageDispatcher` cannot be
     *       obtained.
     *     - `ClientError.argument` if `DataMessageBuilder.builder.source`
     *        or `DataMessageBuilder.builder.format` is `nil` or is the
     *        empty string.
     */
    func processOnChange(updatedAttributes: [VirtualDeviceAttribute : AnyObject]) throws {
        if updatedAttributes.count == 0 {
            return
        }
        
        let builder: DataMessageBuilder = DataMessageBuilder()
        var storageObjects:[StorageObjectImpl] = [StorageObjectImpl]()
        for (key, value) in updatedAttributes {
            do {
                try processOnChange(builder: builder, attribute: key,
                                    newValue: value,
                                    storageObjectList: &storageObjects)
            } catch {
                return
            }
        }
        let dataMessage:DataMessage = try builder.build()
        let messageDispatcher:MessageDispatcherImpl =
            try MessageDispatcher.getMessageDispatcher(
                msgDirectlyConnectedDevice: msgDirectlyConnectedDevice!) as! MessageDispatcherImpl
        for so:StorageObjectImpl in storageObjects {
            messageDispatcher.addStorageObjectDependency(storageObject: so,
                                        clientMsgId: dataMessage.getClientId())
        }
        try messageDispatcher.queue(message: dataMessage)
    }
    
    func toString() -> String {
        return base!.toString()
    }
    
    func getValue(attributeType: DeviceModelAttributeType,
                  attributeName: String? = nil,
                  payload: Foundation.Data,
                  callback: @escaping (AnyObject?, ClientError?)->()) {
        
        var payloadStructure: Dictionary<String, AnyObject>?
        var clientError:ClientError? = nil
        do {
            payloadStructure = try JSONSerialization.jsonObject(
                with: payload,
                options: JSONSerialization.ReadingOptions()) as? Dictionary<String, AnyObject>
        } catch {
            clientError = ClientError.argument(NSLocalizedString(
                "Illegal value", comment: ""))
            callback(nil, clientError)
            return
        }
        let value: AnyObject? = payloadStructure!["value"]
        if value == nil {
            clientError = ClientError.argument(NSLocalizedString(
                "Bad payload", comment: ""))
            callback(nil, clientError)
            return
        }
        self.getValue(attributeType: attributeType,
                      attributeValue: value!,
                      attributeName: attributeName,
                      callback: callback)
    }
    
    func getValue(attributeType: DeviceModelAttributeType,
                  attributeValue: AnyObject,
                  attributeName: String? = nil,
                  callback: @escaping (AnyObject?, ClientError?)->()) {
        
        var value: AnyObject = attributeValue
        var clientError:ClientError? = nil
        switch (attributeType) {
        case .BOOLEAN:
            if !(value is Bool) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break;
            
        case .INTEGER:
            if !(value is Int) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break;
            
        case .NUMBER:
            if !(value is NSNumber) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break;
            
        case .DATETIME:
            if !(value is Int64) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break;
            
        case .STRING:
            if !(value is String) {
                clientError = ClientError.argument(NSLocalizedString(
                    "Illegal attribute value", comment: ""))
            }
            break;
        case .URI:
            if !(value is String) {
                clientError = ClientError.argument(NSLocalizedString(
                    "URI value is not STRING", comment: ""))
                break;
            }
            let uri = value as! String
            if StorageConnectionBase.isStorageCloudURI(uri: uri) {
                msgDirectlyConnectedDevice!.createStorageObject(uri: uri,
                        callback: {(delegate, error) in
                            
                    var clientError:ClientError? = nil
                    if let error = error {
                        clientError = ClientError.network(String(format: NSLocalizedString(
                            "Cannot get value for attribute %@. %@", comment:""),
                                     attributeName!, "\(error)"))
                        value = ExternalObject(uri: uri)
                    } else
                    // TODO: Can error and delegate be nil ?
                    if let delegate = delegate {
                        let storageObjectImpl:StorageObjectImpl =
                            StorageObjectImpl(dcd: self.msgDirectlyConnectedDevice!,
                                              delegate:delegate)
                        storageObjectImpl.setSyncEventInfo(virtualDevice: self,
                                                nameForSyncEvent: attributeName!)
                        value = storageObjectImpl
                    } else {
                        clientError = ClientError.network(String(format: NSLocalizedString(
                            "Cannot get value for attribute %@", comment:""),
                                     attributeName!))
                        value = ExternalObject(uri: uri)
                    }
                    callback(value, clientError)
                })
                // Return don't break, the callback is invoked by
                // createStorageObject
                return
            } else {
                value = ExternalObject(uri: uri)
            }
            break
        }
        callback(value, clientError)
    }
    
    /**
     * - throws: `ClientError.argument` if `attributes` is nil or is the empty
     *           string or if the attribute type is string, if `newValue` is 
     *           nil or the empty string.
     */
    private func processOnChange(builder: DataMessageBuilder,
            attribute: VirtualDeviceAttribute, newValue: AnyObject,
            storageObjectList: inout [StorageObjectImpl]) throws {
        let deviceModelAttribute: DeviceModelAttribute =
                                        attribute.getDeviceModelAttribute()!
        
        let attributeName: String = deviceModelAttribute.getName()!
        builder.setFormat(format: base!.getDeviceModel().getURN() + ":attributes")
        builder.setSource(source: base!.getEndpointId())
        
        switch (deviceModelAttribute.getType()) {
        case .INTEGER:
            if newValue is String {
                try builder.addDataItem(key: attributeName,
                                        value: Double(newValue as! String)!)
            } else {
                try builder.addDataItem(key: attributeName,
                                        value: Double(truncating: newValue as! NSNumber))
            }
            break
        case .NUMBER:
            if newValue is String {
                try builder.addDataItem(key: attributeName,
                                        value: Double(newValue as! String)!)
            } else {
                try builder.addDataItem(key: attributeName,
                                        value: Double(truncating: newValue as! NSNumber))
            }
            break
        case .STRING:
            try builder.addDataItem(key: attributeName,
                                    value: newValue as! String)
            break
        case .BOOLEAN:
            try builder.addDataItem(key: attributeName,
                                    value: newValue.boolValue)
            break
        case .DATETIME:
            try builder.addDataItem(key: attributeName,
                                    value: Double(newValue.stringValue)!)
            break
        case .URI:
            
            if newValue is StorageObjectImpl {
                let so = newValue as! StorageObjectImpl
                if so.getSyncStatus() == SyncStatus.notInSync ||
                        so.getSyncStatus() == SyncStatus.syncPending {
                    storageObjectList.append(so)
                }
                so.setSyncEventInfo(virtualDevice: self, nameForSyncEvent: attributeName)
                try so.sync()
            }
            try builder.addDataItem(key: attributeName, value: (newValue as! ExternalObject).getURI())
            
            break
        }
    }

    func handleStorageObjectStateChange(storageObject: StorageObjectImpl) throws {
        let messageDispatcher:MessageDispatcherImpl =
            try MessageDispatcher.getMessageDispatcher(
                msgDirectlyConnectedDevice: msgDirectlyConnectedDevice!) as! MessageDispatcherImpl
        messageDispatcher.removeStorageObjectDependency(storageObject:storageObject)
    }
}
