/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * Implementation of protocol Equatable
 */
func ==(lhs: VirtualDeviceAttribute, rhs: VirtualDeviceAttribute) -> Bool {
    if lhs === rhs {
        return true
    }
    if !(lhs.virtualDevice === rhs.virtualDevice) {
        return false
    }
    if !(lhs.model == rhs.model) {
        return false
    }
    switch lhs.model.getType(){
    case .STRING:
        if !((lhs.value as! String) == (rhs.value as! String)) {
            return false
        }
        if !((lhs.lastKnownValue as! String) == (rhs.lastKnownValue as! String)) {
            return false
        }
        break
    case .INTEGER:
        if !((lhs.value as! Int) == (rhs.value as! Int)) {
            return false
        }
        if !((lhs.lastKnownValue as! Int) == (rhs.lastKnownValue as! Int)) {
            return false
        }
        break
    case .NUMBER:
        if !((lhs.value as! NSNumber) == (rhs.value as! NSNumber)) {
            return false
        }
        if !((lhs.lastKnownValue as! NSNumber) == (rhs.lastKnownValue as! NSNumber)) {
            return false
        }
        break
    case .BOOLEAN:
        if !((lhs.value as! Bool) == (rhs.value as! Bool)) {
            return false
        }
        if !((lhs.lastKnownValue as! Bool) == (rhs.lastKnownValue as! Bool)) {
            return false
        }
        break
    case .URI:
        if !((lhs.value as! String) == (rhs.value as! String)) {
            return false
        }
        if !((lhs.lastKnownValue as! String) == (rhs.lastKnownValue as! String)) {
            return false
        }
        break
    default: /* DATETIME */
        if !((lhs.value as! Int64) == (rhs.value as! Int64)) {
            return false
        }
        if !((lhs.lastKnownValue as! Int64) == (rhs.lastKnownValue as! Int64)) {
            return false
        }
        break
    }
    /* TODO: FIXME
     if !( lhs.changeCallback === rhs.changeCallback) {
     return false
     }
     if !( lhs.errorCallback === rhs.errorCallback) {
     return false
     }
     */
    return true
}

/**
 * VirtualDeviceAttribute is an attribute in a VirtualDevice model. The
 * `AbstractVirtualDevice.get(String)` and `AbstractVirtualDevice.set(String,
 * AnyObject)` methods
 * are wrappers around the `VirtualDeviceAttribute` instances of a
 * `VirtualDevice`.
 */
class VirtualDeviceAttribute: VirtualDeviceAttributeBase, Hashable {

    var virtualDevice: VirtualDeviceImpl
    
    /**
     * Implementation of protocol Hashable
     */
    var hashValue : Int {
        var result: Int = model.hashValue
        if value != nil {
            switch model.getType(){
            case .STRING:
                result = 31 &* result &+ (value as! String).hashValue
                break
            case .INTEGER:
                result = 31 &* result &+ (value!.int32Value.hashValue).hashValue
                break
            case .NUMBER:
                result = 31 &* result &+ (value!.doubleValue.hashValue).hashValue
                break
            case .BOOLEAN:
                result = 31 &* result &+ (value as! Bool).hashValue
                break
            case .DATETIME:
                result = 31 &* result &+ (value as! Int64).hashValue
                break
            case .URI:
                result = 31 &* result &+ (value as! ExternalObject).hashValue
                break
            }
        } else {
            result = 31 &* result
        }
        return result
    }

    init(virtualDevice: VirtualDeviceImpl, model: DeviceModelAttribute) {
        self.virtualDevice = virtualDevice
        super.init(model: model)
    }
    
    override func set(value: AnyObject) throws {
        // validate throws ClientError.argument if value is not valid
        do {
            try validate(attribute: getDeviceModelAttribute()!, value: value)
        } catch {
            throw ClientError.argument(NSLocalizedString(
                "illegal value", comment: ""))
        }
        
        setLastKnown(lastKnown: value)
        self.value = value
        try (getVirtualDevice()).processOnChange(virtualDeviceAttribute: self,
                                                 newValue: get()!);
    }

    override func isSettable() -> Bool {
        // An attribute is always settable on the device-client side
        return true
    }
    
    func getVirtualDevice() -> VirtualDeviceImpl {
        return virtualDevice
    }
    
    func setLastKnown(lastKnown: AnyObject) {
        self.lastKnownValue = lastKnown
    }
    
    /**
     * Same as set, but doesn't generate an event. Returns true if the value
     * was updated.
     *
     * - parameter value: the new value with which the attribute will be updated
     *
     * - returns: true if the value was updated
     *
     * - throws: `ClientError.argument` if `value` is not the correct type or is
     *           is an invalid value for `attribute`.
     */
    func update(value: AnyObject) throws -> Bool {
        
        // validate throws ClientError.argument if value is not valid
        do {
            try validate(attribute: getDeviceModelAttribute()!, value: value)
        } catch {
            throw ClientError.argument(NSLocalizedString("illegal value", comment: ""))
        }
        
        setLastKnown(lastKnown: value)
        self.value = value
        
        return true
    }

    func toString() -> String {
        let modelName = model.getName()
        var result: String = "name : " + modelName! + "\n"
        result = result + "description : "
        result = result + model.getDescription()! + "\n"
        result = result + "alias : " + model.getAlias()! + "\n"
        result = result + "model : " + model.getModel()! + "\n"
        result = result + "access : " + model.getAccess().rawValue + "\n"
        result = result + "type : " + model.getType().rawValue + "\n"
        
        let modelType: DeviceModelAttributeType = model.getType()
        if (modelType == .NUMBER) || (modelType == .INTEGER) {
            result = result + "lowerBound : " + 
                model.getLowerBound()!.description + "\n"
            result = result + "upperBound : " + 
                model.getUpperBound()!.description + "\n"
        }
        // can only be processed if the optional value value and 
        // lastKnowValue are set
        if(value != nil && lastKnownValue != nil) {
            switch modelType {
            case .STRING:
                result = result + "value : " + (value as! String) + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! String) + "\n"
                break
            case .INTEGER:
                result = result + "value : " + 
                    (value as! Int).description + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! Int).description + "\n"
                break
            case .NUMBER:
                result = result + "value : " + 
                    (value as! NSNumber).description + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! NSNumber).description + "\n"
                break
            case .BOOLEAN:
                result = result + "value : " + 
                    (value as! Bool).description + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! Bool).description + "\n"
                break
            case .URI:
                result = result + "value : " +
                    (value as! ExternalObject).getURI() + "\n"
                break
            default: /* DATETIME */
                result = result + "value : " + 
                    (value as! Int64).description + "\n"
                result = result + "lastKnownValue : " + 
                    (lastKnownValue as! Int64).description + "\n"
                break
            }
        }
        return result
    }
}
