/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

class VDIRequestHandler: AsyncRequestHandler {
    
    var impl: VirtualDeviceImpl?
    
    init(impl: VirtualDeviceImpl?) {
        self.impl = impl
    }
    
    convenience override init() {
        self.init(impl: nil)
    }

    override func handleRequest(requestMessage: RequestMessage?,
        callback: @escaping (ResponseMessage?,ClientError?)->()) throws  {
        
        //
        // Dispatch the request message to the appropriate handler for the method
        //
        
        let method: String = (requestMessage!.getMethod()?.uppercased())!
        
        var responseStatus: StatusCode = .BAD_REQUEST
        
        func getResponseMessage(statusCode:StatusCode,
                callback: @escaping (ResponseMessage?,ClientError?)->()) {
            let rmb: ResponseMessageBuilder =
                ResponseMessageBuilder(requestMessage: requestMessage!)
            rmb.setStatusCode(statusCode: statusCode)
            do {
                callback(try rmb.build(), nil)
            } catch {
                // The only time an exception will be thrown is if 
                // source or destination is nil. But this cannot happen
                // since you can't get a RequestMessage ifsource or destination
                // are nil or empty
                callback(nil, ClientError.network(String(format:
                    NSLocalizedString("Could not create ResponseMessage for statusCode: %@",
                        comment:""), "\(statusCode)") + " : \(error)"))
            }
        }
        
        if "POST" == method {
            let methodOverride: String? =
                requestMessage!.getHeaderValue(name: "X-HTTP-Method-Override")
            if methodOverride != nil && "PATCH" == methodOverride!.uppercased() {
                handlePatch(requestMessage: requestMessage!,
                            callback: {(statusCode) in
                    getResponseMessage(statusCode:statusCode,
                                       callback:callback)
                })
            } else {
                handlePost(requestMessage: requestMessage!,
                           callback: {(statusCode) in
                    getResponseMessage(statusCode:statusCode,
                                       callback:callback)
                })
            }
            
        } else if ("PUT" == method) {
            handlePut(requestMessage: requestMessage!,
                      callback: {(statusCode) in
                getResponseMessage(statusCode:statusCode,
                                   callback:callback)
            })
            
        } else if ("PATCH" == method) {
            handlePatch(requestMessage: requestMessage!,
                        callback: {(statusCode) in
                getResponseMessage(statusCode:statusCode,
                                   callback:callback)
            })
            
        } else {
            throw ClientError.state(NSLocalizedString(
                "Unexpected method", comment: ""))
        }
    }

    // Patch is always a multi-attribute PUT
    private func handlePatch(requestMessage: RequestMessage,
                             callback: @escaping (StatusCode)->()) {
        
        // Set to true on the first attribute that has an onChange handler
        // and never reset
        var attrOnChangeCalled: Bool = false
        
        var root: NamedValue? = nil
        var last: NamedValueImpl? = nil
        
        let patchGroup:DispatchGroup = DispatchGroup()
        let patchQueue:DispatchQueue = DispatchQueue.global(qos: .utility)
        
        let data: Foundation.Data = requestMessage.getBody()
        var json: [String : AnyObject]?
    
        do {
            json = try JSONSerialization.jsonObject(with: data,
                       options: JSONSerialization.ReadingOptions()) as? [String : AnyObject]
        } catch {
            callback(.BAD_REQUEST)
            return
        }
        
        for (attributeName, value) in json! {
            // Throws ClientError.argument if attribute is not in model
            var attribute: VirtualDeviceAttribute? = nil
            do {
                attribute = try impl!.adapter!.getAttribute(name: attributeName) as VirtualDeviceAttribute
            } catch {
                print(error)
                callback(.BAD_REQUEST)
                return
            }
            
            let dma: DeviceModelAttribute = attribute!.getDeviceModelAttribute()!
            
            // Embed function here to capture variable values
            func processAttribute(value: AnyObject?, error: ClientError?) {
                defer {
                    patchGroup.leave()
                }
                if let error = error {
                    print(error)
                    // TODO: is this the right "bad" status ?
                    callback(.BAD_REQUEST)
                    return
                }
                do {
                    try attribute!.validate(attribute: dma, value: value)
                } catch {
                    print(error)
                    callback(.BAD_REQUEST)
                    return
                }
                let nameValue: NamedValueImpl = NamedValueImpl(name: attributeName,
                                                               value: value!)
                if attribute!.getOnChange() != nil {
                    let ce: ChangeEventImpl = ChangeEventImpl(virtualDevice: impl!,
                                                              value: nameValue)
                    attribute!.getOnChange()!(ce)
                    patchQueue.sync(execute: {
                        attrOnChangeCalled = true;
                    })
                }
                patchQueue.sync(execute: {
                    if (last != nil) {
                        last!.setNext(next: nameValue)
                        last = nameValue
                    } else {
                        root = nameValue
                        last = nameValue
                    }
                })
            }
            patchGroup.enter()
            impl!.getValue(attributeType: dma.getType(),
                           attributeValue: value as AnyObject,
                           attributeName: attributeName,
                           callback: processAttribute)
        }
        patchGroup.notify(qos: .utility, queue: patchQueue, execute: {
            
            if (root == nil) {
                callback(.BAD_REQUEST)
                return
            }
            
            // patch is handled by change callback set on VirtualDevice
            let changeCallback: ((ChangeEvent) -> Void)? = self.impl!.base!.getOnChangeCallback()
            if changeCallback == nil {
                if attrOnChangeCalled {
                    callback(.ACCEPTED)
                    return
                }
                
                #if UNIT_TESTS
                #else
                    callback(.NOT_IMPLEMENTED)
                    return
                #endif
            } else {
                changeCallback!(ChangeEventImpl(virtualDevice: self.impl!, value: root!))
            }
            
            do {
                
               var dataItems: [DataItem] = []
                
                // If no exception, then assume the event was handled.
                // Now go update the values.
                var namedValue = root
                while namedValue != nil {
                    let attributeName: String = namedValue!.getName()
                    let attribute: VirtualDeviceAttribute =
                        try self.impl!.adapter!.getAttribute(name: attributeName)
                    let newValue = namedValue!.getValue()
                    
                    // Call update, not set! Update sets the value but does not
                    // cause a DataMessage to be sent. We want to send one
                    // DataMessage with all the updated data items.
                    //
                    // There should be no issues with calling update here. All the
                    // checking was done while building the event chain, and
                    // newValue should be the converted data from the
                    // RequestMessage.
                    let _ = try attribute.update(value: newValue!)
                    
                    let dma: DeviceModelAttribute = attribute.getDeviceModelAttribute()!
                    switch (dma.getType()) {
                    case .INTEGER:
                        try dataItems.append(DataItem(key: attributeName,
                                                      value: newValue as! Int))
                        break
                    case .NUMBER:
                        try dataItems.append(DataItem(key: attributeName,
                                                      value: newValue as! Double))
                        break
                    case .BOOLEAN:
                        try dataItems.append(DataItem(key: attributeName,
                                                      value: newValue as! Bool))
                        break
                    case .STRING:
                        try dataItems.append(DataItem(key: attributeName,
                                                      value: newValue as? String))
                        break
                    case .DATETIME:
                        try dataItems.append(DataItem(key: attributeName,
                                                      value: newValue as! Double))
                        break
                    case .URI:
                        try dataItems.append(DataItem(key: attributeName,
                                value: (newValue as! ExternalObject).getURI()))
                        break
                    }
                    namedValue = namedValue!.next()
                }
                
                let builder: DataMessageBuilder = DataMessageBuilder()
                builder.setFormat(format: self.impl!.getDeviceModel().getURN() + ":attributes")
                builder.setSource(source: self.impl!.getEndpointId())
                builder.addDataItems(dataItems: dataItems)
                let dataMessage: DataMessage = try builder.build()
                
                let messageDispatcher: MessageDispatcher =
                    try MessageDispatcher.getMessageDispatcher(
                        msgDirectlyConnectedDevice: self.impl!.msgDirectlyConnectedDevice!)
                try messageDispatcher.queue(message: dataMessage)
                
                callback(.ACCEPTED)
            } catch {
                // The onChange method may throw exceptions. Since the CL has no
                // knowledge of what these might be, Exception is caught here.
                callback(.INTERNAL_SERVER_ERROR)
            }
            
        }) // end notify
    }
    
    // POST is for actions
    private func handlePost(requestMessage: RequestMessage,
                            callback: @escaping (StatusCode)->()) {
        
        let path: String = requestMessage.getURL()
        let dmURN: String = "deviceModels/" + impl!.getDeviceModel().getURN() + "/actions"
        var actionName: String = path
        if dmURN == path[path.startIndex...dmURN.index(before: dmURN.endIndex)] {
            actionName = String(path[dmURN.endIndex..<path.endIndex])
            actionName = String(actionName[actionName.index(after: actionName.startIndex)..<actionName.endIndex])
        }
        
        if impl!.actionMap != nil && impl!.actionMap![actionName] != nil {
            let callable: ((VirtualDevice, AnyObject?) -> Void)? = impl!.actionMap![actionName]!
            if callable != nil {

                let action: DeviceModelAction =
                    (VirtualDeviceImpl.getDeviceModelAction(deviceModel: impl!.base!.getDeviceModel(),
                                                            actionName: actionName))!

                if action.getArgType() != nil {

                    // Pass the action name as the attributeName
                    impl!.getValue(attributeType: action.getArgType()!,
                                   attributeName: actionName,
                                  payload: requestMessage.getBody(),
                                  callback: {(value, error) in
                        if let error = error {
                            print(error)
                            callback(.BAD_REQUEST)
                            return
                        }
                        callable!(self.impl!, value)
                        callback(.ACCEPTED)
                    })
                } else {
                    callable!(impl!, nil)
                    callback(.ACCEPTED)
                }
            }
        } else {
            callback(.NOT_FOUND)
        }
    }
    
    // PUT is for attributes or for resources
    private func handlePut(requestMessage: RequestMessage,
                           callback: @escaping (StatusCode)->()) {

        let path: String = requestMessage.getURL()
        let dmURN: String = "deviceModels/" + impl!.getDeviceModel().getURN() + "/attributes"
        var attributeName: String = path
        if dmURN == path[path.startIndex...dmURN.index(before: dmURN.endIndex)] {
            attributeName = String(path[dmURN.endIndex..<path.endIndex])
            attributeName = String(attributeName[attributeName.index(after: attributeName.startIndex)..<attributeName.endIndex])
        }
        var virtualDeviceAttribute: VirtualDeviceAttribute? = nil
        do {
            virtualDeviceAttribute = try impl!.adapter!.getAttribute(name: attributeName)
        } catch {
            callback(.BAD_REQUEST)
            return
        }
        let deviceModelAttribute: DeviceModelAttribute =
            virtualDeviceAttribute!.getDeviceModelAttribute()!
        let attributeType: DeviceModelAttributeType = deviceModelAttribute.getType()
        let data: Foundation.Data = requestMessage.getBody()
        
        impl!.getValue(attributeType: attributeType,
                       payload: data,
                       callback: { (newValue, error) in
                        
            if let error = error {
                print(error)
                callback(.BAD_REQUEST)
                return
            } else {
                // TODO: Can error and newValue both be nil ?
                let namedValue = NamedValueImpl(name: attributeName, value: newValue!)
                // Always create the event
                let event:ChangeEventImpl = ChangeEventImpl(
                    virtualDevice: self.impl!, value: namedValue)
                let attrOnChange = virtualDeviceAttribute?.getOnChange()
                if attrOnChange != nil {
                    attrOnChange!(event)
                }
                if self.impl?.base?.getOnChangeCallback() != nil {
                    self.impl!.base?.getOnChangeCallback()!(event)
                } else if attrOnChange == nil {
                    callback(.NOT_IMPLEMENTED)
                    return
                }
                
                // Call set, not update! Update sets the value but does not
                // cause a DataMessage to be sent. Since the put is for a
                // single attribute only, we want the DataMessage to be sent.
                // This is different than handlePatch.
                do {
                    try virtualDeviceAttribute?.set(value: newValue!)
                } catch {
                    callback(.BAD_REQUEST)
                    return
                }
                callback(.ACCEPTED)
                return
            }

        })
    }
}

class VDIAdapter: Adapter {
    
    var impl: VirtualDeviceImpl? = nil
    
    init(impl: VirtualDeviceImpl?) {
        self.impl = impl
    }
    
    convenience init() {
        self.init(impl: nil)
    }
    
    /**
     * Return a `VirtualDeviceAttribute`.
     *
     * - parameter name: the name of the attribute
     *
     * - returns: The `VirtualDeviceAttribute` for the given `name`
     *
     * - throws: `ClientError.argument` if `name` is not a device attribute.
     */
    func getAttribute(name: String) throws -> VirtualDeviceAttribute {
        
        let virtualDeviceAttribute: VirtualDeviceAttribute? = impl!.attributeMap[name]
        
        if (virtualDeviceAttribute == nil) {
            throw ClientError.argument(String(format: NSLocalizedString(
                "Unexpected attribute %@", comment: ""), name))
        }
        
        return virtualDeviceAttribute!
    }
    
    
    /**
     * Sets the the value for attribute by calling the attribute's set method.
     * Invokes the on error handlers on error.
     *
     * - parameter deviceAttribute: the device attribute name
     *
     * - parameter value: the device attribute value
     *
     * - throws: `ClientError.argument` if `attribute` is `nil` or if
     *           `value` is an invalid value<br/>
     *           `ClientError.network` if a `MessageDispatcher` cannot be 
     *           obtained.
     *
     */
    func setValue(deviceAttribute attribute: VirtualDeviceAttribute?,
                           value: AnyObject) throws {
        
        if (attribute == nil) {
            throw ClientError.argument(NSLocalizedString(
                "Attribute must not be nil", comment: ""))
        }
        
        try attribute!.set(value: value)
    }
    
    /**
     * Set all the attributes in an update batch. Errors are handled in the
     * set call, including calling the on error handler.
     *
     * - parameter updatedAttributes: The attributes modified during an update
     *
     * - throws: 
     *    `ClientError.network` if a `MessageDispatcher` cannot be
     *           obtained.
     *    `ClientError.argument` if an attribute is `nil` or is the empty
     *           string or if an attribute type is string, if `newValue` is 
     *           `nil` or the empty string.
     */
    func updateFields(updatedAttributes: [VirtualDeviceAttribute : AnyObject]) throws {
        // TODO: FIXME does not bubble up throws from update.
        // Throws doc is from processOnChange.
        for attribute in updatedAttributes.keys {
            do {
                
                // Here, we assume
                // 1. That attribute is not null. If the attribute were not found
                //    an exception would have been thrown from the VirtualDevice
                //    set(String attributeName, T value) method.
                // 2. That the set method validates the value. The call to
                //    update here should not throw an exception because the
                //    value is bad.
                // 3. The author of this code knew what he was doing.
                
                let _ = try attribute.update(value: updatedAttributes[attribute]!)
                
            } catch {
            }
        }
        
        try impl!.processOnChange(updatedAttributes: updatedAttributes)
        
    }
}

class ErrorCallbackBridge: MessageDispatcher.ErrorCallback {
    
    private var deviceSet: [String :VirtualDeviceImpl]
    
    override init() {
        deviceSet = [ : ]
    }
    
    func add(virtualDevice: VirtualDeviceImpl) throws {
        deviceSet[virtualDevice.getEndpointId()] = virtualDevice
    }
    
    override func failed(messages: [Message], error: ClientError? = nil) throws {
        var collatedMessages: [String : [Message]] = [ : ]
        var list: [Message]?
        
        // Collate the messages by source
        for message in messages {
            let source: String = message.getSource()
            list = collatedMessages[source]
            if (list == nil) {
                list = []
                collatedMessages[source] = list
            }
            list!.append(message)
        }
        
        // Call the onError method of
        for (key, _) in collatedMessages {
            
            let ref: VirtualDeviceImpl? = deviceSet[key]
            if ref == nil {
                throw ClientError.state(NSLocalizedString(
                    "Empty device set", comment: ""))
            }
            let virtualDevice: VirtualDeviceImpl? = ref
            if virtualDevice != nil {
                try invokeErrorCallback(virtualDevice: virtualDevice!,
                            messages: list! /*, exception*/)
            }
        }
    }
    
    // TODO: FIXME pass exception or exception message and forward to errorCallback
    private func invokeErrorCallback(virtualDevice: VirtualDeviceImpl,
                                     messages: [Message] /*, Exception exception*/) throws {
        var values: NamedValueImpl? = nil
        for message in messages {
            var dataItems : [DataItem]? = nil
            if (message is DataMessage) {
                let dataMessage: DataMessage = message as! DataMessage
                dataItems = dataMessage.getDataItems()
            } else if (message is AlertMessage) {
                let alertMessage: AlertMessage = message as! AlertMessage
                dataItems = alertMessage.getDataItems()
            }
            
            if (dataItems != nil) {
                var last: NamedValueImpl? = nil
                for dataItem in dataItems! {
                    
                    let namedValue: NamedValueImpl = NamedValueImpl(
                        name: dataItem.getKey(),
                        value: dataItem.getValue())
                    
                    
                    let attribute: VirtualDeviceAttribute =
                        try virtualDevice.adapter!.getAttribute(name: dataItem.getKey())
                    
                    if (attribute.getOnError() != nil) {
                        attribute.getOnError()!(ErrorEventImpl(virtualDevice: virtualDevice,
                                                               value: namedValue, message: "" ))
                    }
                    
                    
                    if (last != nil) {
                        last!.setNext(next: namedValue)
                        last = namedValue
                    } else {
                        values = namedValue
                        last = namedValue
                    }
                }
            }
        }
        
        if let errrCallback = virtualDevice.base?.getOnErrorCallback() {
            // Need to pass non-nil NamedValueImpl instance
            errrCallback(ErrorEventImpl(virtualDevice: virtualDevice,
                value: values ?? NamedValueImpl(name:"unknown", value: nil),
                message: NSLocalizedString("Unexpected error", comment: "")))
        }
    }
}
