/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * An `Application` lists the characteristics of an IoT CS application.
 *
 * - seealso: `EnterpriseClient.getApplication()`
 *
 */

public class Application {
    
    private let id: String
    private let name: String
    private let description: String
    private let metadata: [String : String]
    
    init(id: String, name: String, description: String,
                                   metadata: [String : String]) {
        self.id = id
        self.name = name
        self.description = description
        self.metadata = metadata
    }
    
    static func fromJson(object: JSON) -> Application {
        
        var mt: [String : String] = [:]
        let jsonDictionary = object.dictionaryValue
        let metadata = jsonDictionary["metadata"]
        if metadata!.count != 0 {
            for i in 0 ..< metadata!.count {
                mt[(metadata![i].array?.first?.stringValue)!] =
                                        metadata![i].array?.last?.stringValue
            }
        }
        
        let appDescription = (jsonDictionary["description"] != nil) ?
                             (jsonDictionary["description"]?.stringValue)! : ""
        
        return Application(id: (jsonDictionary["id"]?.stringValue)!,
                           name: (jsonDictionary["name"]?.stringValue)!,
                           description: appDescription, metadata: mt)
    }
   
    /**
     * Application identifier.
     *
     * - returns: The identifier of the application.
     */
    public func getId() -> String {
        return self.id
    }
    
    /**
     * Name of the application.
     *
     * - returns: The name of the application.
     */
    public func getName() -> String {
        return self.name
    }
    
    /**
     * Free form description of the application.
     *
     * - returns: The description of the application.
     */
    public func getDescription() -> String {
        return self.description
    }
    
    /**
     * Get the application Metadata for the specified `key`.
     *
     * - parameter key: The metadata to look for.
     *
     * - returns: The value corresponding to the specified `key`.
     */
    public func getMetadata(key: String) -> String? {
        if self.metadata[key] == nil {
            return nil
        }
        return self.metadata[key]!
    }
    
    /**
     * Get the application Metadata.
     *
     * - returns: A Dictionary associating keys and values.
     */
    public func getMetadata() -> [String : String] {
        return self.metadata
    }
}
