/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * A `Device` lists the characteristics of a device.
 *
 * - seealso: `EnterpriseClient.getActiveDevices(deviceModel:callback:)`
 */
public class Device {
    
    /**
     * The `Field` enum defines the characteristics of a device.
     */
    public class Field : Hashable {
        
        /**
         * The constant values of this enum.
         */
        public enum Values : String {
            /// A unique value identifying the device
            case ID = "id"

            /// A string representing the device unique identifier.
            case DEVICE_UID = "deviceUID"

            /// The device type
            case TYPE = "type"

            /// Optional free form text describing the device.
            case DESCRIPTION = "description"

            /// Date and time the endpoint was created as a String in ISO-8601 
            /// time format.
            case CREATED = "created"

            /// The device state.
            case STATE = "state"

            /// The device name.
            case NAME = "name"

            /// A string (generally an OUI) describing the manufacturer of the 
            /// device.
            case MANUFACTURER = "manufacturer"

            /// A string representing the specific model of the device.
            case MODEL_NUMBER = "modelNumber"

            /// A string (which should be unique across all devices of this 
            /// modelNumber / manufacturer) uniquely identifying the specific 
            /// device.
            case SERIAL_NUMBER = "serialNumber"

            /// A string representing the hardware revision.
            case HARDWARE_REVISION = "hardwareRevision"

            /// A string representing the software revision.
            case SOFTWARE_REVISION = "softwareRevision"

            /// A string representing the software version. This field is 
            /// available only for gateway devices.
            case SOFTWARE_VERSION = "softwareVersion"

            /// `true` if the device is currently enabled.
            case ENABLED = "enabled"

            /// Connection status of the device.
            case CONNECTIVITY_STATUS = "connectivityStatus"

            /// Last time the device sent a message.
            case LAST_HEARD_TIME = "lastHeardTime"

            /// A list of device models
            case DEVICE_MODELS = "deviceModels"
            
            /// Additional user-defined key / value pairs describing the 
            /// endpoint.
            case METADATA = "metadata"
            
            /// The device location.
            case LOCATION = "location"

            /// For endpoints which are not themselves directly connected to the
            /// IoT network, this points to the endpoint which is handling 
            /// communications on their behalf.
            case DIRECTLY_CONNECTED_OWNER = "directlyConnectedOwner"
            
            /// To access all values
            static let allValues = [ID, DEVICE_UID, TYPE, DESCRIPTION, CREATED,
                                    STATE, NAME, MANUFACTURER, MODEL_NUMBER,
                                    SERIAL_NUMBER, HARDWARE_REVISION,
                                    SOFTWARE_REVISION, SOFTWARE_VERSION,
                                    ENABLED, CONNECTIVITY_STATUS,
                                    LAST_HEARD_TIME, DEVICE_MODELS,
                                    METADATA, LOCATION,
                                    DIRECTLY_CONNECTED_OWNER]

        }
        
        /// The alias used by server REST API
        let valias: String
        
        /// Hashable implementation
        public var hashValue: Int {
            return alias().hashValue
        }
        
        /**
         * Get the `Field` instance associated with the specified alias. The
         * String must match exactly, including case.
         *
         * - parameter alias: The alias to look for
         *
         * - returns: The Field instance matching the specified alias
         *
         * - seealso: `Field`
         */
        public static func fromAlias(alias: String) -> Field {
            for v in Values.allValues {
                if v.rawValue == alias {
                    return Field(alias: alias)
                }
            }
           return Field(alias: "")
        }
        
        /**
         * Get a new value set initialized with all possible `Field` values.
         *
         * - returns: A new `Field.Values` of all possible `Field` values.
         */
        public static func all() -> [Field.Values] {
            return Values.allValues
        }
        
        /**
         * Get the alias of this `Field`. The alias is the name the server uses
         * for the field.
         *
         * - returns: The alias of this `Field`.
         *
         * - seealso: `Field`
         */
        public func alias() -> String {
            return self.valias
        }
        
        /**
         * Create a `Field` instance.
         *
         * - parameter valias: The alias name used by server
         */
        init(alias: String) {
            self.valias = alias
        }
        
    }
    
    /**     
     * Build a new `Device` instance.
     */
    init() {
        
    }
    
    /**
     * Get the list of device models supported by this device.
     *
     * - returns: A list of device model URN.
     */
    
    public func getDeviceModels() -> [String] {
        fatalError("NOT IMPLEMENTED")
    }
    
    
    /**
     * Get the value of the specified `field` or an empty String if the value is
     * unknown.
     *
     * - parameter field: The field to look for
     *
     * - returns: The field value or an empty String.
     */
    
    public func getValue(field: Device.Field) -> String {
        fatalError("NOT IMPLEMENTED")
    }
    
    
    /**
     * Return whether the device is directly connected to the server or is 
     * connected through a gateway.
     *
     * - returns: `true` if the device is directly connected, `false` otherwise.
     */
    
    public func isDirectlyConnected() -> Bool {
        fatalError("NOT IMPLEMENTED")
    }
    
    
    /**
     * Get the device identifier.
     *
     * - returns: The device identifier.
     */
    
    public func getId() -> String {
        fatalError("NOT IMPLEMENTED")
    }
    
    
    /**
     * Return the device metadata. Might be empty.
     *
     * - returns: A map of device metadata
     */
    
    public func getMetadata() -> [String : String] {
        fatalError("NOT IMPLEMENTED")
    }
    
    /**
     * Return the value of the metadata field `key`.
     *
     * - parameter key: The metadata key to return.
     *
     * - returns: Value for metadata `key` or `nil` if not found.
     */
    
    public func getMetadata(key: String) -> String? {
        fatalError("NOT IMPLEMENTED")
    }
    
    /**
     * Return the device's location.
     * 
     * - returns: The device's location.
     */
    public func getLocation() -> Location? {
        fatalError("NOT IMPLEMENTED")
    }
    
    /**
     * The `Device.Location` class describing the location of a device.
     */
    public class Location {
        /// The device's latitude in decimal degrees.
        private let latitude: Float
        
        /// The device's longitude in decimal degrees.
        private let longitude: Float
        
        /// The device's altitude in meters.
        private let altitude: Float?
        
        /// The device's uncertainty in meters.
        private let uncertainty: Float?
        
        /**
         * Build a new `Location` instance.
         */
        public init(latitude: Float, longitude: Float, altitude: Float?, uncertainty: Float?) {
            self.latitude = latitude
            self.longitude = longitude
            self.altitude = altitude
            self.uncertainty = uncertainty
        }
        
        /**
         * Gets the latitude in decimal degrees.
         *
         * - returns: the latitude
         */
        public func getLatitude() -> Float {
            return self.latitude
        }
        
        /**
         * Gets the longitude in decimal degrees.
         *
         * - returns: the longitude
         */
        public func getLongitude() -> Float {
            return self.longitude
        }
        
        /**
         * Gets the altitude in meters.
         *
         * - returns: the altitude
         */
        public func getAltitude() -> Float? {
            return self.altitude
        }
        
        /**
         * Gets the uncertainty in meters.
         *
         * - returns: the uncertainty
         */
        public func getUncertainty() -> Float? {
            return self.uncertainty
        }
    }
}

