/*
 *  Copyright © 2016 Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * A *DeviceApp* represents an application, running on a device, and implementing its own device model.
 *
 */

public class DeviceApp {
    
    /**
     * The *Field* enum defines the characteristics of a device.
     */
    
    public class Field : Hashable {
 
        public enum Values : String {
            

            /// A unique value identifying the device application
            
            case ID = "id"

            /// The endpoint type
            
            case TYPE = "type"
            
            /// Optional free form text describing the device application.
            
            case DESCRIPTION = "description"
            
            /// The device application name.
            
            case APP_NAME = "appName"
           
            /// The device application name.
            
            case APP_VERSION = "appVersion"
            
            /// The device application type.
            
            case APP_TYPE = "appType"
            
            /// For endpoints which are not themselves directly connected to the IoT network, this points to the endpoint which is handling communications on their behalf.
            
            case DIRECTLY_CONNECTED_OWNER = "directlyConnectedOwner"
            
            /// Additional user-defined key / value pairs describing the endpoint.
            
            case METADATA = "metadata"
        
            ///String in ISO-8601 time format indicating the date and time the endpoint was created.
     
            case CREATED = "created"

            /// The device state.

            case STATE = "state"

            ///A list of device models

            case DEVICE_MODELS = "deviceModels"
            
            /// To access all values
            static let allValues = [ID, TYPE, DESCRIPTION, APP_NAME, APP_VERSION, APP_TYPE, DIRECTLY_CONNECTED_OWNER, METADATA, CREATED, STATE, DEVICE_MODELS]
            
        }
        
        /// The alias used by server REST API
        let valias: String
        
        // Hashable implementation
        public var hashValue: Int {
            return alias().hashValue
        }
        
        /**
         * Get the *Field* instance associated with the specified alias.
         *
         *  - Parameter alias: The alias to look for
         *
         *  - Returns: The *Field* instance matching the specified alias
         *
         *  - SeeAlso: Field
         */
        public static func fromAlias(alias: String) -> Field {
            for v in Values.allValues {
                if v.rawValue == alias {
                    return Field(alias: alias)
                }
            }
            return Field(alias: "")
        }
        
        /**
         * Get a new value set initialized with all possible *Field* values.
         *
         *  - Returns: A new *Field.Values*} of all possible *Field* values.
         */
        public static func all() -> [Field.Values] {
            return Values.allValues
        }
        
        /**
         * Get the alias of this *Field*.
         *
         *  - Returns: The alias of this *Field*.
         */
        public func alias() -> String {
            return self.valias
        }
        
        /**
         * Create a *Field* instance
         *
         *  - Parameter alias: The alias name used by server
         */
        public init(alias: String) {
            self.valias = alias
        }
    }
    
}

// Hashable implementation
public func ==(lhs: DeviceApp.Field, rhs: DeviceApp.Field) -> Bool {
    return lhs.valias == rhs.valias
}
