/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * A *DeviceAppEnumerator* can be used to enumerate device applications.
 */
public class DeviceAppEnumerator {
    
    /// The application identifier: namespace to use to look for messages
    private let appID: String?
    
    /// The secure connection to use
    private let secureConnection: HttpSecureConnection

    /**
     * Creates a message enumerator to retrieve messages.
     *
     * - Parameter client: The client to use to retrieve messages
     */
    public init(client: EnterpriseClient) {
        if client.getApplication().getId().isEmpty == false {
            self.appID = client.getApplication().getId()
        }
        else {
            self.appID = ""
        }
    
        self.secureConnection = SecureHttpConnectionMap.getSecureHttpConnection(client: client)
    }
    
    /**
     * Return a *DeviceAppPageable* collection of *DeviceApp* matching the
     * specified filter.
     *
     * The *Filter* forms a query. If not *nil*, only device applications that 
     * satisfy the *Filter* are returned.
     *
     * - Parameter filter: A filter to constrain the collection of device apps. 
     *                     If *nil*, all the device applications are returned.
     *
     * - Parameter iterator: An *DeviceAppPageable* collection of *DeviceApp* 
     *                       that satisfy the *Filter*. If no matching device 
     *                       application is found, an *DeviceAppPageable* with 
     *                       no elements is returned.
     *
     * - Throws: IOException if an I/O error occurred when trying to retrieve 
     *           data from server
     *
     * - Throws: GeneralSecurityException when key or signature algorithm class 
     *           cannot be loaded, or the key is not in the trusted assets 
     *           store, or the private key is invalid. If User Authentication 
     *           is used, then *UserAuthenticationException* will be thrown if 
     *           the session cookie has expired.
     *
     * - SeeAlso: DeviceAppPageable
     *
     * - SeeAlso: DeviceApp
     *
     */
    public func getDeviceApps(filter: Filter?, iterator: @escaping (DeviceAppPageable?, ClientError?) -> Void) throws -> Void {
        let deviceAppsIterator = DeviceAppPageable(appID: self.appID!, offset: 0, limit: 10,
                                                       filter: filter, secureConnection: self.secureConnection)
        
        try deviceAppsIterator.setSizeOnCount(response: { httpIterator, error in
            if (error != nil) {
                iterator(nil, error)
            } else {
                if (httpIterator != nil) {
                    iterator(deviceAppsIterator, nil)
                } else {
                    iterator(nil, ClientError.argument(NSLocalizedString("DeviceAppEnumerator getDeviceApps() httpIterator should not be nil", comment: "")))
                }
            }
        })
    }
    
    /**
     * Creates a request for *GET iot/api/v2/deviceApps* or *GET iot/api/v2/apps/
     * {a-id}/deviceApps*.
     */
    public class DeviceAppEnumerationRequest {

        /// List of headers required for the REST API

        private let HEADERS = ["Content-Type" : "application/json", "Accept" : "application/json"]
       

        /// The REST request including parameters
        
        private let REQUEST: String
        
        /**
         * Creates a DeviceAppEnumerationRequest matching the specified criteria
         * *"/iot/api/v2/apps/{applicationId}/devices/{path}?offset={offset}&limit={limit}"*
         *
         * - Parameter applicationID: The application identifier in which we 
         *                            should look for devices
         *
         * - Parameter path: The path to extend the request. If *nil*, not 
         *                   added to the request
         *
         * - Parameter limit: The maximum number of result to return
         *
         * - Parameter offset: The offset of the results in the entire list of 
         *                     devices matching the specified criteria
         *
         * - Parameter filter: Filter to be applied
         */
        public init(applicationID: String, path: String, offset: Int, limit: Int, filter: Filter?) {
            var req: String = ""
            var sep: String = "?"
            
            if applicationID.isEmpty == true {
                req = RestApi.V2.getReqRoot()
            } else {
                req = RestApi.V2.getReqRoot() + "/apps/" + applicationID
            }
            
            req += "/deviceApps"
            if path.isEmpty == false {
                req += "/" + path
            }
            
            if limit > 0 {
                req += sep + "limit=" + String(limit)
                sep = "&"
            }
            
            if offset > 0 {
                req += sep + "offset=" + String(offset)
                sep = "&"
            }
            req += sep + "totalResults=true"
            
            // build query
            if filter != nil {
                req += "&q="
                req += filter!.toJsonURI()
            }

            self.REQUEST = req
            
        }
        
        /**
         * Get the REST call (including parameters) representing this request
         *
         * - Returns: The REST call (including parameters)
         */
        public func request() -> String {
            return self.REQUEST
        }
        
        /*
         * Get the list of HTTP headers required for this request
         *
         * - Returns: HTTP headers required to perform this request
         */
        public func headers() -> [String : String] {
            return self.HEADERS
        }
    }
    
    /**
     * A new iterator for device applications.
     */
    public final class DeviceAppPageable : PageableEnumerator<DeviceApp> {

        /// Filter to apply to constrain the result

        private let filter: Filter?
        
        /// The application identifier: namespace to use to look for devices

        private let appID: String
        

        /// The secure connection to use

        private let secureConnection: HttpSecureConnection
        
         /**
          * Instantiate a new iterator.
          *
          * - Parameter appID: The application identifier if looking for devices
          *                    of this application, otherwise *nil*.
          *
          * - Parameter offset: The offset of the first item to fetch
          *
          * - Parameter limit: The maximum number of items requested per REST 
          *                    call
          *
          * - Parameter filter: List of filters to apply to the result
          *
          * - Parameter secureConnection: The server connection
          */
        init(appID: String, offset: Int, limit: Int, filter: Filter?, secureConnection: HttpSecureConnection) {
            self.appID = appID
            self.filter = filter
            self.secureConnection = secureConnection
            super.init(offs: offset, lim: limit)

        }
        
        public func setSizeOnCount(response: @escaping (HttpResponse?, ClientError?) -> Void) throws -> Void {
            try self.getCount(countResponse: { countResponse, error in
                if (error != nil) {
                    response(nil, error)
                } else {
                    if let countObject = countResponse {
                        let data = countObject.getData()
                        if data.count == 0 {
                            // TODO: why countResponse callback is not called?
                            return
                        }
                        var jsonObject:JSON!
                        do {
                            jsonObject = try JSON(data: data as Foundation.Data)
                        } catch {
                            response(nil, ClientError.network(String(format:NSLocalizedString(
                                "%@: JSON threw exception: %@",
                                comment:""), "DeviceAppEnumerator.setSizeOnCount()", "\(error)")))
                            return
                        }
                        self.setSize(size: (jsonObject.dictionaryValue["count"]?.intValue)!)
                        response(countObject, nil)
                    } else {
                        response(nil, ClientError.argument(NSLocalizedString("DeviceAppEnumerator setSizeOnCount() countResponse should not be nil", comment: "")))
                    }
                }
            })
        }
        
        /**
         * Performs a new REST call to get the count of deviceApps matching the 
         * filter
         *
         * - Returns: The count response of device applications matching the 
         *            request
         */
        private func getCount(countResponse: @escaping (HttpResponse?, ClientError?) -> Void) throws -> Void {
            let req = DeviceAppEnumerationRequest(applicationID: self.appID, path: "count", offset: self.voffset, limit: self.vlimit, filter: self.filter)
            try self.secureConnection.get(restApi: req.request(), headers: req.headers(), response: { responseGet, error in
                if (error != nil) {
                    countResponse(nil, error)
                } else {
                    if let getObject = responseGet {
                        countResponse(getObject, nil)
                    } else {
                        countResponse(nil, ClientError.argument(NSLocalizedString("DeviceAppEnumerator getCount() responseGet should not be nil", comment: "")))
                    }
                }
            })
        }
        
        public func next(enumerator: @escaping (DeviceAppPageable?, ClientError?) -> Void) throws -> Void {
            self.offset(offset: self.voffset + self.velements.count, pageResponse: { retPage, err in
                if (err != nil) {
                    enumerator(nil, err)
                } else {
                    if (retPage != nil) {
                        if self.velements.isEmpty == false {
                            enumerator(self, nil)
                        }
                        // TODO else {} ?
                    } else {
                        enumerator(nil, ClientError.state(NSLocalizedString("DeviceAppEnumerator next() retPage should not be nil", comment: "")))
                    }
                }                
            })
        }
        
        private func offset(offset: Int, pageResponse: @escaping (DeviceAppEnumerationResponse?, ClientError?) -> Void) {
            do {
                try self.load(offset: offset, limit: self.vlimit, pageResponse: { returnPage, error in
                    if (error != nil) {
                        pageResponse(nil, error)
                    } else {
                        if let pageObject = returnPage {
                            self.velements = pageObject.elements()
                            self.vhasMore = pageObject.hasMore()
                            self.voffset = pageObject.offset()
                            self.vsize = pageObject.total()
                            pageResponse(pageObject, nil)
                        } else {
                            pageResponse(nil, ClientError.argument(NSLocalizedString("DeviceAppEnumerator offset() returnPage should not be nil", comment: "")))
                        }
                    }
                })
            } catch {
                pageResponse(nil, ClientError.network(NSLocalizedString("Error while loading device application page at offset " + offset.description, comment: "")))
            }
        }
        
        private func load(offset: Int, limit: Int, pageResponse: @escaping (DeviceAppEnumerationResponse?, ClientError?) -> Void) throws -> Void  {
            let req = DeviceAppEnumerationRequest(applicationID: self.appID, path: "", offset: offset, limit: limit, filter: self.filter)
            try self.secureConnection.get(restApi: req.request(), headers: req.headers(), response: { getResponse, error in
                if (error != nil) {
                    pageResponse(nil, error)
                } else {
                    if let getObject = getResponse {
                        let data = getObject.getData()
                        do {
                            let jsonObject = try JSON(data: data as Foundation.Data)
                            let page = try DeviceAppEnumerationResponse.fromJson(jsonObject: jsonObject)
                            pageResponse(page, nil)
                        } catch {
                            pageResponse(nil, ClientError.network(NSLocalizedString("Error while parsing device application page data", comment: "")))
                        }
                    } else {
                        pageResponse(nil, ClientError.argument(NSLocalizedString("DeviceAppEnumerator load() getResponse should not be nil", comment: "")))
                    }
                }
            })
        }
        
    }
    
    /**
     * Response to *GET iot/api/v2/deviceApps* or 
     * *GET iot/api/v2/apps/{a-id}/deviceApps*.
     */
    fileprivate class DeviceAppEnumerationResponse : PagedResponse<DeviceAppImpl> {
        
        init(deviceApps: [DeviceAppImpl], offset: Int, total: Int, hasMore: Bool, links: [String : String]) {
            super.init(elements: deviceApps, offset: offset, total: total, hasMore: hasMore, links: links)
        }
        
        
        /**
         * Creates a *DeviceAppEnumerationResponse* from a JSON object 
         * received in response to a *DeviceAppEnumerationRequest*.
         *
         * - Parameter jsonObject: The JSON object to parse.
         *
         * - Returns: A *DeviceAppEnumerationResponse* instance.
         *
         * - Throws: JsonException if the jsonObject has not the expected format.
         *
         * - SeeAlso: DeviceAppEnumerationRequest
         */
        fileprivate static func fromJson(jsonObject: JSON) throws -> DeviceAppEnumerationResponse {
            let items = jsonObject.dictionaryValue["items"]?.arrayValue
            let links = jsonObject.dictionaryValue["links"]?.arrayValue
            var hasMore = jsonObject.dictionaryValue["hasMore"]?.boolValue
            if hasMore == nil {
                hasMore = false
            }
            
            var offset = jsonObject.dictionaryValue["offset"]?.intValue
            if offset == nil {
                offset = -1
            }
            
            var total = jsonObject.dictionaryValue["total"]?.intValue
            if total == nil {
                total = -1
            }
            
            var list: [DeviceAppImpl] = [DeviceAppImpl]()
            if items != nil && items?.count != 0 {
                for item: JSON in items! {
                    let device = try DeviceAppImpl.from(object: item)
                    list.append(device)
                }
            }
            
            var map: [String : String] = [String : String]()
            if links != nil && links?.count != 0 {
                for value: JSON in links! {
                    let rel = value.dictionaryValue["rel"]?.stringValue
                    let href = value.dictionaryValue["href"]?.stringValue
                    if rel?.isEmpty == false && href?.isEmpty == false {
                        map[rel!] = href
                    }
                }
            }
            
            return DeviceAppEnumerationResponse(deviceApps: list, offset: offset!, total: total!, hasMore: hasMore!, links: map)
        }
    }
}
