/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * `DeviceApp` implementation.
 */
class DeviceAppImpl : DeviceApp {
    
    private static let DEVICE_APPLICATION = "DEVICE_APPLICATION"
    
    private let fields: [DeviceApp.Field : String]
    private let id: String
    private let deviceModels: [String]
    private let metadata: [String : String]
    
    init(id: String, fields: [DeviceApp.Field : String], deviceModels: [String],
         metadata: [String : String]) {
        self.id = id
        self.fields = fields
        self.deviceModels = deviceModels
        self.metadata = metadata
    }
    
    /**
     * Creates a new `DeviceApp` instance from the given json object.
     *
     * - Parameter object: The json object to use to build the `DeviceApp` 
     *                     instance.
     *
     * - Returns: A new `DeviceApp` instance created from the given Json object.
     *
     * - throws: `ClientError.state` if `object` is not a `DEVICE_APPLICATION`
     *           or the ID is not set.
     */
    static func from(object: JSON) throws -> DeviceAppImpl {
        var fields: [DeviceApp.Field : String] = [:]
        let entrySet = object.dictionaryValue
        
        for e in entrySet {
            let key = e.0
            let f = DeviceApp.Field.fromAlias(alias: key)
            if f.alias().isEmpty != false {
                fields[f] = e.1.stringValue
            }
        }

        let t: String =
            fields[DeviceApp.Field.fromAlias(alias: Field.Values.TYPE.rawValue)]!
        if DEVICE_APPLICATION != t {
            throw ClientError.state(String(format: NSLocalizedString(
                "Not a device Application: %@", comment: ""), t))
        }
        
        let id: String =
            fields[DeviceApp.Field.fromAlias(alias: Field.Values.ID.rawValue)]!
        if id.isEmpty == true {
            throw ClientError.state(NSLocalizedString(
                "Id is missing", comment: ""))
        }
        
        // get device models
        let jsonModels =
            entrySet[DeviceApp.Field.Values.DEVICE_MODELS.rawValue]?.arrayValue
        var deviceModels: [String] = [String]()
        
        if jsonModels != nil && jsonModels?.count != 0 {
            for model in jsonModels! {
                let urn: String = (model.dictionaryValue.first?.0)!
                deviceModels.append(urn)
            }
        }
        
        // get metadata
        let jsonMetadata =
            entrySet[DeviceApp.Field.Values.METADATA.rawValue]?.arrayValue
        var metadata: [String : String] = [String : String]()
        

        if jsonMetadata != nil && jsonMetadata?.count != 0 {
            for entry in jsonMetadata! {
                let key = entry.first!.0
                let value = entry.first!.1
                metadata[key] = value.stringValue
            }
        }
        
        return DeviceAppImpl(id: id, fields: fields, deviceModels: deviceModels,
                             metadata: metadata)
    }
    
    /**
     * Get the endpoint identifier.
     *
     * - Returns: The endpoint identifier.
     */
    func getId() -> String {
        return self.id
    }
    
    /**
     * Return the value of the metadata field `key`.
     *
     * - Parameter key: The metadata key to return.
     *
     * - Returns: Value for metadata `key` or `nil` if not found.
     */
    func getMetadata(key: String) -> String? {
        if self.metadata[key] == nil {
            return nil
        }
        return self.metadata[key]!
    }
    
    /**
     * Get the list of device models supported by this endpoint.
     *
     * - Returns: A list of device model URN.
     */
    func getDeviceModels() -> [String] {
        return self.deviceModels
    }
    
    /**
     * Return the endpoint metadata. Might be empty.
     *
     * - Returns: A map of device metadata
     */
    func getMetadata() -> [String : String] {
        return self.metadata
    }
    
    /**
     * Get the value of the specified `field` or an empty String if the value 
     * is unknown.
     *
     * - Parameter field: The field to look for.
     *
     * - Returns: The field value or an empty String.
     */
    func getValue(field: Field) -> String {
        return self.fields[field]!
    }
}
