/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * A *DeviceEnumerator* can be used to enumerate devices matching a set of 
 * criteria. When iterating, it may generates REST calls to get the next chunk of 
 * devices.
 */
public class DeviceEnumerator: PageableEnumerator<Device> {
    
    /// The secure connection to use
    private let secureConnection: HttpSecureConnection
    
    
    /// The application identifier: namespace to use to look for devices
    private let appID: String
    
    /// Set of fields requested
    private let fields: [Device.Field.Values]
    
    /// Filter to apply to constrain the result
    private let filter: Filter
    
    /// Whether or not the results should include decommissioned devices
    private let includeDecommissioned: Bool
    
    /**
     * Instantiate a new iterator.
     *
     * - parameter appID: The application identifier if looking for devices of this application, otherwise `nil`.
     *
     * - parameter offset: The offset of the first item to fetch.
     *
     * - parameter limit: The maximum number of items requested per REST call.
     *
     * - parameter includeDecommissioned: Whether or not the results should include decommissioned devices.
     *
     * - parameter fields: Set of fields requested.
     *
     * - parameter filter: List of filters to apply to the result.
     *
     * - parameter secureConnection: The server connection.
     */
    init(appID: String, offset: Int, limit: Int, includeDecommissioned: Bool,
         fields: [Device.Field.Values], filter: Filter, secureConnection: HttpSecureConnection)  {
        
        self.appID = appID
        self.fields = fields
        self.filter = filter
        self.includeDecommissioned = includeDecommissioned
        self.secureConnection = secureConnection
        super.init(offs: offset, lim: limit)
    }
    
    /**
     * Iterate through the pages.
     *
     * - parameter callback: Callback containing the `Pageable<T>` If
     *                       an error occurs `page` is set to `nil` and
     *                       `error` indicates the error.
     */
    func iterate(callback: @escaping (Pageable<Device>?, ClientError?) -> Void)  {
        do {
            try self.getCount(countResponse: { countResponse, error in
                if (error != nil) {
                    callback(nil, error)
                } else {
                    if let countObject = countResponse {
                        let data = countObject.getData()
                        
                        if data.count == 0 {
                            self.setSize(size: -1)
                        }
                        else {
                            var jsonObject:JSON!
                            do {
                                jsonObject = try JSON(data: data as Foundation.Data)
                            } catch {
                                callback(nil, ClientError.network(String(format:NSLocalizedString(
                                    "%@: JSON threw exception: %@",
                                         comment:""),"DeviceEnumerator.iterate()", "\(error)")))
                                return
                            }
                            if let cnt = jsonObject.dictionaryValue["count"]?.intValue {
                                self.setSize(size: cnt)
                            }
                            else {
                                self.setSize(size: -1)
                            }
                        }
                        callback(self, nil)
                    } else {
                        callback(nil, ClientError.network(NSLocalizedString("DeviceEnumerator iterate() countResponse should not be nil", comment: "")))
                    }
                }
            })
        } catch {
            callback(nil, ClientError.network(NSLocalizedString(
                "Cannot retrieve device list", comment: "")))
        }
    }
    
    
    /**
     * Performs a new REST call to get the count of devices matching the filter
     *
     * - Returns: The count of devices matching the request, or *-1* if unknown.
     */
    private func getCount(countResponse: @escaping (HttpResponse?, ClientError?) -> Void) throws -> Void {
        let req = DeviceEnumerationRequest(applicationID: self.appID, path: "count", offset: self.voffset, limit: self.vlimit, includeDecommissioned: self.includeDecommissioned, fields: self.fields, filter: self.filter)
        try self.getSecureConnection().get(restApi: req.request(), headers: req.headers(), response: { getResponse, error in
            if (error != nil) {
                countResponse(nil, error)
            } else {
                if let getObject = getResponse {
                    countResponse(getObject, nil)
                } else {
                    countResponse(nil, ClientError.argument(NSLocalizedString("DeviceEnumerator getCount(): getResponse should not be nil", comment: "")))
                }
            }
        })
    }    
    
    override func load(offset: Int, limit: Int, callback: @escaping (PagedResponse<Device>?, ClientError?) -> Void) throws {
        let req = DeviceEnumerationRequest(applicationID: self.appID, path: "", offset: offset, limit: limit, includeDecommissioned: self.includeDecommissioned, fields: self.fields, filter: self.filter)
        try self.getSecureConnection().get(restApi: req.request(), headers: req.headers(), response: { getResponse, error in
            if (error != nil) {
                callback(nil, error)
            } else {
                if let getObject = getResponse {
                    let data = getObject.getData()
                    do {
                        let jsonObject = try JSON(data: data as Foundation.Data)
                        let devicePage = try DeviceEnumerationResponse.fromJson(jsonObject: jsonObject)
                        callback(devicePage, nil)
                    } catch {
                        callback(nil, ClientError.network(NSLocalizedString("Error while retrieving device page at offset " + offset.description, comment: "")))
                    }
                } else {
                    callback(nil, ClientError.argument(NSLocalizedString("DeviceEnumerator load(): getResponse should not be nil", comment: "")))
                }
            }
        })
    }
    
    
    func getSecureConnection() -> HttpSecureConnection {
        return self.secureConnection
    }
    
    /**
     * Creates a request for *GET iot/api/v2/devices* or *GET iot/api/v2/apps/{a-id}/devices*
     */
    class DeviceEnumerationRequest {
        
        /// List of headers required for the REST API

        private let vheaders = ["Content-Type" : "application/json", "Accept" : "application/json"]
        
        /// The REST request including parameters

        private var vrequest: String = ""
        
        /**
         * Creates a *DeviceEnumerationRequest* matching the specified criteria.
         *
         * - Parameter limit: The maximum number of result to return.
         *
         * - Parameter offset: The offset of the results in the entire list of devices matching the specified criteria.
         *
         * - Parameter includeDecommissioned: Whether or not the result should include decommissioned devices.
         *
         * - Parameter fields: Set of expected fields.
         *
         * - Parameter filter: Filter to be applied.
         */
        convenience init(offset: Int, limit: Int, includeDecommissioned: Bool, fields: [Device.Field.Values], filter: Filter) {
            let p: String = ""
            self.init(applicationID: p, path: p, offset: offset, limit: limit, includeDecommissioned: includeDecommissioned, fields: fields, filter: filter)
        }
        
        /**
         * Creates a *DeviceEnumerationRequest* matching the specified criteria.
         *
         * - Parameter applicationID: The application identifier in which we should look for devices.
         *
         * - Parameter limit: The maximum number of result to return.
         *
         * - Param offset: The offset of the results in the entire list of devices matching the specified criteria.
         *
         * - Parameter includeDecommissioned: Whether or not the result should include decommissioned devices.
         *
         * - Parameter fields: Set of expected fields.
         *
         * - Parameter filter: Filter to be applied.
         */
        convenience init(applicationID: String, offset: Int, limit: Int, includeDecommissioned: Bool, fields: [Device.Field.Values], filter: Filter) {
            self.init(applicationID: applicationID, path: String(""), offset: offset, limit: limit, includeDecommissioned: includeDecommissioned, fields: fields, filter: filter)
        }
        
        /**
         * Creates a *DeviceEnumerationRequest* matching the specified criteria *"/iot/api/v2/apps/{applicationId}/devices/{path}?offset={offset}&limit={limit}"*
         *
         * - Parameter applicationID: The application identifier in which we should look for devices.
         *
         * - Parameter path: The path to extend the request. If *null*, not added to the request.
         *
         * - Parameter limit: The maximum number of result to return.
         *
         * - Parameter offset: The offset of the results in the entire list of devices matching the specified criteria.
         *
         * - Parameter includeDecommissioned: Whether or not the result should include decommissioned devices.
         *
         * - Parameter fields: Set of expected fields.
         *
         * - Parameter filter: Filter to be applied.
         */
        init(applicationID: String, path: String, offset: Int, limit: Int, includeDecommissioned: Bool, fields: [Device.Field.Values], filter: Filter?) {
            var req: String = ""
            var sep: String = "?"
            
            if applicationID.isEmpty == true {
                req = RestApi.V2.getReqRoot()
            } else {
                req = RestApi.V2.getReqRoot()+"/apps/" + applicationID
            }
            
            req += "/devices"
            
            if path.isEmpty == false {
                req += "/" + path
            }
            
            if limit > 0 {
                req += sep + "limit=" + String(limit)
                sep = "&"
            }
            
            if offset > 0 {
                req += sep + "offset=" + String(offset)
                sep = "&"
            }
            
            req += sep + "includeDecommissioned=" + String(includeDecommissioned)
            req += "&totalResults=true"
            req += "&expand=location,metadata"
            
            // always ask for the type and id: required to instantiate a endpoint.
            var reqFields: [Device.Field.Values] = [] 
            if fields.count == 0 {
                reqFields = fields
                reqFields.append(Device.Field.Values.TYPE)
                reqFields.append(Device.Field.Values.ID)
            }
            else {
                reqFields = Device.Field.all()
            }
            
            if fields.count != 0 {
                req += "&fields=";
                req += joinFields(delimiter: ",", list: reqFields)
            }
            
            // build query
            if filter != nil {
                req += "&q="
                req += filter!.toJsonURI()
            }

            self.vrequest = req
        }
        
        
        private func joinFields(delimiter: String, list: [Device.Field.Values]) -> String {
            var sb: String = String()
            var first: Bool = true
            for f in list {
                if first == true {
                    first = false
                } else {
                    sb.append(delimiter)
                }
                sb.append(f.rawValue)
            }
            return sb
        }
        
        /**
         * Get the REST call (including parameters) representing this request
         *
         * - Returns: The REST call (including parameters)
         */
        func request() -> String {
            return self.vrequest
        }
        
        /**
         * Get the list of HTTP headers required for this request
         *
         * - Returns: HTTP headers required to perform this request
         */
        func headers() -> [String : String] {
            return vheaders
        }
    }
    
    /**
     * Response to *GET iot/api/v2/devices* or *GET iot/api/v2/apps/{a-id}/devices*.
     */
    class DeviceEnumerationResponse : PagedResponse<Device> {
    
        private init(devices: [DeviceImpl], offset: Int, total: Int, hasMore: Bool, links: [String : String]) {
            super.init(elements: devices, offset: offset, total: total, hasMore: hasMore, links: links)
        }
    
        /**
         * Creates a *DeviceEnumerationResponse* from a JSON object received in response to a *DeviceEnumerationRequest*
         *
         * - Parameter jsonObject: The JSON object to parse.
         *
         * - Returns: A *DeviceEnumerationResponse* instance.
         *
         * - Throws: JsonException if the jsonObject has not the expected format.
         *
         * - SeeAlso: DeviceEnumerationRequest.
         */
        static func fromJson(jsonObject: JSON) throws -> DeviceEnumerationResponse {
            let items = jsonObject.dictionaryValue["items"]?.arrayValue
            let links = jsonObject.dictionaryValue["links"]?.arrayValue
            var hasMore = jsonObject.dictionaryValue["hasMore"]?.boolValue
            if hasMore == nil {
                hasMore = false
            }
            
            var offset = jsonObject.dictionaryValue["offset"]?.intValue
            if offset == nil {
                offset = -1
            }
            
            var total = jsonObject.dictionaryValue["totalResults"]?.intValue
            if total == nil {
                total = -1
            }
            
            var list: [DeviceImpl] = [DeviceImpl]()
            if items != nil && items?.count != 0 {
                for item: JSON in items! {
                    let device = try DeviceImpl.from(object: item)
                    list.append(device)
                }
            }
            
            var map: [String : String] = [String : String]()
            if links != nil && links?.count != 0 {
                for value: JSON in links! {
                    let rel = value.dictionaryValue["rel"]?.stringValue
                    let href = value.dictionaryValue["href"]?.stringValue
                    if rel?.isEmpty == false && href?.isEmpty == false {
                        map[rel!] = href
                    }
                }
            }
            
            return DeviceEnumerationResponse(devices: list, offset: offset!, total: total!, hasMore: hasMore!, links: map)
        }
        
    }
    
}
