/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * Device implementation.
 */
class DeviceImpl : Device {
    
    private static let DCD = "DIRECTLY_CONNECTED_DEVICE";
    private static let ICD = "INDIRECTLY_CONNECTED_DEVICE";
    
    private static let LOCATION_LATITUDE = "latitude"
    private static let LOCATION_LONGITUDE = "longitude"
    private static let LOCATION_ALTITUDE = "altitude"
    private static let LOCATION_UNCERTAINTY = "uncertainty"
    
    private let fields: [Device.Field : String]
    private let visDirectlyConnected: Bool
    private let id: String
    private let deviceModels: [String]
    private let metadata: [String : String]
    private let location: Location?
    
    private init(id: String, fields: [Device.Field : String], deviceModels: [String], metadata: [String : String], location: Device.Location?, isDCD: Bool) {
        self.id = id
        self.fields = fields
        self.deviceModels = deviceModels
        self.metadata = metadata
        self.visDirectlyConnected = isDCD
        self.location = location
    }
    
    static func from(object: JSON) throws -> DeviceImpl {
        var fields:[Field : String] = [:]
        var entrySet = object.dictionaryValue
        
        for e in entrySet {
            let key = e.0
            let f = Field.fromAlias(alias: key)
            if f.alias().isEmpty == false {
                fields[f] = e.1.stringValue
            }
        }
        
        let t: String = fields[Field.fromAlias(alias: Field.Values.TYPE.rawValue)]!
        let isDCD: Bool = (t.isEmpty == false) && (t == DCD)
        
        let id: String = fields[Field.fromAlias(alias: Field.Values.ID.rawValue)]!
        
        // get device models
        let jsonModels = entrySet[Field.Values.DEVICE_MODELS.rawValue]?.arrayValue
        var deviceModels: [String] = []
        
        for i in 0..<jsonModels!.count {
                let urnObject = jsonModels![i]
                deviceModels.append(urnObject["urn"].stringValue)
        }
        
        // add device models in fields
        var fieldsDeviceModelString: String = "["
        for i in jsonModels! {
            let stringItem = try String(data: i.rawData(), encoding: .utf8)!
            fieldsDeviceModelString = fieldsDeviceModelString + stringItem + ","
        }
        fieldsDeviceModelString.remove(at: fieldsDeviceModelString.index(before: fieldsDeviceModelString.endIndex))
        fieldsDeviceModelString = fieldsDeviceModelString + "]"
        
        var f = Field.fromAlias(alias: Field.Values.DEVICE_MODELS.rawValue)
        fields[f] = fieldsDeviceModelString
        
        
        // get metadata
        let jsonMetadata = entrySet[Field.Values.METADATA.rawValue]
        let metaDictionary = jsonMetadata!.dictionaryValue
        var metadata: [String : String] = [:]
        
        for item in metaDictionary {
            let metaKey = item.0
            switch (item.1.type) {
            case .string:
                let metaValue = item.1.stringValue
                metadata[metaKey] = metaValue
                break
            
            case .array:
                var sb: String = "[{"
                let metaArray = item.1.arrayValue
                for i in 0..<metaArray.count {
                    let metaObject = metaArray[i].dictionaryValue
                    for j in metaObject {
                        sb = sb + "\"" + j.0 + "\""
                        sb = sb + ":" + "\"" + j.1.stringValue + "\","
                    }
                }
                sb.remove(at: sb.index(before: sb.endIndex))
                sb = sb + "}]"
                
                metadata[metaKey] = sb
                break
            default:
                metadata[metaKey] = ""
                break
            }
        }
        
        
        // add metadata in fields
        var fieldsMetadataString: String = "{"
        let stringItem1 = try String(data: (jsonMetadata?.rawData())!,
                                     encoding: .utf8)!
        fieldsMetadataString = fieldsMetadataString + stringItem1 + "}"

        
        f = Field.fromAlias(alias: Field.Values.METADATA.rawValue)
        fields[f] = fieldsMetadataString
        
        
        //get location
        var latitude: Float?
        var longitude: Float?
        var altitude: Float?
        var uncertainty: Float?
        
        let jsonLocation = entrySet[Field.Values.LOCATION.rawValue]
        if (jsonLocation != nil) {
            let locDictionary = jsonLocation!.dictionaryValue
            
            for item in locDictionary {
                let locKey = item.0
                let locValue = item.1.floatValue
                
                if (locKey == LOCATION_LATITUDE) {
                    latitude = locValue
                } else if (locKey == LOCATION_LONGITUDE) {
                    longitude = locValue
                } else if (locKey == LOCATION_ALTITUDE) {
                    altitude = locValue
                } else if (locKey == LOCATION_UNCERTAINTY) {
                    uncertainty = locValue
                }
            }
            
            // add location in fields
            let fieldsLocationString = try String(data: (jsonLocation?.rawData())!,
                                                  encoding: .utf8)!
            f = Field.fromAlias(alias: Field.Values.LOCATION.rawValue)
            fields[f] = fieldsLocationString
        }
        
        var location: Location?
        if (latitude != nil && longitude != nil) {
            location = Location(latitude: latitude!, longitude: longitude!, altitude: altitude, uncertainty: uncertainty)
        }
        
        return DeviceImpl(id: id, fields: fields, deviceModels: deviceModels, metadata: metadata, location: location, isDCD: isDCD)
    }
    
    /**
     *  Get the list of device models supported by this device.
     
     *  - returns: A list of device model URN.
     */
   
    override func getDeviceModels() -> [String] {
        return self.deviceModels
    }
    

    /**
     * Get the value of the specified *field* or an empty String if the value is unknown.
     
     * - parameter field: The field to look for
     
     * - returns: The field value or an empty String.
     */
    
    override func getValue(field: Device.Field) -> String {
        return fields[field]!
    }
    

    /**
     * Return whether the device is directly connected to the server or is connected through a gateway.
     
     * - returns: *true* if the device is directly connected, *false* otherwise.
     */
   
    override func isDirectlyConnected() -> Bool {
        return self.visDirectlyConnected
    }
    

    /**
     * Get the device identifier.
     
     * - returns: The device identifier.
     */
   
    override func getId() -> String {
        return self.id
    }
    

    /**
     * Return the device metadata. Might be empty.
     
     * - returns: A map of device metadata
     */
    
    override func getMetadata() -> [String : String] {
        return self.metadata
    }
    
    /**
     * Return the value of the metadata field *key*.
     
     * - parameter key: The metadata key to return.
     
     * - returns: Value for metadata *key* or *null* if not found.
     */
    
    override func getMetadata(key: String) -> String? {
        if self.metadata[key] == nil {
            return nil
        }
        
        return self.metadata[key]!
    }
    
    /**
     * Return the device's location.
     
     * - returns: The device's location.
     */
    
    override func getLocation() -> Device.Location? {
        return self.location
    }

}
