/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func < <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l < r
  case (nil, _?):
    return true
  default:
    return false
  }
}

// FIXME: comparison operators with optionals were removed from the 
// Swift Standard Libary.
// Consider refactoring the code to use the non-optional operators.
fileprivate func > <T : Comparable>(lhs: T?, rhs: T?) -> Bool {
  switch (lhs, rhs) {
  case let (l?, r?):
    return l > r
  default:
    return rhs < lhs
  }
}

/**
 * The `EnterpriseClient` is an enterprise application, client of the Oracle
 * IoT Cloud Service.
 */
public class EnterpriseClient : Client, Hashable {

    private let iotApp: Application
    private let secureConnection: HttpSecureConnection
    private var virtualDevices: Dictionary<String, Dictionary<DeviceModelImpl, VirtualDeviceImpl>>?
    private var lastUntil: Int64?
    private var bulkMonitorInProgress: Bool = false
    
    // Checked by all methods. If a method is called and isClosed is true
    // call the callback with the error or throw ClientError.State
    // depending on the method signature.
    private var isClosed: Bool = false
    
    var monitor: Monitor!
    
    // %@ - app id
    // %d - device id
    private let BULKDATA_RESOURCE_FORMAT: String = RestApi.V2.getReqRoot() +
                                    "/apps/%@/devices/data?formatLimit=%d"
    private let MONITOR_MAX_FORMATS = Config.MONITOR_MAX_FORMATS

    /**
     * Constructs a new `EnterpriseClient` instance. The specified context
     * (if any) could be used by platform specific implementation to retrieve 
     * information like internal storage area.
     *
     * - parameter iotApp: The IoT application.
     *
     * - parameter secureConnection: A reference to a secure connection object.
     */
    init(iotApp: Application, secureConnection: HttpSecureConnection) {
        self.iotApp = iotApp
        self.secureConnection = secureConnection
        self.virtualDevices = Dictionary<String, Dictionary<DeviceModelImpl, VirtualDeviceImpl>>()
        // Must occur before Monitor()
        SecureHttpConnectionMap.putSecureConnection (client: self,
                                    secureConnection: secureConnection)
        self.monitor = Monitor(ec:self)
    }
    
    /**
     * Creates an instance of `EnterpriseClient` associated with the specified
     * IoT application.
     *
     * The created instance will use a custom or default `TrustedAssetsManager`
     * to store, load and handle the trust material. The class name of a custom
     * `TrustedAssetsManager` to instantiate can be specified using the
     * "oracle.iot.client.tam" system property; the specified
     * `TrustedAssetsManager` class must have a public constructor accepting an
     * `Object` parameter.
     *
     * - parameter appName: The name of the IoT application to connect with.
     *                      If not `nil`, a REST call is performed to find the 
     *                      application-id matching with the specified name.
     *                      If `nil`, a REST call is performed to find the 
     *                      application that has, in its list of integrations 
     *                      identifier, the one retrieved from the
     *                      `TrustedAssetsManager`.
     *
     * - parameter path: the path to the provisioning file
     *
     * - parameter password: the password to decrypt the provisioning file
     *                       must be the same used during creation of the file
     *
     * - parameter callback: Callback for enterprise client object.
     *                      If creating client is not successful then
     *                      the object will be `nil` and an error object is 
     *                      passed in the form `callback(client, error)` and the
     *                      reason can be taken from error
     *
     * - seealso: `getApplication()`
     *
     */
    public static func newClient(appName: String, path: String,
                                 password: String,
                callback: @escaping (EnterpriseClient?, ClientError?) -> Void) {
        let tam: TrustedAssetsManager?
        do {
            tam = try TrustedAssetsManagerFactory.createTrustedAssetsManager(path: path,
                                                            password: password)!
        } catch ClientError.file(let err) {
            callback(nil, ClientError.file(err))
            return
        } catch  {
            callback(nil, ClientError.state(NSLocalizedString("\(error)", comment: "")))
            return
        }
        
        var secureHttpConnection: HttpSecureConnection
        
        // check if endpointId is null in truststore in order to use webapi 
        // (user auth) or api (endpoint auth)
        if tam!.getEndpointId() == "" {
            RestApi.USE_WEBAPI = true
            secureHttpConnection = HttpSecureConnectionUserAuth.createUserAuthSecureConnection(tam: tam!)
        }
        else {
            secureHttpConnection = HttpSecureConnection.createHttpSecureConnection(tam: tam!,
                                                    useOnlySharedSecret: true)
        }
        
        let filter: Filter
        if appName.isEmpty {
            filter = Filter.eq(op1: "integrations.id", op2: tam!.getEndpointId())
        }
        else {
            filter = Filter.eq(op1: "name", op2: appName)
        }
        
        let r = ApplicationEnumerator.ApplicationEnumerationRequest.init(applicationId: nil,
                            path: nil, filter: filter, offset: -1, limit: -1)
        do {
            try httpGet(secureConnection: secureHttpConnection,
                        headers: r.getHeaders(), request: r.getRequest(),
                        jsonObject: {(jsonGet, error) -> Void in
                if (error != nil) {
                    callback(nil, error)
                } else {
                    if let httpResponse = jsonGet {
                        // get application array
                        let apps = httpResponse["items"]
                        if apps.count != 1 {
                            var msg = ""
                            if appName.isEmpty {
                                msg = String(format: NSLocalizedString(
                                "Application with integration.id = %@ cannot be retrieved.",
                                        comment: ""), tam!.getEndpointId())
                            } else {
                                msg = String(format: NSLocalizedString(
                                "Application with name = %@ cannot be retrieved.",
                                        comment: ""), appName)
                            }
                            callback(nil, ClientError.network(msg))
                            return
                        }
                        
                        let app = Application.fromJson(object: apps[0])
                        
                        // sanity check
                        if appName.isEmpty == false {
                            if appName != app.getName() {
                                callback(nil, ClientError.network(String(format:
                                    NSLocalizedString(
                                        "Incorrect server response. Received [%@] instead of %@",
                                        comment: ""), app.getName(), appName)))
                               
                                return
                            }
                        }
                        
                        let client = EnterpriseClient(iotApp: app,
                                            secureConnection: secureHttpConnection)
                        callback(client, nil)
                        return
                    } else {
                        callback(nil, ClientError.argument(NSLocalizedString(
                            "EnterpriseClient newClient() httpResponse should not be nil",
                            comment: "")))
                        return
                    }
                }
            })
        } catch  {
            callback(nil, ClientError.state(NSLocalizedString("\(error)", comment: "")))
        }
    }
    
    static func httpGet(secureConnection: HttpSecureConnection,
                        headers: [String : String],
                        request: String,
                        jsonObject: @escaping (JSON?, ClientError?) -> ()) throws {

        try secureConnection.get(restApi: request, headers:
            headers, response: { getHttp, error in
            if (error != nil) {
                jsonObject(nil, error)
            } else {
                if let httpResponse = getHttp {
                    let status = httpResponse.getStatus()
                    if status != StatusCode.OK.getCode() {
                        let s = StatusCode.valueOf(code: status)
                        jsonObject(nil, ClientError.network(NSLocalizedString(
                            "Http get failed with \(s)", comment: "")))
                        return
                    }
                    
                    let dataResponse = httpResponse.getData()
                    if dataResponse.count == 0 {
                        jsonObject(nil, ClientError.network(NSLocalizedString(
                            "Http get failed: no data received.", comment: "")))
                        return
                    }
                    
                    var dataObject:JSON!
                    do {
                        dataObject = try JSON(data: dataResponse)
                    } catch {
                        jsonObject(nil, ClientError.argument(String(format:NSLocalizedString(
                            "EnterpriseClient.httpGet(): JSON threw exception: %@",
                            comment: ""), "EnterpriseClient.httpGet()", "\(error)")))
                        return
                    }
                    jsonObject(dataObject, nil)
                    return
                } else {
                    jsonObject(nil, ClientError.argument(NSLocalizedString(
                        "EnterpriseClient httpGet() httpResponse should not be nil",
                        comment: "")))
                }
            }
        })
    }
    
    /**
     * Get the all the applications that the user has access to.
     *
     * - parameter callback: Called with a `Pageable<Application>` containing
     *                       the applications associated with this 
     *                       `EnterpriseClient`.
     *                       If an error occurs `page` is set to `nil` and
     *                       `error` indicates the error.
     */
    func getApplications(callback: @escaping (Pageable<Application>?, ClientError?) -> Void) {
        if (isClosed) {
            callback(nil, ClientError.state(NSLocalizedString(
                "Client is closed", comment: "")))
            return
        }
        // TODO: Why arent's exceptions bubbled up through the callback ?
        ApplicationEnumerator(offset: 0, limit: 200, filter: nil,
                              secureConnection: self.secureConnection)
            .iterate( callback: { page, error in
                if (error != nil) {
                    callback(nil, ClientError.state(NSLocalizedString(
                        "Cannot create application iterator", comment: "")))
                } else {
                    do {
                        try page!.first(callback: { page, error in
                            callback(page, error)
                        })
                    } catch {
                        callback(nil, ClientError.network(NSLocalizedString(
                            "Cannot retrieve applications", comment: "")))
                    }
                }
            })
    }

    /**
     * Get the list of device models available for this application.
     * The callback is called with a `DeviceModelIterator`.
     *
     * - parameter callback: Called with a `Pageable<String>` containing the
     *                       device model urns associated with this
     *                       `EnterpriseClient`.
     *                       If an error occurs `page` is set to `nil` and
     *                       `error` indicates the error.
     */
    public func getDeviceModels(callback: @escaping (Pageable<String>?, ClientError?) -> Void) {
        if (isClosed) {
            callback(nil, ClientError.state(NSLocalizedString(
                "Client is closed", comment: "")))
            return
        }
        DeviceModelIterator(appId: self.iotApp.getId(), offset: 0, limit: 200,
                            secureConnection: self.secureConnection)
            .iterate( callback: { page, error in
                if (error != nil) {
                    callback(nil, ClientError.state(NSLocalizedString(
                        "Cannot create device model iterator", comment: "")))
                } else {
                    do {
                        try page!.first(callback: { page, error in
                            callback(page, error)
                        })
                    } catch {
                        callback(nil, ClientError.network(NSLocalizedString(
                            "Cannot retrieve device models", comment: "")))
                    }
                }
            })
    }
    
    /**
     * Get details on the IoT application for this client.
     *
     * - returns: Details on the specified application.
     */
    public func getApplication() -> Application {
        // NOTE: Should the application information be on demand in
        // order to provide up to the minute state ?
        
        // TODO verify 'isClosed' value in order to know if client has been 
        // closed
        
        return self.iotApp
    }
    
    /**
     * The `callback` is called with a `Pageable` collection of `Device`.
     * The fields that will be available from returned `Device` instances will
     * be limited to the one defined in `fields`. Use `Device.Field.Values` to
     * get them all.
     *
     * The `filter` forms a query. Only devices that satisfy the `filter` and
     * have a connectivity status of `status` are returned.
     *
     * - parameter fields: Set of fields `Device.Field` to return for the
     *                     selected devices. Can be `nil`.
     *
     * - parameter filter: A filter. Can be `nil`.
     *
     * - parameter callback: Callback containing the `Pageable<Device>` If
     *                       an error occurs `page` is set to `nil` and
     *                       `error` indicates the error.
     *
     *
     * - seealso: 
     *      - `Pageable`
     *      - `Device`
     */
    public func getDevices(fields: [Device.Field.Values],
                           filter: Filter,
                    callback: @escaping (Pageable<Device>?, ClientError?) -> Void) {
        if (isClosed) {
            callback(nil, ClientError.state(NSLocalizedString(
                "Client is closed", comment: "")))
            return
        }
        DeviceEnumerator(appID: self.iotApp.getId(), offset: 0, limit: 200,
                         includeDecommissioned: false, fields: fields,
                         filter: filter,
                         secureConnection: self.secureConnection)
            .iterate( callback: { page, error in
                do {
                    try page!.first(callback: { page, error in
                        callback(page, error)
                    })
                } catch {
                    callback(nil, ClientError.network(NSLocalizedString(
                        "Cannot retrieve device list.", comment: "")))
                }
            })
    }

    /**
     * Get the `DeviceModel` for the device model urn.
     * - parameter deviceModelUrn: The URN of the device model
     *
     * - parameter callback: Called with the device model in the activated 
     *                       state. If activation fails then `deviceModel`
     *                       will be `nil` and `error` will not be `nil` and
     *                       identify the error. If both `deviceModel` and
     *                       `error` are `nil` then the device model does not
     *                       exist.
     * - throws: `ClientError.state` if `close` has been called.
     */
    public func getDeviceModel(deviceModelUrn: String,
            callback: @escaping (DeviceModel?, ClientError?) -> Void) throws {
        if (isClosed) {
            throw ClientError.state(NSLocalizedString(
                "Client is closed", comment: ""))
        }
        // TODO: FIXME getDeviceModel does not really throw, exceptions occur in
        // nested completion handler.
        try DeviceModelFactory.getDeviceModel(aid: self.iotApp.getId(),
                secureConnection: secureConnection, urn: deviceModelUrn,
                callback: callback)
    }

    /**
     * Get the active devices implementing the specified device model.
     *
     * - parameter deviceModel: The URN of the device model.
     *
     * - parameter callback: Called with a `Pageable<Device>` containing
     *                       the devices associated with this 
     *                       `EnterpriseClient`.
     *                       If an error occurs `page` is set to `nil` and
     *                       `error` indicates the error.
     */
    public func getActiveDevices(deviceModel: String,
                callback: @escaping (Pageable<Device>?, ClientError?) -> Void) {
        if (isClosed) {
            callback(nil, ClientError.state(NSLocalizedString(
                "Client is closed", comment: "")))
            return
        }
        let filds = Device.Field.all()
                
        var filterArray = [Filter]()
        filterArray.append(Filter.eq(op1: Device.Field.Values.CONNECTIVITY_STATUS.rawValue, op2: "ONLINE"))
        filterArray.append(Filter.eq(op1: Device.Field.Values.STATE.rawValue, op2: "ACTIVATED"))
        filterArray.append(Filter.eq(op1: Device.Field.Values.DEVICE_MODELS.rawValue + ".urn", op2: deviceModel))
                
        let filtrs = Filter.and(filters: filterArray)
        self.getDevices(fields: filds, filter: filtrs, callback: { (devicePage, deviceError) in
            callback(devicePage, deviceError)
        })
    }
    
    /**
     * Closes the resources used by this client.
     *
     * - throws: `ClientError.network` if the `SecureConnection` cannot be
     *           closed
     */
    public func close() throws -> Void {
        if (!isClosed) {
            self.monitor.stop()
            for item in self.virtualDevices! {
                let dmDictionary = item.1
                for model in dmDictionary {
                    let virtDevice = model.1
                    self.removeVirtualDevice(virtualDevice: virtDevice)
                }
            }
            
            self.secureConnection.close()

            defer {
                isClosed = true
            }
        }
    }
    
    /**
     * Create a `VirtualDevice` instance with the given device model for the
     * given device identifier. This method creates a new `VirtualDevice`
     * instance for the given parameters. The client library does not cache 
     * previously created `VirtualDevice` objects.
     *
     * - parameter endpointId: The application client identifier of the device 
     *                         being modeled.
     *
     * - parameter deviceModel: The device model URN that this application 
     *                          implements.
     *
     * - returns: A new `VirtualDevice`.
     */
    public func createVirtualDevice(deviceId endpointId: String, deviceModel: DeviceModel) -> VirtualDevice {
        // TODO: verify 'isClosed' value in order to know if client has been closed
        // TODO: IOT-36343 This method must expose the async nature of the
        // call. Workaround implemented for IOT-36038
        // Like Java CL must not return an instance and if possible
        // throw exception early or return an error in the callback if
        // - monitor cannot be started
        // - no attributes are returned during initial bulk load if the
        // the device has attributes.
        let virtDev = VirtualDeviceImpl(enterpriseClient: self, secureConnection: secureConnection,
                                        endpointId: endpointId, model: deviceModel as! DeviceModelImpl)
        self.deviceMonitorInitialization(virtualDevice: virtDev, retryCnt: 0)
        return virtDev
    }
    
    // Note the format of the following JSON artifacts, the variable is typically
    // called "attributesDictionary"
    // Each entry is an attribute where
    //
    // <entry>.0 -> attribute name
    // <entry>.1 -> attribute value
    
    private func deviceMonitorInitialization(virtualDevice: VirtualDevice,
                                             retryCnt: Int) -> Void {
        let virtualDeviceImpl:VirtualDeviceImpl = virtualDevice as! VirtualDeviceImpl
        let deviceId = virtualDeviceImpl.getEndpointId()
        let urn = virtualDeviceImpl.getURN()
        
        let jsonString = "{\"\(deviceId)\":[\"\(urn)\"]}"
        
        let appId = self.iotApp.getId()
        let secConn = virtualDeviceImpl.getSecureConnection()
        
        let val = Double(MONITOR_MAX_FORMATS)
        let maxFormats = Int(val > 0 ? val : 10)
            
        let payload = jsonString.data(using: .utf8)
        
        // When will this be <= 0 ?
        if payload!.count <= 0 {
            return
        }

        var since = lastUntil
        if since == nil || since == 0 {
            since = TimeManager.currentTimeMillis() - Config.MESSAGE_POLLING_INTERVAL
        }

        let resource = String(format: BULKDATA_RESOURCE_FORMAT, appId, maxFormats) +
            "&formatSince=" + String(since!)
        
        do {
            try request(secureConnection: secConn, method: "POST",
                        resource: resource, payload: payload!,
                        callBackResponseMonitor: { response, error in
                if (error == nil) && (response != nil) {
                    // Used to delay monitor startup until all attributes are
                    // processed and the device callback has returned.
                    let monitorGroup:DispatchGroup = DispatchGroup()
                    let responseDictionary = response!.dictionaryValue
                    
                    virtualDeviceImpl.getEnterpriseClient().addVirtualDevice(virtualDevice: virtualDeviceImpl)
                    
                    if responseDictionary["data"]?.count > 0 {
                        let dataDictionary = responseDictionary["data"]!.dictionaryValue
                    
                        if dataDictionary.count > 0 &&
                                responseDictionary["until"]!.int64Value > 0 &&
                                    dataDictionary.first?.0 == deviceId {
                            let urnDictionary = dataDictionary[deviceId]
                    
                            if urnDictionary?.first!.0 == urn {
                                virtualDeviceImpl.getEnterpriseClient().setLastUntil(until: responseDictionary["until"]!.int64Value)
                                
                                let attributesDictionary = urnDictionary![urn].dictionaryValue["attributes"]
                                var root: NamedValue? = nil
                                var last: NamedValueImpl? = nil
                                
                                let attrGroup:DispatchGroup = DispatchGroup()
                                let attrQueue = DispatchQueue.global(qos:.utility)
                                
                                for eachAttribute in attributesDictionary! {
                                    let attributeName = eachAttribute.0
                                    let attributeValue = eachAttribute.1.rawValue
                                    let attributeList = virtualDeviceImpl.attributeMap
                                    
                                    // TODO: This process is different than SE
                                    // SE converts JSON attributes into VirtualDeviceAttributes
                                    // Why isn't that done here ? 
                                    // What happens when the onChange handler processes
                                    // the NamedValues ? They should be virtualAttribute types.
                                    for eachVirtualAttribute in attributeList {
                                        if eachVirtualAttribute.1.model.getName() != attributeName {
                                            continue
                                        }
                                        
                                        let attrType:DeviceModelAttributeType =
                                            eachVirtualAttribute.1.model.getType()
                                        let attrOnChange =
                                            eachVirtualAttribute.1.getOnChange()
                                        
                                        // Embed this function here to take advantage
                                        // of closure behavior to capture variable
                                        // values
                                        func processAttribute(value:AnyObject) {
                                            
                                            let _ = virtualDeviceImpl.initValue(attributeName: attributeName,
                                                                                value: value)
                                            
                                            let namedValue = NamedValueImpl(name: attributeName,
                                                                            value: value)
                                            if let attrOnChange = attrOnChange {
                                                let changeEvent: ChangeEventImpl =
                                                    ChangeEventImpl(virtualDevice: virtualDeviceImpl,
                                                                    value: namedValue)
                                                attrOnChange(changeEvent)
                                            }
                                            attrQueue.sync(execute: {
                                                if (last != nil) {
                                                    last!.setNext(next: namedValue)
                                                    last = namedValue
                                                } else {
                                                    root = namedValue
                                                    last = namedValue
                                                }
                                            })
                                        }
                                        
                                        switch attrType {
                                        case .URI:
                                            // Enter group in createStorageObject
                                            let uri:String = attributeValue as! String
                                            self.createStorageObject(
                                                    virtualDevice:virtualDeviceImpl,
                                                    uri:uri, fieldName:attributeName,
                                                    cbGroup:attrGroup,
                                                    callback: processAttribute)
                                            break
                                        default:
                                            // TODO: these enter and leaves are
                                            // not absolutely necessary
                                            attrGroup.enter()
                                            processAttribute(value: attributeValue as AnyObject)
                                            attrGroup.leave()
                                            break
                                        }
                                    }
                                }
                                if virtualDeviceImpl.base?.getOnChangeCallback() != nil && root != nil {
                                    let changeEvent: ChangeEventImpl = ChangeEventImpl(virtualDevice: virtualDeviceImpl, value: root!)
                                    // Hold up the monitor start until all attributes are processed
                                    monitorGroup.enter()
                                    attrGroup.notify(qos: .utility,
                                                     queue: attrQueue, // Should this be a different queue ?
                                                     execute: {
                                        virtualDeviceImpl.base?.getOnChangeCallback()!(changeEvent)
                                        monitorGroup.leave()
                                    })
                                }
                            }
                        }
                    } else {
                        print("Device monitor initialization: no data field in json response");
                    }
                    // Start monitoring after initial bulk load.
                    monitorGroup.notify(qos:.utility, queue: DispatchQueue.global(qos:.utility),
                                        execute: {
                        virtualDeviceImpl.getEnterpriseClient().startMonitoring()
                    })
                } else {
                    if (error != nil) {
                        print("Retry " + String(retryCnt) + ": " + error!.localizedDescription)
                    } else {
                        print("Retry " + String(retryCnt) + ": no response from device monitor initialization")
                    }
                    if (retryCnt < 3) {
                        self.deviceMonitorInitialization(virtualDevice: virtualDevice, retryCnt: (retryCnt + 1))
                    } else {
                        // TODO: error handling when retry fails
                        // TODO: forwarding http request error from Transport 
                        // class instead of nil value
                        print("deviceMonitorInitialization failed after 3 attempts")
                    }
                }
            })
        } catch {
            // TODO : error handling
            print(error)
        }
    }
    
    func setLastUntil(until: Int64) -> Void {
        self.lastUntil = until
    }
    
    func addVirtualDevice(virtualDevice: VirtualDeviceImpl) -> Void {
        self.removeVirtualDevice(virtualDevice: virtualDevice)
        let endpointId = virtualDevice.getEndpointId()
        
        if self.virtualDevices?.count > 0 && self.virtualDevices![endpointId] != nil { 
            if var urnDictionary = self.virtualDevices![endpointId] {
                urnDictionary[virtualDevice.getDeviceModel() as! DeviceModelImpl] = virtualDevice
                self.virtualDevices![endpointId] = urnDictionary
            }
        }
        else {
            self.virtualDevices![endpointId] = [ virtualDevice.getDeviceModel() as! DeviceModelImpl : virtualDevice ]
        }
    }
    
    private func startMonitoring() {
        /*
        if self.virtualDevices?.count == 1 {
            self.monitor.start()
        }
        */
        // Always call start monitoring, if this call hinders performance
        // the qos of the monitors "syncQueue" can be increased.
        // Use a function variable for the "start" and replace it with
        // a "noop" the first time the monitor is started.
        // The problem with the above logic is if the count goes to 2
        // before the first call to startMonitoring the monitor is never
        // started
        self.monitor.start()
    }
    
    func removeVirtualDevice(virtualDevice: VirtualDeviceImpl) -> Void {
        let deviceId = virtualDevice.getEndpointId()
        var vdDictionary = self.virtualDevices![deviceId]
        
        if ((vdDictionary != nil) && (vdDictionary?.count > 0)) {
            vdDictionary!.removeValue(forKey: virtualDevice.getDeviceModel() as! DeviceModelImpl)
        }
        if vdDictionary?.count == 0 {
            let _ = self.virtualDevices?.removeValue(forKey: deviceId)
        }
        
        if self.virtualDevices?.count == 0 {
            self.monitor.stop()
        }
    }
    
    /*
     * Convert the json response into VirtualDeviceAttributes and notify
     * the respective devices of the attribute changes. The bulk json data
     * has the following format.
     *
     *   {
     *       "data":{
     *           "0-AI":{
     *               "urn:com:oracle:iot:device:thermometer":{
     *                   "attributes":{
     *                       "scale":1.0,
     *                       "temperature":20.935437036685517
     *                   }
     *                   "formats": {
     *                       "format1" : [{
     *                           "eventTime": 1234,
     *                            "severity": "HIGH",
     *                            "fields": {
     *                                 "name1": "value1",
     *                                 "name2": "value2"
     *                            },...,
     *                            {...}
     *                       ]
     *                   }
     *               }, .... ,
     *               "urn:com:oracle:iot:device:YY":{
     *                   "attributes":{
     *                       "M": W,
     *                       "N": Z
     *                   }
     *               }
     *           }, ... ,
     *           "0-NN":{
     *               "urn:com:oracle:iot:device:XX":{
     *                   "attributes":{
     *                       "X":J,
     *                       "Y":K
     *                   }
     *                   "formats": {...}
     *               }
     *           }
     *       }
     *   }
     */

    func remoteBulkMonitor() {
        if self.bulkMonitorInProgress == true {
            return
        }
        
        self.bulkMonitorInProgress = true
        
        if self.virtualDevices?.count == 0 {
            self.bulkMonitorInProgress = false
            return
        }
        
        let appId = self.iotApp.getId()
        let secConn = self.secureConnection
        
        let val = Double(self.MONITOR_MAX_FORMATS)
        let maxFormats = Int(val > 0 ? val : 10)
        
        var jsonString = "{"
        
        var devices = self.virtualDevices
        
        for d in devices! {
            jsonString = jsonString + "\"\(d.0)\":["
            let deviceModelDictionary = d.1
            for dm in deviceModelDictionary {
                let urn = dm.0.getURN()
                jsonString = jsonString + "\"\(urn)\","
            }
            
            jsonString = String(jsonString[jsonString.startIndex..<jsonString.index(before: jsonString.endIndex)])
            jsonString = jsonString + "],"
        }
        
        jsonString = String(jsonString[jsonString.startIndex..<jsonString.index(before: jsonString.endIndex)])
        jsonString = jsonString + "}"
        
        let payload = jsonString.data(using: .utf8)
        
        if payload!.count > 0 {

            var resource = String(format: BULKDATA_RESOURCE_FORMAT, appId, maxFormats)
            let since = lastUntil
            if ((since != nil) && (since != 0)) {
                resource += "&since=" + String(since!)
            }
            
            do {
                try self.request(secureConnection: secConn, method: "POST", resource: resource,
                                 payload: payload!, callBackResponseMonitor: { response, error in
                    if (error != nil) {
                        // TODO error logging/handling
                        print("remoteBulkMonitor() error: " + error!.localizedDescription)
                        self.bulkMonitorInProgress = false
                    } else {
                        if let responseObj = response {
                            self.bulkMonitorInProgress = false
                            
                            let responseDictionary = responseObj.dictionaryValue
                            
                            if (responseDictionary["data"] == nil || responseDictionary["until"] == nil) {
                                //TODO: error
                                return
                            }
                            let dataDictionary = responseDictionary["data"]!.dictionaryValue
                            let untilDictionary = responseDictionary["until"]!.int64Value
                            
                            if dataDictionary.count == 0 && untilDictionary <= 0 {
                                // throw error
                                return
                            }
                            
                            self.setLastUntil(until: responseDictionary["until"]!.int64Value)
                            for deviceId in dataDictionary {
                                if let dmDictionary = devices![deviceId.0] {
                                    for dm in dmDictionary {
                                        self.processMonitorData(data: dataDictionary, deviceId: deviceId.0, virtualDevice: dm.1)
                                    }
                                }
                            
                            }
                        }
                    }
                })
            } catch {
                // TODO error logging/handling
                print("Exception during remoteBulkMonitor()")
            }
        }
        else {
            self.bulkMonitorInProgress = false
        }
    }

    func processMonitorData(data: [String : JSON], deviceId: String, virtualDevice: VirtualDeviceImpl) {
        
        let urn = virtualDevice.getURN()
        
        var root: NamedValue? = nil
        var last: NamedValueImpl? = nil

        var attributesDictionary: [String : JSON]?
        var formatsDictionary: [String : JSON]?
        
        var isEmpty : Bool = true
        
        if let urnDictionary = data[deviceId]?.dictionaryValue[urn] {
            if urnDictionary.count > 0 {
                attributesDictionary = urnDictionary["attributes"].dictionaryValue
                formatsDictionary = urnDictionary["formats"].dictionaryValue
                isEmpty = false
            }
        }
        
        if isEmpty {
            return
        }
        
        let attrQueue:DispatchQueue = DispatchQueue.global(qos:.utility)

        let attrGroup:DispatchGroup = DispatchGroup()
        for attribute in attributesDictionary! {
            //var namedValue: NamedValueImpl?
            let attributeName = attribute.0
            var attributeValue:JSON = attribute.1
            
            var vdAttribute:VirtualDeviceAttribute? = nil
            do {
                vdAttribute = try virtualDevice.getAttribute(name: attributeName)
            } catch {
                print(String(format: NSLocalizedString(
                    "Cannot get attribute %@. %@", comment: ""),
                    attributeName, "\(error)"))
                continue
            }
            
            let attTypeList = (virtualDevice.getDeviceModel() as! DeviceModelImpl).getDeviceModelAttributes()
            let updated = vdAttribute?.remoteUpdate(attributeName: attributeName,
                                                   attribute: attributeValue,
                                                   model: attTypeList)
            // Embed this function here to take advantage
            // of closure behavior to capture variable values
            func processAttribute(value:AnyObject) {
                
                let namedValue = NamedValueImpl(name: attributeName,
                                                value: value)
                if vdAttribute?.getOnChange() != nil {
                    let changeEvent: ChangeEventImpl =
                        ChangeEventImpl(virtualDevice: virtualDevice,
                                        value: namedValue)
                    vdAttribute?.getOnChange()!(changeEvent)
                }
                attrQueue.sync(execute: {
                    if (last != nil) {
                        last!.setNext(next: namedValue)
                        last = namedValue
                    } else {
                        root = namedValue
                        last = namedValue
                    }
                })
            }
            
            if updated == true {
                let value:AnyObject = attributeValue.rawValue as AnyObject
                let attrType:DeviceModelAttributeType =
                    (vdAttribute!.getDeviceModelAttribute()?.getType())!
                switch attrType {
                case .URI:
                    let uri:String = attributeValue.stringValue
                    //value =
                    self.createStorageObject(
                        virtualDevice:virtualDevice,
                        uri:uri, fieldName:attributeName,
                        //onchange: vdAttribute?.getOnChange(),
                        cbGroup: attrGroup,
                        callback: processAttribute)
                    break
                default:
                    // TODO: These enter and leave are not absolutely necessary
                    attrGroup.enter()
                    processAttribute(value: value)
                    attrGroup.leave()
                    break
                }
            }
        }
        
        if ((root != nil) && (virtualDevice.base?.getOnChangeCallback() != nil)) {
            let changeEvent: ChangeEventImpl = ChangeEventImpl(virtualDevice: virtualDevice, value: root!)
            attrGroup.notify(qos: .utility,
                             queue: attrQueue, // Should this be a different queue ?
                             execute: {
                virtualDevice.base?.getOnChangeCallback()!(changeEvent)
            })
        }

        let noMsgCallbacks:Bool = virtualDevice.getOnAlertCallback() == nil &&
            virtualDevice.getOnDataCallback() == nil
        
        for format in formatsDictionary! {
            let formatName = format.0
            var vdFormat:VirtualDeviceFormat? = nil
            
            do {
                vdFormat = try virtualDevice.getFormat(formatName: formatName)
            } catch {
                print(String(format:
                    NSLocalizedString("Cannot get format %@. %@", comment: ""),
                    formatName, "\(error)"))
                continue
            }
            
            // Skip if there are no handlers
            if noMsgCallbacks && vdFormat!.getOnAlert() == nil &&
                    vdFormat!.getOnData() == nil {
                continue
            }
            let formatGroup:DispatchGroup = DispatchGroup()
            let formatQueue:DispatchQueue = DispatchQueue.global(qos:.utility)
            let formatArray = format.1.arrayValue
            for item in formatArray {
                
                let formatURN = vdFormat?.dmFormats?.getURN()
                let formatType:DeviceModelFormatType = (vdFormat?.dmFormats?.getType())!
                let vdFormatFields:[DeviceModelFormatField] = (vdFormat?.dmFormats?.getFields())!
                
                let fieldsDictionary = item.dictionaryValue["fields"]
                let msgTime = item.dictionaryValue["eventTime"]
                let timeValue = TimeInterval(Double(msgTime!.stringValue)! / 1000)
                let time = Date(timeIntervalSince1970: timeValue)
                
                var alertMsgList: NamedValue? = nil
                var lastAlertMsgList: NamedValueImpl? = nil
                var dataMsgList: NamedValue? = nil
                var lastDataMsgList: NamedValueImpl? = nil
                

                for msgField in fieldsDictionary! {
                    let fieldName = msgField.0
                    var fieldValue:AnyObject = msgField.1.rawValue as AnyObject
                    
                    // Embed this function here to take advantage
                    // of closure behavior to capture variable values
                    func processFormat(value:AnyObject) {
                        let namedValue = NamedValueImpl(name: fieldName, value: value)
                        switch formatType {
                        case .ALERT:
                            if let onAlert = vdFormat?.getOnAlert() {
                                let alertEvent: AlertEvent = AlertEventImpl(
                                    virtualDevice: virtualDevice, urn: formatURN!,
                                    value: namedValue, eventTime: time)
                                onAlert(alertEvent)
                            }
                            formatQueue.sync(execute: {
                                if alertMsgList == nil {
                                    alertMsgList = namedValue
                                    lastAlertMsgList = namedValue
                                } else {
                                    lastAlertMsgList!.setNext(next: namedValue)
                                    lastAlertMsgList = namedValue
                                }
                            })
                            break
                        case .DATA:
                            if let onData = vdFormat?.getOnData() {
                                let dataEvent: DataEvent = DataEventImpl(
                                    virtualDevice: virtualDevice, urn: formatURN!,
                                    value: namedValue, eventTime: time)
                                onData(dataEvent)
                            }
                            formatQueue.sync(execute: {
                                if dataMsgList == nil {
                                    dataMsgList = namedValue
                                    lastDataMsgList = namedValue
                                } else {
                                    lastDataMsgList!.setNext(next: namedValue)
                                    lastDataMsgList = namedValue
                                }
                            })
                            break
                        }
                    }

                    for field in vdFormatFields {
                        if field.getName() != fieldName {
                            continue
                        }
                        switch field.getType() {
                        case .URI:
                            let uri:String = fieldValue as! String
                            self.createStorageObject(
                                virtualDevice:virtualDevice,
                                uri:uri, fieldName:fieldName,
                                cbGroup: formatGroup,
                                callback: processFormat)
                            break
                        default:
                            formatGroup.enter()
                            processFormat(value: fieldValue)
                            formatGroup.leave()
                            break
                        }
                    }
                }
                
                formatGroup.notify(qos:.utility,
                                   queue:formatQueue, //Should this be a different queue ?
                                   execute: {
                    if alertMsgList != nil &&
                            virtualDevice.getOnAlertCallback() != nil {
                        let alertEvent: AlertEvent = AlertEventImpl(
                            virtualDevice: virtualDevice, urn: formatURN!,
                            value: alertMsgList!, eventTime: time)
                        
                        if virtualDevice.getOnAlertCallback() != nil {
                            virtualDevice.getOnAlertCallback()!(alertEvent)
                        }
                    }

                    if dataMsgList != nil &&
                            virtualDevice.getOnDataCallback() != nil {
                        let dataEvent: DataEvent = DataEventImpl(
                            virtualDevice: virtualDevice, urn: formatURN!,
                            value: dataMsgList!, eventTime: time)
                        
                        if virtualDevice.getOnDataCallback() != nil {
                            virtualDevice.getOnDataCallback()!(dataEvent)
                        }
                    }
                })
            }
        }
    }
    
    /**
     * Make an http request to the cloud service.
     *
     * - parameter secureConnection: The secure connection to the cloud service.
     *
     * - parameter method: The http method one of `GET` or `POST`.
     *
     * - parameter resource: The REST resource including query parameters.
     *
     * - parameter payload: The request payload, can be `nil`.
     *
     * - parameter callBackResponseMonitor: Callback for REST call response as 
     *             json object.
     */
    private func request(secureConnection: HttpSecureConnection, method: String,
            resource: String, payload: Foundation.Data,
            callBackResponseMonitor: @escaping (JSON?, ClientError?) -> Void) throws -> Void {
        
        let vheaders = ["Content-Type" : "application/json",
                        "Accept" : "application/json"]
        
        if method == "GET" {
            try secureConnection.get(restApi: resource, headers: vheaders,
                                     response:  { jsonHttpResponse, error in
                if (error != nil) {
                    callBackResponseMonitor(nil, error)
                } else {
                    if let jsonObj = jsonHttpResponse {
                        let jsonData = jsonObj.getData()
                        var json:JSON!
                        do {
                            json = try JSON(data: jsonData)
                        } catch {
                            callBackResponseMonitor(nil, ClientError.network(String(format:
                                NSLocalizedString(
                                    "%@: JSON threw exception: %@",
                                    comment: ""),"EnterpriseClient.request()", "\(error)")))
                            return
                        }
                        callBackResponseMonitor(json, nil)
                    } else {
                        callBackResponseMonitor(nil, ClientError.argument(NSLocalizedString("EnterpriseClient request() jsonHttpResponse should not be nil", comment: "")))
                    }
                }
            })
        } else if method == "POST" {
            try secureConnection.post(restApi: resource, payload: payload, headers: vheaders, response:  { jsonHttpResponse, error in
                if (error != nil) {
                    callBackResponseMonitor(nil, error)
                } else {
                    if let jsonObj = jsonHttpResponse {
                        let jsonData = jsonObj.getData()
                        var json:JSON!
                        do {
                            json = try JSON(data: jsonData)
                        } catch {
                            callBackResponseMonitor(nil, ClientError.argument(String(format:NSLocalizedString(
                                "%@: JSON threw exception: %@", comment:""),
                                "EnterpriseClient.request()", "\(error)")))
                            return
                        }
                        callBackResponseMonitor(json, nil)
                    } else {
                        callBackResponseMonitor(nil, ClientError.argument(NSLocalizedString("EnterpriseClient request() jsonHttpResponse should not be nil", comment: "")))
                    }
                }
            })
        }
    }

    /// Override Hashable
    public var hashValue: Int {
        return self.iotApp.getId().hashValue
    }

    /**
     * Create a new `StorageObject` that will have a name with the given
     * object name prefixed with the application's ID and a directory
     * separator. The prefix addition can be disabled by setting the
     * `Config.DISABLE_STORAGE_OBJECT_PREFIX` to `true`.
     *
     * If `contentType` is nil, the mime-type defaults to
     * "application/octet-stream".
     *
     * - parameter name: the unique name to be used to reference the content in
     *             storage
     * 
     * - parameter contentType: The mime-type of the content or `nil`
     * 
     * - returns: a `StorageObject`
     *
     * - throws:
     *        - `ClientError.network` if there is an exception raised by the
     *                                runtime, or an abnormal response from the 
     *                                storage cloud
     *        - `ClientError.security` if there is an exception establishing
     *                                 a secure connection to the storage cloud
     */
    public func createStorageObject(name:String,
                                contentType:String? = OCTET_STREAM) throws -> StorageObject {
        let storageConnection:StorageConnectionBase =
            StorageConnectionMap.getStorageConnection(secureConnection:secureConnection)
        let delegate:AdvancedStorageObject =
            try storageConnection.createStorageObject(
                clientId:secureConnection.getEndpointId() ?? "",
                                    name:name, contentType:contentType)
        return StorageObjectImpl(delegate:delegate)

    }
    
    /**
     * Create a new `StorageObject` from the `uri` for a named object in storage.
     *
     * - parameter uri: the URI of the object in the storage cloud
     *
     * - parameter callback: This api is makes an asynchronous network call.
     *                       `callback` is invoked when the network call
     *                       completes.
     *
     * - returns: a `StorageObject` for the object in storage
     *
     * - throws: `ClientError.network` if there is an exception raised by the runtime,
     *                                or a failure reported by the storage cloud
     *      `ClientError.security` if there is an exception establishing a secure 
     *                             connection to the storage cloud
     */
    public func createStorageObject(uri:String,
                callback: @escaping (StorageObject?,ClientError?)->()) throws {
        // TODO: Return to public access when asynchronous implementation is in place
        let storageConnection:StorageConnectionBase =
            StorageConnectionMap.getStorageConnection(secureConnection:secureConnection)
        do {
            //let delegate:AdvancedStorageObject = try storageConnection.createStorageObject(storageUrl:uri)!
            //callback(StorageObjectImpl(delegate:delegate), nil)
            try storageConnection.createStorageObject(storageUrl:uri,
                    callback: {(storageObject, error) in
                if let error = error {
                    callback(nil, error)
                } else {
                    callback(StorageObjectImpl(delegate:storageObject!), nil)
                }
            })
        } catch {
            throw error
        }
    }
    
    // For use by monitor
    private func createStorageObject(virtualDevice:VirtualDeviceImpl,
                                     uri:String, fieldName:String,
                                     cbGroup:DispatchGroup,
                                     callback: @escaping (AnyObject)->()) {
        // Enter the group to delay the device onChange until
        // createStorageObject completes
        cbGroup.enter()
        if StorageConnectionBase.isStorageCloudURI(uri:uri) {
            do {
                try createStorageObject(uri:uri, callback:{(storageObject, error) in
                    defer {
                        cbGroup.leave()
                    }
                    if let error = error {
                        //If SCS object in inaccessable, log it and create
                        // ExternalObject value instead.
                        print(String(format:NSLocalizedString(
                            "Storage CS object access failed: %@", comment:""),
                                     "\(error)"))
                        callback(ExternalObject(uri: uri) as AnyObject)
                        return
                    }
                    (storageObject as! StorageObjectImpl).setSyncEventInfo(virtualDevice:virtualDevice,
                                                                           nameForSyncEvent: fieldName)
                    callback(storageObject as AnyObject)
                })
            } catch {
                let storageObject:AnyObject = ExternalObject(uri: uri) as AnyObject
                callback(storageObject)
                cbGroup.leave()
            }
        } else {
            let storageObject:AnyObject = ExternalObject(uri: uri) as AnyObject
            callback(storageObject)
            cbGroup.leave()
        }
        
    }

    /*
     * Private only
     */
    fileprivate class StorageConnectionMap  {
        // TODO: Is it ok to use endpoint id to uniquiely map to a
        // secureConnection ? It is problematic to make SecureConnection
        // Hashable.
        
        private static
        var STORAGE_CONNECTION_MAP:Dictionary<String,StorageConnectionImpl> =
            Dictionary<String,StorageConnectionImpl>();
        
        fileprivate class func getStorageConnection(secureConnection:SecureConnection)
                -> StorageConnectionImpl {
            var storageConnection = STORAGE_CONNECTION_MAP[secureConnection.getEndpointId()!]
            if storageConnection == nil {
                storageConnection = StorageConnectionImpl(secureConnection:secureConnection)
                STORAGE_CONNECTION_MAP[secureConnection.getEndpointId()!] = storageConnection
            }
            return storageConnection!
        }
    }
}
