/*
 *  Copyright © 2016 Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * The *Response* protocol is used to retrieve the result of a request on a *Resource*.
 */
public protocol Response {
    
    /**
     * Returns the HTTP status code
     *
     * - Returns: The status code.
     */
    func getStatusCode() -> Int
    
    /**
     * Returns the message id of the http request that this is a response for.
     *
     * - Returns: The request id.
     */
    func getRequestId() -> String
    
    /**
     * Get request message Path.
     *
     * - Returns: Path, never *nil*.
     */
    func getPath() -> String
    
    /**
     * Get HTTP response message body. If message body contains string value, default encoding is UTF-8.
     *
     * - Returns: Body, never *nil*.
     *
     * - Throws: ClientError.State if the response is not complete.
     *
     * - SeeAlso: isDone()
     */
    func getBody() throws -> Foundation.Data
    
    /**
     * - Returns: The Response headers.
     *
     * - Throws: ClientError.State if the response is not complete.
     *
     * - SeeAlso: isDone()
     */
    func getHeaders() throws -> [String : String]
    
    /**
     * Call *isDone* to determine if the response is complete. Resource requests to the cloud service can be asynchronous. Before obtaining the body *isDone* must be called and return *true* before calling *getBody*, *getBodyString* or *getHeaders*. If *isDone()* returns *false* the application must call *poll()* to try to update the response.
     *
     * - Returns: *true* if the response is complete, *false* otherwise.
     */
    func isDone() -> Bool
    
    /**
     * This method is used to look for an available response. It should typically be called when *isDone()* has returned *false*.
     *
     * - Throws: IOException if network error occurs.
     *
     * - Throws: GeneralSecurityException when key or signature algorithm class cannot be loaded, or the key is not in the trusted assets store, or the private key is invalid.
     */
    func poll() throws -> Void
    
}
