/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * Implements *Response*.
 */
class ResponseImpl : Response {
    
    // request data: immutable
    
    /// Request identifier
    private let requestId: String?
    
    /// Request path
    private let path: String
    
    /// Request URI used to re-emit the request. Might be null is processed 
    /// synchronously
    private let uri: String

    /// Request headers used to re-emit the request. Might be null if response 
    /// is processed synchronously
    private let requestHeaders: [String : String]
    
    // response data: mutable
    
    /// Response body
    private var body: Data?

    /// Response headers
    private var responseHeaders: [String : String]

    /// Secure connection
    private var secureConnection: HttpSecureConnection
    

    /// Response status
    private var statusCode: Int?
    

    /// `true` if request is done.
    private var done: Bool
    
    private init(requestId: String, path: String, body: Foundation.Data?,
                 status: Int, requestURI: String?,
                 secureConnection: HttpSecureConnection,
                 reqHeaders: [String : String],
                 respHeaders: [String : String], done: Bool) {
        // TODO: look for URI instead of String
        self.requestId = requestId
        self.path = path
        self.uri = requestURI!
        self.requestHeaders = reqHeaders
        self.secureConnection = secureConnection
        self.body = body
        self.statusCode = status
        self.responseHeaders = respHeaders
        self.done = done
    }
    
    /**
     * Create a `ResponseImpl` instance from the specified data.
     *
     * - Parameter path: The request path.
     *
     * - Parameter reqHeaders: The request http headers.
     *
     * - Parameter res: The http response received.
     *
     * - Parameter secureConnection: The secure connection to use.
     *
     * - Returns: A `ResponseImpl` instance.
     *
     * - throws: ClientError.network if response `data` is empty, there are
     *           no `links`, the `requestURI` link is empty, status is not 200
     *           or 202
     */
    static func from(path: String, reqHeaders: [String : String], res: HttpResponse,
                     secureConnection: HttpSecureConnection) throws -> ResponseImpl {

        let data = res.getData()
        
        if data.count == 0 {
            let errorCause = "No data received."
            print(errorCause)
            throw ClientError.network(NSLocalizedString(errorCause, comment:""))
        }
        
        let jsonBody = try JSON(data: data as Foundation.Data)
        let requestId = jsonBody["id"].stringValue
        
        switch res.getStatus() {
            case 200:
                // Request is done
                var respHeaders: [String : String] = [String : String]()
                let headers = res.getHeaders()
                
                if headers.count != 0 {
                    for header in headers {
                        let values = header.1
                        for v in values {
                            respHeaders[header.0] = v
                        }
                    }
                }
                
                let resp: ResponseImpl = ResponseImpl(requestId: requestId,
                    path: path, body: nil, status: StatusCode.OK.getCode(),
                    requestURI: nil, secureConnection: secureConnection,
                    reqHeaders: reqHeaders, respHeaders: respHeaders, done: true)
                
                resp.updateFrom(httpResponse: res.getData() as Foundation.Data)
                return resp;
            
            case 202:
                // Request is not yet done
                let links = jsonBody["links"].arrayValue
                if links.count == 0 {
                    let errorCause = "No links in response."
                    print(errorCause)
                    throw ClientError.network(NSLocalizedString(errorCause, comment: ""))
                }
            
                var requestURI: String = ""
                for i in 0..<links.count {
                    let rel = links[i].dictionaryValue["rel"]
                    if rel == "requests" {
                        requestURI = (links[i].dictionaryValue["href"]?.stringValue)!
                        break
                    }
                }
                if requestURI.isEmpty == true {
                    let errorCause = "Incorrect response: missing request link."
                    print(errorCause)
                    throw ClientError.network(NSLocalizedString(errorCause, comment: ""))
                }
                
                return ResponseImpl(requestId: requestId, path: path,
                                    body: res.getData() as Foundation.Data,
                                    status: StatusCode.ACCEPTED.getCode(),
                                    requestURI: requestURI,
                                    secureConnection: secureConnection,
                                    reqHeaders: reqHeaders,
                                    respHeaders: [:], done: false)
            
        default:
            print("GET " + path + ":" + String(res.getStatus()))
            throw ClientError.network(String(format: NSLocalizedString("GET %@: $%@", comment: ""),
                                             path, String(res.getStatus())))
        }
    }
    
    
    private func updateFrom(httpResponse: Foundation.Data) -> Void {
        var jsonObject:JSON!
        do {
            jsonObject = try JSON(data: httpResponse)
        } catch {
            print(String(format:NSLocalizedString(
                "%@: JSON threw exception: %@",
                    comment:""), "ResponseImpl.updateFrom()", "\(error)"))
            self.statusCode = StatusCode.BAD_REQUEST.getCode()
            self.done = true
            return
        }
        
        let respObject = jsonObject["response"].dictionaryValue
        let reqStatus: String = jsonObject["status"].stringValue
        
        if reqStatus == "COMPLETED" {
            
            self.statusCode = respObject["statusCode"]!.intValue
            let bodyData = respObject["body"]!.stringValue
            self.body = bodyData.data(using: .utf8)
            
            let jsonHeaders = respObject["headers"]?.dictionaryValue

            self.responseHeaders = [String : String]()
            
            for header in jsonHeaders! {
                self.responseHeaders[header.0] = header.1.stringValue
            }
            
            self.done = true
            return
        }
        
        if reqStatus == "FAILED" {
            self.statusCode = StatusCode.BAD_REQUEST.getCode()
            self.done = true
        }
    }
    
    
    
    func getStatusCode() -> Int {
        return self.statusCode!
    }
    

    func getRequestId() -> String {
        return self.requestId!
    }
    

    func getPath() -> String {
        return self.path
    }
    

    func getBody() throws -> Foundation.Data {
        return self.body!
    }
    

    func getHeaders() throws -> [String : String] {
        return self.responseHeaders
    }
    

    func isDone() -> Bool {
        return self.done
    }
    

    func poll() throws -> Void {
        if self.done {
            return
        }
        
        try self.secureConnection.get(restApi: self.uri,
                                      headers: self.requestHeaders,
                                      response: { (httpResponse, error) -> Void in
            if (error != nil) {
                print(error!.localizedDescription)
            } else {
                if let jsonObj = httpResponse {
                    let status = jsonObj.getStatus()
                    if status != StatusCode.ACCEPTED.getCode() &&
                        status != StatusCode.OK.getCode() {
                        self.statusCode = status
                        self.done = true
                        return
                    }
                    
                    let data = jsonObj.getData()
                    self.updateFrom(httpResponse: data)
                } else {
                    print("ResponseImpl poll() httpResponse should not be nil")
                }
            }
        })
    }
}
