/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * VirtualDeviceAttribute is an attribute in the device model.
 */
class VirtualDeviceAttribute: VirtualDeviceAttributeBase {

    var lastUpdate: Date?
    var localUpdateRequest: Bool?
    var error: Bool?
    
    /**
     * Set the value of an attribute. This method is used by the application to
     * synchronize the *VirtualDevice* state with the physical device. This 
     * results in a REST API call to the server. For an enterprise client, an 
     * endpoint resource REST API call is made to the server. For a device 
     * client, an endpoint DATA message REST API call is made to the server.
     * The value is validated according to the constraints in the device model. 
     * If the value is not valid, an IllegalArgumentException is raised.
     *
     * - parameter value: The value to set.
     */
    override func set(value: AnyObject) -> Void {
        do {
            try validate(attribute: model, value: value)
            self.value = value
        } catch { }
    }
    
    /**
     * Check if the attribute can be set. On the device-client, an attribute 
     * should always be settable. On the enterprise-client, an attribute is 
     * settable if the model says the attribute is writable.
     *
     * - returns: *true* if the attribute can be set.
     */
    override func isSettable() -> Bool {
        return self.model.getAccess() !=
            .READ_ONLY
    }
    
    override init(model: DeviceModelAttribute) {
        super.init(model: model)
    }
    
    convenience init(model: DeviceModelAttribute, value: AnyObject) {
        self.init(model: model)
        
        do {
            try validate(attribute: model, value: value)
        } catch {
            self.error = true
        }
        
        self.value = value
        self.lastKnownValue = value
        self.localUpdateRequest = false
        
    }
    
    func remoteUpdate(attributeName: String, attribute: JSON,
                      model: [String : DeviceModelAttribute]) -> Bool {
        
        var attType: DeviceModelAttributeType?
        var rangeLow: Double?
        var rangeHigh: Double?
        var attAccess: DeviceModelAttribute.Access?
        
        for item in model {
            if item.0 == attributeName {
                attType = item.1.getType()
                attAccess = item.1.getAccess()
                rangeLow = Double(truncating: item.1.getLowerBound()!)
                rangeHigh = Double(truncating: item.1.getUpperBound()!)
            }
        }
        
        switch(attType!) {
            case .NUMBER:
                let doubleRange = Double(attribute.stringValue)
                let validRange = checkRange(value: doubleRange!,
                                            rangeLow: rangeLow,
                                            rangeHigh: rangeHigh)
                
                if  validRange == false {
                    self.error = true
                    return false
                }
        
                break
            case .INTEGER:
                let intRange = Double(attribute.stringValue)
                let validRange = checkRange(value: intRange!,
                                            rangeLow: rangeLow,
                                            rangeHigh: rangeHigh)
                
                if  validRange == false {
                    self.error = true
                    return false
                }
            
                break
            case .BOOLEAN, .STRING, .DATETIME, .URI:
            break
        }
        
        
        self.lastUpdate = Date()
        
        let newValueObject = attribute.rawValue
        if self.lastKnownValue != nil {
            let areEqual = equalValues(newValue: newValueObject as AnyObject,
                                       oldValue: self.lastKnownValue!,
                                       type: attType!)
            if  areEqual == true {
                return false
            }
        }

        self.lastKnownValue = newValueObject as AnyObject?
        
        
        if self.localUpdateRequest == true &&
                attAccess == .READ_WRITE {
            return false
        }
        else {
            self.value = newValueObject as AnyObject?
        }
        
        return true
    }
    
    func checkRange(value: Double, rangeLow: Double?,
                    rangeHigh: Double?) -> Bool {
        if let low = rangeLow {
            if let high = rangeHigh {
                if value < low || value > high {
                    return false
                }
            }
        }
        
        return true
    }
    
    func equalValues(newValue: AnyObject, oldValue: AnyObject,
                     type: DeviceModelAttributeType) -> Bool {
        
        switch type {
        case .DATETIME:
            if newValue is Date && oldValue is Date {
                let nv = newValue as! Date
                let ov = oldValue as! Date
                if nv == ov {
                    return true
                }
            }
            else {
                if newValue.stringValue == oldValue.stringValue {
                    return true
                }
            }
            break
            
        case .STRING:
            let nv = newValue as! String
            let ov = oldValue as! String
            if nv == ov {
                return true
            }
            break
            
        case .INTEGER:
            let nv = newValue as! NSInteger
            let ov = oldValue as! NSInteger
            if nv == ov {
                return true
            }
            break
            
        case .NUMBER:
            let nv = newValue as! NSNumber
            let ov = oldValue as! NSNumber
            if nv == ov {
                return true
            }
            break
            
        case .BOOLEAN:
            let nv = newValue as! Bool
            let ov = oldValue as! Bool
            if (nv == true && ov == true) || (nv == false && ov == false) {
                return true
            }
            break
        case .URI:
            if newValue is ExternalObject && oldValue is ExternalObject {
                let oe = (oldValue as! ExternalObject)
                let ne = (newValue as! ExternalObject)
                return ne.getURI() == oe.getURI()
            }
            break
        }
        return false
    }
}
