/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * VirtualDeviceFormat is an attribute in the device model.
 */
class VirtualDeviceFormat {
    
    var dmFormats: DeviceModelFormat?
    
    var value: [String : AnyObject] = [:]

    var lastKnownValue: [String : AnyObject] = [:]
    
    var lastUpdate: Date?
    var localUpdateRequest: Bool?
    
    var onAlert: ((AlertEvent) -> Void)?
    var onData: ((DataEvent) -> Void)?
    
    init(dmFormats: DeviceModelFormat) {
        self.dmFormats = dmFormats
    }
    
    convenience init(model: DeviceModelFormat, value: AnyObject) {
        self.init(dmFormats: model)
        
        self.localUpdateRequest = false

    }
    
    /**
     * Get the callback for an alert.
     
     * - returns: *callback*, which may be *nil*.
     */
    
    func getOnAlert() -> ((AlertEvent) -> Void)? {
        return self.onAlert
    }
    
    /**
     * Set a callback that is invoked when the value of an alert changes on the remote virtual device.
     *
     * - parameter callback: A callback to invoke when an alert value changes.
     */
    
    func setOnAlert(callback: @escaping (AlertEvent) -> Void) -> Void {
        self.onAlert = callback
    }
    
    /**
     * Get the callback for a custom data message.
     
     * - returns: *callback*, which may be *nil*.
     */
    
    func getOnData() -> ((DataEvent) -> Void)? {
        return self.onData
    }
    
    /**
     * Set a callback that is invoked when the value of a custom data message changes on the remote virtual device.
     *
     * - parameter callback: A callback to invoke when a custom data message value changes.
     */
    
    func setOnData(callback: @escaping (DataEvent) -> Void) -> Void {
        self.onData = callback
    }
    
    
    func get() -> [String : AnyObject] {
        return self.value
    }
    
    func remoteFormatUpdate(formatName: String, format: JSON) -> Bool {
        
        self.lastUpdate = Date()
        
        let newValueObject = format.rawValue
        
        self.lastKnownValue[formatName] = newValueObject as AnyObject?
        
        if self.value[formatName]?.stringValue == (newValueObject as AnyObject).stringValue {
            return false
        }
        
        // update the format value
        self.value[formatName] = newValueObject as AnyObject?
        return true
        
    }
    
}
