/*
 * Copyright (c) 2017, Oracle and/or its affiliates.  All rights reserved.
 *
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

//package com.oracle.iot.client

import Foundation

// TODO: Do subclasses inherit hash and "==" overrides ?
/*
@Override
public boolean equals(Object obj) {
    if (this == obj) return true;
    if(obj == null) return false;
    if (this.getClass() != obj.getClass()) return false;
    StorageObject other = (StorageObject)obj;
    if (!this.getURI().equals(other.getURI())) return false;
    if (!this.getName().equals(other.getName())) return false;
    if (!this.getType().equals(other.getType())) return false;
    return true;
}

@Override
public int hashCode() {
    int hash = 5;
    hash = 61 * hash + this.name.hashCode();
    hash = 61 * hash + this.type.hashCode();
    hash = 61 * hash + this.getURI().hashCode();
    return hash;
}
*/

/**
 * StorageObject provides information about content in cloud storage.
 */
public class AdvancedStorageObject: AdvancedExternalObject {

    private var name:String!
    private var contentType:String!
    private var encoding:String?
    private var date:Date?
    private var length:Int64 = -1
    private var inputStream:InputStream?
    private var outputStream:OutputStream?

    init(uri:String, name:String, contentType:String?, encoding:String?,
         date:Date?, length:Int64) {
        super.init(uri:uri)
        self.name = name
        self.contentType = contentType ?? OCTET_STREAM
        self.encoding = encoding
        self.date = date
        self.length = length
    }

    /**
     * Get the the name of this object in the storage cloud. This is name and
     * path of the file that was uploaded to the storage cloud.
     * - returns: the name of this object in the storage cloud
     */
    public func getName() -> String {
        return name
    }

    /**
     * Get the mime-type of the content. See [ANA Media Types](
     * http://www.iana.org/assignments/media-types/media-types.xhtml)
     *
     * - returns: The mime-type of the content
     */
    public func getContentType() -> String {
        return contentType
    }

    /**
     * Get the compression scheme of the content.
     * - returns: the compression scheme of the content, or `nil` if the
     *            content is not compressed
     */
    public func getEncoding() -> String? {
        return encoding
    }

    /**
     * Get the length of the content in bytes. This is the number of bytes 
     * required to upload or download the content.
     *
     * - returns: The length of the content, or -1 if unknown
     */
    public func getLength() -> Int64 {
        return length
    }

    // TODO: Should these methods return String to conform to IOS 8601 ?
    // Use ISO8601DateFormatter string(Date) date(String)
    /**
     * Get the date and time the content was created or last modified in cloud 
     * storage. This may be `nil` if the content has not been uploaded.
     * The date and time stamp format is `ISO 8601`.
     *
     * - returns: The date the content was last modified in cloud storage, or
     *            `nil` if the content has not been uploaded.
     */
    public func getDate() -> Date? {
        return date
    }

    /**
     * Called by the library to obtain the input stream when uploading content
     *
     * - returns: The input stream for the content being uploaded
     */
    public func getInputStream() -> InputStream? {
        return inputStream
    }

    /**
     * Set an input stream for content to be uploaded. The implementation
     * allows for either the input stream to be set, or the output stream to be 
     * set, but not both. If `inputStream` is not `nil`,
     * the output stream will be set to `nil.
     *
     * - parameter inputStream: A stream from which the content will be read
     *
     * - throws: `ClientError.state` if a transfer is currently in progress
     */
    public func setInputStream(inputStream:InputStream?) {
        self.inputStream = inputStream
        if inputStream != nil {
            self.outputStream = nil
        }
    }

    /**
     * Called by the library to obtain the output stream when downloading
     * content
     *
     * - returns: The output stream for the content being downloaded
     */
    public func getOutputStream() -> OutputStream? {
        return outputStream
    }

    /**
     * Set an output stream for content to be downloaded. The implementation
     * allows for either the input stream to be set, or the output stream to be 
     * set, but not both. If `outputStream` is not `nil`, the input stream will 
     * be set to `nil`.
     *
     * - parameter outputStream A stream to which the content will be written
     *
     * - throws: `ClientError.state` if a transfer is currently in progress
     */
    public func setOutputStream(outputStream:OutputStream?) {
        self.outputStream = outputStream
        if outputStream != nil {
            self.inputStream = nil
        }
    }
    /**
     * Used internally by the client library to update fields after upload.
     * Not intended for general use.
     * - parameter date: the last-modified date for the object in the Storage
     *                    Cloud
     * - parameter length: length the length for the object in the Storage 
     *                      Cloud
     */
    func setMetadata(date:Date, length:Int64) {
        self.date = date
        self.length = length
    }
    
    /**
     * Synchronize content with the Storage Cloud Service.
     *
     * - parameter callback: `callback` is called when the network activity is
     *                       complete.
     *
     * - throws: `ClientError.network` if there is an raised by the runtime,
     *                                 or a failure reported by the storage cloud
     *      `ClientError.security` if there is an exception establishing a 
     *                             secure connection to the storage cloud
     *      `ClientError.argument` if the `StorageObject` does not return an 
     *                             `InputStream` or `OutputStream`
     */
    public func sync(callback: @escaping (ClientError?) -> ()) throws {
        fatalError("Subclass need to implement the 'sync()' method.")
    }
}
