/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

public let OCTET_STREAM = "application/octet-stream"

/**
 * Client of the Oracle IoT Cloud Service. A client is a directly-connected
 * device, a gateway, or an enterprise application.
 *
 */
public protocol Client {

    /**
     * Create an `AbstractVirtualDevice` instance with the given device model
     * for the given device identifier. This method creates a new `VirtualDevice`
     * instance for the given parameters. The client library does not
     * cache previously created VirtualDevice objects.
     *
     * - parameter deviceId: The device identifier of the device being modeled
     *
     * - parameter deviceModel: The device model URN that this device implements
     *
     * - throws: `ClientError.network` if a `MessageDispatcher` cannot be 
     *           obtained.
     *
     * - returns: a new `VirtualDevice`
     */
    func createVirtualDevice(deviceId: String,
                             deviceModel: DeviceModel) throws -> VirtualDevice
    
    /**
     * Get the `DeviceModel` for the device model URN. This method may
     * return `nil` if there is no device model for the URN. nil may also be
     * returned if the device model is a "draft" and the property
     * `com.oracle.iot.client.device.allow_draft_device_models` is set to
     * `false`, which is the default.
     *
     * - parameter deviceModelUrn: The URN of the device model
     *
     * - parameter callback: This function is called with the retrieved device
     *                       model. If retrieving the device model is not 
     *                       successful then the object will be null and an 
     *                       error object is passed in the form 
     *                       `callback(device, error)` and the reason can be 
     *                       taken from error
     *
     * On error `callback` receives a `ClientError.network` error if:
     *     the object is not found (`StatusCode.NOT_FOUND`),
     *     the request fails (`!StatusCode.OK`), or
     *     the returned data is `nil` or JSON parsing fails
     */
    func getDeviceModel(deviceModelUrn: String,
            callback: @escaping (DeviceModel?, ClientError?) -> Void) throws

    /**
     * Release the resources held by the client. This call is idempotent
     *
     * - throws: `ClientError.network` if a `SecureConnection` cannot be closed.
     */
     func close() throws

    /**
     * Create a new `StorageObject` with the given object name and mime-type.
     * If `contentType` is `nil`, the mime-type defaults to "application/octet-stream".
     *
     * - parameter name: the unique name to be used to reference the content in storage
     *
     * - parameter contentType: The mime-type of the content
     *
     * - returns: a `StorageObject`
     *
     * - throws: 
     *       - `ClientError.network` if there is an exception raised by the runtime,
     *                                or an abnormal response from the storage cloud
     *       - `ClientError.security` if there is an exception establishing a 
     *                                secure connection to the storage cloud
     */
    func createStorageObject(name:String, contentType:String?) throws -> StorageObject
}
