/*
 *  Copyright © 2016 Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

// TODO: This should be in "Settings" format so app can use the "Settings"
// api to display configuration to user.
/**
 * Class containing properties, constants and error codes
 * as static vars / lets.
 * The properties are the pendant to Java SE's system properties.
 */
class Config {
    //------ USER CONFIGURABLE SYSTEM PROPERTIES: Enterprise -------
    
    /**
     * Default: 3000 milliseconds
     *
     * The amount of delay, in milliseconds, between polls for messages by the MessagePoller.
     * This setting affects how often a MessageEnumerator.MessageListener instance may be called.
     */
    static let MESSAGE_POLLING_INTERVAL: Int64 = 3000
    
    /**
     * Default: 10
     *
     * The number of alerts each request the enterprise client performs for virtual device updates can have.
     */
    static let MONITOR_MAX_FORMATS: Int = 10
    
    
    //------ USER CONFIGURABLE SYSTEM PROPERTIES: Device -------
    
    /**
     * Default: 4192 bytes
     *
     * The total size all of incoming messages received from the server that have not been
     * handled by callbacks to the application.
     */
    static let REQUEST_BUFFER_SIZE: Int = 4192
    
    /**
     * Default: false
     *
     * Long polling is used by default, set this to true to disable it.
     */
    static let DISABLE_LONG_POLLING: Bool = false
    
    /**
     * Default: 600
     *
     * Long polling timeout value (on server side, default value is 540 seconds)
     */
    static let LONG_POLLING_TIMEOUT: Double = 600
    
    /**
     * Default: 10000 milliseconds
     *
     * Only used when long polling is disabled. The minimum time, in milliseconds, since the
     * last send before polling the server for incoming messages if the receive buffer is empty.
     * A larger value may cause a delay in the receipt of incoming messages, but will cause less
     * network overhead. A value of zero will cause the library to skip the poll for incoming
     * messages when calling receive if the receive buffer is empty.
     */
    static let SEND_RECEIVE_TIMEOUT: Int64 = 10000
    
    /**
     * Default: 2
     *
     * The number of attempts to upload/download or get access token
     */
    static let STORAGE_CONNECTION_RETRY_LIMIT: Int = 2
    /**
     * Default: "storage.oraclecloud.com"
     *
     * The default storage connection host name
     */
    static let STORAGE_CONNECTION_HOST_NAME: String = "storage.oraclecloud.com"
    /**
     * Default: 1MB
     *
     *The number of bytes to chunk when streamin a downloaded file to the
     * StorageObject outputstream
     */
    static let STORAGE_CONNECTION_CHUNK_SIZE: Int = 1024 * 1024
    // Should this be a percentage of the total ?
    /** 
     * Default: 1MB
     *
     * Number of bytes transferred between progress updates
     */
    static let STORAGE_CONNECTION_PROGRESS_UPDATE_INTERVAL:Int64 = 1024*1024

    /**
     * Default: "1"
     *
     * The current software version
     */
    static let VERSION: String? = "1"

    /**
     * Default: "127.1.0.0"
     *
     * The default IP address
     */
    static let IP_ADDRESS: String = "127.1.0.0"

    /**
     * Default: "Unknown"
     *
     * The default MAC address
     */
    static let MAC_ADDRESS: String? = "Unknown"
    
    
    //------ USER CONFIGURABLE SYSTEM PROPERTIES: Shared -------

    /**
     * Default: false
     *
     * If true, a device can be activated without a device model on the server for the device.
     * The server will create a draft device model which contains no attributes. A draft device
     * model allows devices to be brought on-line ahead of the server being provisioned with the
     * device model. However, attempts by the device to set an attribute when using a draft
     * device model will fail.
     */
    static let ALLOW_DRAFT_DEVICE_MODELS: Bool = true

    /**
     * Default: 15000 milliseconds
     *
     * Only used when long polling is disabled. The amount of time, in milliseconds, to wait for
     * the server to respond to an HTTPS request.
     */
    static let HTTP_RESPONSE_TIMEOUT = 15000
    
    /**
     * The library will use a prefix comprised of the endpoint ID and a
     * path separator, "<endpointId>/<storage object name>" when creating
     * a `StorageObject`. Set the `DISABLE_STORAGE_OBJECT_PREFIX` to `true`
     * to disable this feature and the `StorageObject` name will be the value
     * of the name given when creating the `StorageObject`
     */
    static let DISABLE_STORAGE_OBJECT_PREFIX:Bool = false
    
    
    //------ SYSTEM PROPERTIES: NOT USER CONFIGURABLE -------

    /**
     * Defines maximum length of string keys in JSON
     */
    static let MAX_KEY_LENGTH = 2048

    /**
     * Error domain
     */
    static let ERROR_DOMAIN: String = "CLErrorDomain"

    /**
     * Error codes
     */
    static let ERROR_DCD_NOT_ACTIVATED: Int = 103
    static let ERROR_REQUEST_HANDLER_NIL: Int = 104
    static let ERROR_REQUEST_MESSAGE_NIL: Int = 105
    static let ERROR_WRONG_MESSAGE_TYPE: Int = 106
    static let ERROR_WRONG_FORMAT: Int = 107
    static let ERROR_WRONG_SOURCE: Int = 108
    static let ERROR_WRONG_SOURCE_DESTINATION: Int = 109
    static let ERROR_ILLEGAL_PROPERTY: Int = 110
    static let ERROR_ILLEGAL_RELIABILITY: Int = 111
    static let ERROR_WRONG_EVENT_TIME: Int = 112
    static let ERROR_ILLEGAL_DIRECTION: Int = 113
    static let ERROR_WRONG_RECEIVED_TIME: Int = 114
    static let ERROR_WRONG_SENT_TIME: Int = 115
    static let ERROR_ILLEGAL_DATA_ITEM_KEY: Int = 116
    static let ERROR_ILLEGAL_DATA_ITEM_VALUE: Int = 117
    static let ERROR_MESSAGE_PROPERTY_KEY: Int = 118
    static let ERROR_MESSAGE_PROPERTY_VALUE: Int = 119
    static let ERROR_ILLEGAL_REQUEST: Int = 120
    static let ERROR_ILLEGAL_REQUEST_MESSAGE_HEADER: Int = 121
    static let ERROR_ILLEGAL_REQUEST_MESSAGE_PARAMETER: Int = 122
    static let ERROR_ILLEGAL_REQUEST_MESSAGE_CLIENT: Int = 123
    static let ERROR_ILLEGAL_REQUEST_MESSAGE_PAYLOAD: Int = 124
    static let ERROR_ILLEGAL_REQUEST_MESSAGE_METHOD: Int = 125
    static let ERROR_ILLEGAL_REQUEST_MESSAGE_URL: Int = 126
    static let ERROR_ILLEGAL_RESPONSE_MESSAGE_HEADER: Int = 128
    static let ERROR_ILLEGAL_RESPONSE_MESSAGE_CONTENT_TYPE: Int = 129
    static let ERROR_ILLEGAL_RESPONSE_MESSAGE_PAYLOAD: Int = 130
    static let ERROR_ILLEGAL_RESPONSE_MESSAGE_REQUEST_ID: Int = 131
    static let ERROR_ILLEGAL_RESOURCE_MESSAGE: Int = 132
    static let ERROR_ILLEGAL_RESOURCE_NAME: Int = 133
    static let ERROR_ILLEGAL_RESOURCE_METHODS: Int = 134
    static let ERROR_DEVICE_MODEL_ISSUE: Int = 136
    static let ERROR_VIRTUAL_DEVICE_ISSUE: Int = 137
    static let ERROR_HTTP_ERROR: Int = 138
    static let ERROR_SECURITY_ISSUE: Int = 139
    static let ERROR_ILLEGAL_RESOURCES: Int = 144
    static let ERROR_JSON: Int = 145
    static let ERROR_ILLEGAL_ATTRIBUTE_VALUE = 146
    static let ERROR_UNSUPPORTED_OPERATION: Int = 147
    static let ERROR_NO_DATA: Int = 148

    /**
     * Count of POSTs
     */
    static var postCounter: Int = 0
}


