/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * An enumeration of the data types an attribute may contain.
 */
public enum DeviceModelAttributeType: String {
    /** The type of the attribute is a Number. */
    case NUMBER
    /** The type of the attribute is a String. */
    case STRING
    /** The type of the attribute is a Boolean */
    case BOOLEAN
    /**  The type of the attribute is an Integer. */
    case INTEGER
    /**
     * The type of the attribute is a Long, which represents
     * the millisecond time since the epoch.
     */
    case DATETIME
    /** The type of the attribute is a URI. */
    case URI
}

// Do we want this too in addition to the one defined in the prototype file ?
/**
 * Implementation of protocol Equatable
 * /
func ==(lhs: DeviceModelAttributeImpl, rhs: DeviceModelAttributeImpl) -> Bool {
    if (lhs === rhs) {
        return true
    }
    if (lhs.getModel() == rhs.getModel()) && (lhs.getName() == rhs.getName()) {
        return true
    }
    return false
}
*/

/**
 * Implementation of protocol Equatable
 *
 * parameter lhs: one item to be compared
 *
 * parameter rhs: the other item to be compared
 *
 * - returns: *true* if equal, *false* otherwise
 */
public func ==(lhs: DeviceModelAttribute, rhs: DeviceModelAttribute) -> Bool {
    if lhs === rhs {
        return true
    }
    if !(lhs.getName() == rhs.getName())  {
        return false
    }
    if !(lhs.getModel() == rhs.getModel())  {
        return false
    }
    if !(lhs.getDescription() == rhs.getDescription())  {
        return false
    }
    if !(lhs.getType() == rhs.getType())  {
        return false
    }
    if !(lhs.getLowerBound() == rhs.getLowerBound())  {
        return false
    }
    if !(lhs.getUpperBound() == rhs.getUpperBound())  {
        return false
    }
    if !(lhs.getAlias() == rhs.getAlias())  {
        return false
    }
    return (lhs.getAccess() == rhs.getAccess())
}

/**
 * DeviceModelAttribute is the model of an attribute in a {@link DeviceModel}
 */
public class DeviceModelAttribute: Hashable {
    
    internal let name: String?
    internal let urn: String?
    internal let description: String?
    internal let type: DeviceModelAttributeType
    internal let lowerBound: NSNumber?
    internal let upperBound: NSNumber?
    internal let access: Access
    internal let alias: String?
    internal let defaultValue: AnyObject?
 
    /**
     * Access rules for an attribute.
     */
    public enum Access: String {
        /**
         * The attribute value can be read but may not be set.
         */
        case READ_ONLY
        /**
         * The attribute value can be set but may not be read.
         */
        case WRITE_ONLY
        /**
         * The attribute value can be set and can be read
         */
        case READ_WRITE
        /**
         * The attribute is a action
         */
        case EXECUTABLE
    }

    public convenience init(urn: String, name: String, description: String?, type: DeviceModelAttributeType,
                            lowerBound: NSNumber?, upperBound: NSNumber?, access: Access, alias: String?) {
        self.init(urn: urn, name: name, description: description, type: type, lowerBound: lowerBound, upperBound: upperBound, access: access, alias: alias, defaultValue: nil)
    }
    
    public init(urn: String, name: String, description: String?, type: DeviceModelAttributeType,
                lowerBound: NSNumber?, upperBound: NSNumber?, access: Access, alias: String?, defaultValue: AnyObject?) {
        self.urn = urn
        self.name = name
        self.description = description
        self.type = type
        self.lowerBound = lowerBound
        self.upperBound = upperBound
        self.access = access
        self.alias = alias
        self.defaultValue = defaultValue
    }
    
    // TODO: Why did this implementation override the other ?
    /**
     * Implementation of protocol Hashable
     * /
    override var hashValue : Int {
        var hash: Int = 37;
        hash = 37 &* hash &+ (getModel() != nil ? getModel()!.hashValue : 0);
        hash = 37 &* hash &+ (getName()  != nil ? getName()!.hashValue  : 0);
        return hash;
    }
    */
    /**
     * Implementation of protocol Hashable
     *
     * - returns: the hash value
     */
    public var hashValue : Int {     // <--- create overflow
        var result: Int = (getName()?.hashValue)!
        if getModel() != nil {
            result = result &* 31 &+ (getModel()?.hashValue)!
        }
        result = result &* 31 &+ (getDescription()!.hashValue)
        result = result &* 31 &+ (getType().hashValue)
        result = result &* 31 &+ (getLowerBound()!.hashValue)
        result = result &* 31 &+ (getUpperBound()!.hashValue)
        
        if getAlias() != nil {
            result = result &* 31 &+ (getAlias()!.hashValue)
        }
        
        result = result &* 31 &+ (getAccess().hashValue)
        return result
    }
    
    /**
     * Get the attribute name.
     *
     * - returns: the attribute name from the device model
     */
    public func getName() -> String? {
        return self.name!
    }
    
    /**
     * Get the URN of the device type model this attribute comes from.
     *
     * - returns: the URN of the device type model
     */
    public func getModel() -> String? {
        return self.urn!
    }
    
    /**
     * A human friendly description of the attribute. If the model does not
     * contain a description, this method will return an empty String.
     *
     * - returns: the attribute description, or an empty string
     */
    public func getDescription() -> String? {
        return self.description
    }
    
    /**
     * The data type of the attribute. If the access type of the attribute is
     * executable, this method will return nil.
     *
     * - returns: the attribute's data type, or nil
     */
    public func getType() -> DeviceModelAttributeType {
        return self.type
    }
    
    // TODO: Why did these override the others ?
    /*
    override func getLowerBound() -> NSNumber {
        return self.lowerBound!
    }
    
    override func getUpperBound() -> NSNumber {
        return self.upperBound!
    }
    */

    /**
     * For *Type.NUMBER* only, give the lower bound of the
     * acceptable range of values. nil is always returned for attributes
     * other than *.NUMBER* type.
     *
     * - returns: a Number, or nil if no lower bound has been set
     */
    public func getLowerBound() -> NSNumber? {
        return self.lowerBound
    }
    
    /**
     * For *Type.NUMBER* only, give the upper bound of the
     * acceptable range of values. nil is always returned for attributes
     * other than *.NUMBER* type.
     *
     * - returns: a Number, or nil if no upper bound has been set
     */
    public func getUpperBound() -> NSNumber? {
        return self.upperBound
    }
    
    /**
     * An alternative name for the attribute. This optional attribute provides a
     * more descriptive name for the attribute. For example, "temperature" might
     * be an alias for the attribute name "5700". The alias can be used where ever
     * the attribute name would be used in the *AbstractVirtualDevice* API.
     *
     * - returns: The attribute alias.
     */
    public func getAlias() -> String? {
        return self.alias
    }
    
    /**
     * Return the access rules for the attribute. The default is READ-ONLY
     *
     * - returns: the access rule for the attribute
     */
    public func getAccess() -> DeviceModelAttribute.Access {
        return self.access
    }
    
    /**
     * Get the default value of the attribute as defined by the device model. If there is no
     * `defaultValue` for the attribute in the device model, then this method will return
     * `nil`. The value `nil` is not a default value.
     *
     * - returns:  the default value of the attribute, or `nil` if no default is defined
     */
    public func getDefaultValue() -> AnyObject? {
        return self.defaultValue
    }
    
    /*
    public func toString() -> String {
        var str: String = "urn = " + urn! + ", name = " + name!
        str = str + ", description = " + description! + ", type = " + type.rawValue
        str = str + ", lowerBound = " + lowerBound!.description
        str = str + ", upperBound = " + upperBound!.description
        str = str + ", access = " + access.rawValue + ", alias = " + alias!
        return str
    }
    */
}
