/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import SwiftyJSON

/**
 * DeviceModel Factory.
 */
class DeviceModelFactory {
    static let DM_PREFIX: String  = "dm-"
    static let FILE_TYPE: String = ".json"
    static let ALLOW_DRAFT_MODELS :Bool = Config.ALLOW_DRAFT_DEVICE_MODELS
    
    private init() {}
    
    /**
     * Get a device model from the IoT server.
     *
     * - parameter secureConnection: a connection to the server.
     *
     * - parameter urn: the URN of desired device model
     *
     * - parameter callback: This function is called with the retrieved device
     *                       model. If retrieving the device model is not 
     *                       successful then the object will be null and an 
     *                       error object is passed in the form 
     *                       `callback(device, error)` and the
     *                       reason can be taken from error
     */
    static func getDeviceModel(secureConnection: HttpSecureConnection,
                               urn: String,
                               callback: @escaping (DeviceModel?, ClientError?) -> Void) throws {
        // TODO: FIXME Does not really throw
        try getDeviceModel(aid: nil, secureConnection: secureConnection,
                           urn: urn, callback: callback)
    }
    
    private static func parseJsonStr(str: String) throws -> [String : AnyObject]? {
        if let data = str.data(using: .utf8) {
            return try JSONSerialization.jsonObject(with: data,
                                        options: []) as? [String : AnyObject]
        }
        return nil
    }
    
    /**
     * Get a device model from the IoT server.
     *
     * - parameter aid: the IoT application identifier
     *
     * - parameter secureConnection: a connection to the server.
     *
     * - parameter urn: the URN of desired device model
     *
     * - parameter callback: This function is called with the retrieved device
     *                       model. If retrieving the device model is not 
     *                       successful then the object will be null and an 
     *                       error object is passed in the form 
     *                       `callback(device, error)` and the
     *                       reason can be taken from error
     *
     * On error `callback` receives a `ClientError.network` error if:
     *    the object is not found (StatusCode.NOT_FOUND),
     *    the request fails (!StatusCode.OK), or
     *    the returned data is nil or JSON parsing fails
     */
    static func getDeviceModel(aid: String?, secureConnection: HttpSecureConnection,
                               urn: String,
                               callback: @escaping (DeviceModel?,ClientError?) -> Void) throws {
        // TODO: FIXME does not really throw
        let documentDirectoryURL = try! FileManager.default.url(
                                for: .documentDirectory, in: .userDomainMask,
                                appropriateFor: nil, create: true)
        
        let encodedUrn = urn.addingPercentEncoding(withAllowedCharacters: .urlPathAllowed)
        let documentFileURL = documentDirectoryURL.appendingPathComponent(DM_PREFIX + encodedUrn! + FILE_TYPE)
        
        var foundFileURL: URL?
        do {
            if try documentFileURL.checkResourceIsReachable() == true {
                // Check Documents folder first to see if device model was cached 
                // (dynamically provisioned) from an earlier request
                foundFileURL = documentFileURL
            } else {
                // Check bundles folder next to see if it was statically 
                // provisioned with the app
                if let bundleFileStr = Bundle.main.path(
                        forResource: DM_PREFIX + encodedUrn!, ofType: FILE_TYPE) {
                    let bundleFileURL = URL(fileURLWithPath: bundleFileStr)
                    if try bundleFileURL.checkResourceIsReachable() == true {
                        foundFileURL = bundleFileURL
                    }
                }
            }
        } catch {
            // The device model isn't cached
        }
        
        if (foundFileURL != nil) {
            do {
                let dm = try DeviceModelParser.fromJson(root: JSON(parseJSON:
                    String(contentsOf: foundFileURL!)))
                callback(dm, nil)
                return
            } catch {
                print (error)
                callback(nil, ClientError.file(NSLocalizedString("\(error)", comment: "")))
            }
        }
        
        // Download device model from server and cache it before returning to caller
        var uri: String
        
        if (aid != nil) {
            uri = RestApi.V2.getReqRoot() + "/apps/" + aid! + "/deviceModels"
            uri = uri + "?q=" + "{\"" + "urn" + "\""
            uri = uri + ":{\"" + "$eq"
            uri = uri + "\":\"" + urn + "\"}}"
        } else {
            /*
             * TODO: The server does not allow devices to list device
             *       models and use a filter like enterprise clients can,
             *       so we filter in the method below, when the server
             *       allows it, use the the code below instead.
             *
             * Filter f = Filter.and(Filter.eq("urn", urn),
             *            Filter.eq("draft", false));
             * uri = RestApi.V2.getReqRoot() +  "/deviceModels"
             * uri = uri + "&q=" + f.toJson().toString()
             */
            uri = RestApi.V2.getReqRoot() + "/deviceModels/" + urn
        }
        
        getObject(aid: aid, secureConnection: secureConnection, uri: uri,
                  path: documentFileURL, callback: callback)
    }

    private static func getObject(aid: String?,
                                  secureConnection: HttpSecureConnection,
                                  uri: String, path: URL,
                callback: @escaping (DeviceModel?, ClientError?) -> Void) {
        // TODO: FIXME exceptions are thrown from with the secureConnection 
        // completion handler
        // TODO: FIXME http errors are checked on return from 
        // callback not before.
        let accessToken = secureConnection.accessToken
        let headers: [String : String] = [
            "Content-Type" : "application/json",
            "Accept" : "application/json",
            "X-EndpointId" : secureConnection.trustedAssetsManager.getClientId(),
            "Authorization" : accessToken.getTokenType() + " " +
                accessToken.getToken()
        ]
        
        var status: Int = 0
        var data: Foundation.Data?
        var dm: DeviceModel?
        
        do {
            try secureConnection.get(restApi: uri,
                                     headers: headers,
                                     response: { httpResponse, err in
                if (err != nil) {
                    callback(nil, err)
                } else {
                    if let httpObject = httpResponse {
                        do {
                            status = httpObject.getStatus()
                            if status == StatusCode.OK.getCode() {
                                data = httpObject.getData()
                                
                                var dmj:JSON!
                                do {
                                    dmj = try JSON(data: data!)
                                } catch {
                                    let m = String(format:NSLocalizedString(
                                        "%@: JSON threw exception: %@", comment:""),
                                        "DeviceModelFactory.getObject()", "\(error)")
                                    print(m)
                                    callback(nil, ClientError.network(m))
                                    return
                                }
                                if let items = dmj.dictionaryValue["items"]?.array {
                                    if items.count > 0 {
                                        dmj = items[0]
                                    }
                                }
                                
                                dm = try DeviceModelParser.fromJson(root: dmj)
                                
                                do {
                                    try dmj.rawString()?.write(to: path,
                                                               atomically: true,
                                                               encoding: .utf8)
                                } catch {
                                    throw ClientError.file(NSLocalizedString("\(error)", comment: ""))
                                }
                                
                                callback(dm, nil)
                                
                                if (status == StatusCode.NOT_FOUND.getCode()) {
                                    return
                                }
                                if (status != StatusCode.OK.getCode()) {
                                    throw ClientError.state(NSLocalizedString(
                                        "Device Model: no data", comment: ""))
                                }
                                
                                if (data == nil) {
                                    throw ClientError.state(NSLocalizedString(
                                        "Device Model: no data", comment: ""))
                                }
                            } else {
                                callback(nil, ClientError.network(NSLocalizedString(
                                    "Get device model request failed.",
                                    comment: "")))
                            }
                        } catch let error as ClientError {
                            callback(nil, error)
                        } catch {
                            callback(nil, ClientError.network(NSLocalizedString("\(error)", comment: "")))
                        }
                    } else {
                        callback(nil, ClientError.argument(NSLocalizedString("DeviceModelFactory error: httpObject is nil", comment: "")))
                    }
                }
            })
        } catch {
            callback(nil, ClientError.state(NSLocalizedString("\(error)", comment: "")))
        }
        return
    }
}
