/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import CommonCrypto


/**
 * This is wrapper class for the CommonCrypto HMAC functions
 */
class HMAC {
    
    // MARK: - Types
    
    enum Algorithm {
        case sha1
        case md5
        case sha256
        case sha384
        case sha512
        case sha224
        
        var algorithm: CCHmacAlgorithm {
            switch self {
            case .md5: return CCHmacAlgorithm(kCCHmacAlgMD5)
            case .sha1: return CCHmacAlgorithm(kCCHmacAlgSHA1)
            case .sha224: return CCHmacAlgorithm(kCCHmacAlgSHA224)
            case .sha256: return CCHmacAlgorithm(kCCHmacAlgSHA256)
            case .sha384: return CCHmacAlgorithm(kCCHmacAlgSHA384)
            case .sha512: return CCHmacAlgorithm(kCCHmacAlgSHA512)
            }
        }
        
        var digestLength: Int {
            switch self {
            case .md5: return Int(CC_MD5_DIGEST_LENGTH)        // 16
            case .sha1: return Int(CC_SHA1_DIGEST_LENGTH)      // 20
            case .sha224: return Int(CC_SHA224_DIGEST_LENGTH)  // 28
            case .sha256: return Int(CC_SHA256_DIGEST_LENGTH)  // 32
            case .sha384: return Int(CC_SHA384_DIGEST_LENGTH)  // 48
            case .sha512: return Int(CC_SHA512_DIGEST_LENGTH)  // 64
            }
        }
        

    }
    
    static func getAlgorithm(algorithm: String) -> Algorithm {
        switch algorithm.lowercased() {
        case "hmacsha1": return .sha1
        case "hmacmd5": return .md5
        case "hmacsha256": return .sha256
        case "hmacsha384": return .sha384
        case "hmacsha512": return .sha512
        case "hmacsha224": return .sha224
        default: return .sha256
        }
    }
    // MARK: - Signing
    /**
     * This functions signs a collection of bytes
     *
     * - parameter data: the bytes to sign
     *
     * - parameter algorithm: the algorithm identifier, must be one of the 
     *                        identifier given in this class
     *
     * - parameter key: the key to generate the HMAC
     *
     * - returns: the signature as Foundation.Data
     */
    static func sign(data: Foundation.Data, algorithm: Algorithm,
                     key: Foundation.Data) -> Foundation.Data {
        let signature = UnsafeMutablePointer<CUnsignedChar>.allocate(
                                            capacity: algorithm.digestLength)
        data.withUnsafeBytes({(databytes: UnsafePointer) in
            key.withUnsafeBytes({(keybytes: UnsafePointer) in
                CCHmac(algorithm.algorithm, keybytes, key.count, databytes,
                       data.count, signature)
            })
        })
        return Foundation.Data(bytes: UnsafePointer<UInt8>(signature),
                               count: algorithm.digestLength)
    }
    
    /**
     * This functions signs a a String
     *
     * - parameter message: the String to sign
     *
     * - parameter algorithm: the algorithm identifier, must be one of the
     *                      identifier given in this class
     *
     * - parameter key: the key to generate the HMAC
     *
     * - returns: the signature as String or nil if the message or key can not 
     *             be converted to Data
     */
    static func sign(message: String, algorithm: Algorithm, key: String) -> String? {
        guard let messageData = message.data(using: .utf8, allowLossyConversion: false),
            let keyData = key.data(using: .utf8)
            else { return nil }
        
        let data = sign(data: messageData, algorithm: algorithm, key: keyData)
        
        var hash = ""
        data.enumerateBytes { bytes, begin, _ in
            /*
            for i in bytes.startIndex..<range {
                hash += String(format: "%02x", bytes[i])
            }
            */
            for i in begin..<bytes.endIndex {
                hash += String(format: "%02x", bytes[i])
            }
        }
        return hash
        
    }
}

//-----------------------------------------------
//
//  PBKDF2.swift
//  CryptoSwift
//
//  Created by Marcin Krzyzanowski on 05/04/16.
//  Copyright © 2016 Marcin Krzyzanowski. All rights reserved.
//
struct PBKDF2 {
        
        enum PBKDF2Error: Error {
            case invalidInput
        }
        
        private let salt: [UInt8]
        fileprivate let iterations: Int  // c
        private let numBlocks: UInt  // l
        fileprivate let hmac: HMACholder
        private let expectedKeyLength: Int
 
    
        init(password: [UInt8], salt: [UInt8], iterations: Int = 1000, /* c */
             keyLength: Int? = nil /* dkLen */ ,
             algorithm: HMAC.Algorithm = .sha256) throws {
            
            // check the input and instantiate a HMAC object without really 
            // performing any HMAC
 
            guard let hmac = HMACholder(key: password, algorithm: algorithm),
                    (iterations > 0) && (password.count > 0) && (salt.count > 0) else {
                throw PBKDF2Error.invalidInput
            }

            
            let keyLengthFinal: Int
            if let kl = keyLength {  // if a keyLength is given
                keyLengthFinal = kl
                self.expectedKeyLength = kl
            } else {
                keyLengthFinal = algorithm.digestLength
                self.expectedKeyLength = keyLengthFinal
            }
            
            let hLen = Double(algorithm.digestLength)
            /* 
             * TODO: FIXME
             * keyLength should be keyLengthFinal
             * Int(((pow(2,32) as Double) - 1) * hLen) exceeds MAXINT on the
             * device
            if keyLength > Int(((pow(2,32) as Double) - 1) * hLen) {
                throw Error.InvalidInput
            }
            */
            
            self.salt = salt
            self.iterations = iterations
            self.hmac = hmac
            
            self.numBlocks = UInt(ceil(Double(keyLengthFinal) / hLen))
        }
        
        func calculate() -> [UInt8] {
            var ret = [UInt8]()
            for i in 1...self.numBlocks {
                // for each block T_i = U_1 ^ U_2 ^ ... ^ U_iter
                if let value = calculateBlock(salt: self.salt, blockNum: i) {
                    ret.append(contentsOf: value)
                }
            }
            // the following code is added to cut of extra bytes when the digest
            // is longer then the expected key length, this is here to make the 
            // function compatible with the bouncycastle implementation for java
            if(ret.count > self.expectedKeyLength) {
               let retKLsl = ret[0...(self.expectedKeyLength-1)]
               let retKL = Array(retKLsl)
               return retKL
            }
            return ret
        }
    }

private extension PBKDF2 {
    func INT(i: UInt) -> [UInt8] {
        var inti = [UInt8](repeating: 0, count: 4)
        inti[0] = UInt8((i >> 24) & 0xFF)
        inti[1] = UInt8((i >> 16) & 0xFF)
        inti[2] = UInt8((i >> 8) & 0xFF)
        inti[3] = UInt8(i & 0xFF)
        return inti
    }
    
    // F (P, S, c, i) = U_1 \xor U_2 \xor ... \xor U_c
    // U_1 = PRF (P, S || INT (i))
    func calculateBlock(salt: [UInt8], blockNum: UInt) -> [UInt8]? {
        // here is the actual HMAC
        guard let u1 = self.hmac.authenticate(message: salt + INT(i: blockNum)) else {
            return nil
        }
        
        var u = u1
        var ret = u
        if self.iterations > 1 {
            for _ in 2...self.iterations {
                u = self.hmac.authenticate(message: u)!
                for x in 0..<ret.count {
                    ret[x] = ret[x] ^ u[x]
                }
            }
        }
        return ret
    }
}

final class HMACholder {
    
    var key:[UInt8]
    var keyData:Foundation.Data
    private let algorithm: HMAC.Algorithm
    
    init? (key: [UInt8], algorithm:HMAC.Algorithm = .sha256) {
        self.algorithm = algorithm
        self.key = key
        self.keyData =  Foundation.Data(bytes: UnsafePointer<UInt8>(self.key),
                                        count: self.key.count)
    }
    
    func authenticate(message:[UInt8]) -> [UInt8]? {
        // fix it conversion from UInt8 to NSdata and back to UInt8
        let data = Foundation.Data(bytes: UnsafePointer<UInt8>(message),
                                   count: message.count)
        let ret:Foundation.Data = HMAC.sign(data: data,
                                            algorithm: self.algorithm,
                                            key: self.keyData)
        return ret.withUnsafeBytes({(hmacBytes: UnsafePointer) in
            return Array(UnsafeBufferPointer(start: hmacBytes, count: ret.count))
        })
    }
}

//----- Crypto Helper function to perform AES128 en/Decryption based on the 
// CommonCrypto ---------

final class CryptoHelper {
    
    class func generateRandomBytes(count: Int) -> Foundation.Data? {
        var result = [UInt8](repeating: 0, count: count)
        // Warning is now error need to use result.
        let _ = SecRandomCopyBytes(kSecRandomDefault, count, &result)
        
        return Foundation.Data(bytes: UnsafePointer<UInt8>(result), count: count)
    }

    class func blockSize(algorithm: CCAlgorithm) -> Int {
        switch Int(algorithm) {
        case kCCAlgorithmAES128, kCCAlgorithmAES:
            return kCCBlockSizeAES128
        case kCCAlgorithmDES:
            return kCCBlockSizeDES
        case kCCAlgorithm3DES:
            return kCCBlockSize3DES
        case kCCAlgorithmCAST:
            return kCCBlockSizeCAST
        case kCCAlgorithmRC2:
            return kCCBlockSizeRC2
        case kCCAlgorithmBlowfish:
            return kCCBlockSizeBlowfish
        default:
            return 0
        }
    }

    class func encrypt(data: Foundation.Data, key: Foundation.Data,
            iv: Foundation.Data, algorithm: CCAlgorithm) -> Foundation.Data? {
        let dataLength = data.count
        // Becomes encryptedData
        var result: Foundation.Data =
            Foundation.Data(count: dataLength + key.count + iv.count)
        return result.withUnsafeMutableBytes({(resultBytes: UnsafeMutablePointer<UInt8>) in
            return data.withUnsafeBytes({(dataBytes: UnsafePointer<UInt8>) in
                let keyLength = size_t(key.count)
                return key.withUnsafeBytes({(keyData: UnsafePointer<UInt8>) in
                    return iv.withUnsafeBytes({(ivData: UnsafePointer<UInt8>) in
                        
                        let encryptedDataLength = size_t(result.count)
                        var encryptedLength: size_t = 0
                        
                        let status = CCCrypt(CCOperation(kCCEncrypt), algorithm,
                            CCOptions(kCCOptionPKCS7Padding), keyData, keyLength,
                                ivData, dataBytes, dataLength, resultBytes,
                                encryptedDataLength, &encryptedLength)
                        if status == Int32(kCCSuccess) {
                            result.count = Int(encryptedLength)
                            return result
                        } else {
                            return nil
                        }
                    })
                })
            })
        })
    }

    // This is the special method encryption method for the unified Provisioner 
    // it encrypts the given text but then adds in front of the encrypted text 
    // the initialization vector
    // return: iv concatenated with encrypted text
    //
    // The algorithm is AES128/CBC/PKCS7 padding
    // PKCS7 padding is the only padding supported by CommonCrypto, PKCS5 which 
    // is used in the Java implementation is a subset of PKCS7. PKCS5 is only 
    // defined for 8 byte bloch cipher so will not work for AES.
    // the Java implementation substitute PKCS5 with PKCS7 when set as padding 
    // for AES
    //
    class func encryptUP(data: Foundation.Data, key: Foundation.Data,
            iv: Foundation.Data, algorithm: CCAlgorithm) -> Foundation.Data? {
        let encryptedData = encrypt(data: data, key: key,iv: iv,
                                    algorithm: algorithm)
        var data = Foundation.Data()
        data.append(iv)
        data.append(encryptedData!)
        return data as Foundation.Data
    }
    
    // This is the special method for decryption for the unified Provisioner, 
    // it expects that the first 16 bytes are the iv for the AES decryption. If 
    // the data.length is less the 16 or there is no data to decrypt and the 
    // method returns nil
    // return: decrypted Data
    //
    class func decryptUP(data: Foundation.Data, key: Foundation.Data,
                         algorithm: CCAlgorithm) -> Foundation.Data? {
        // First block is the IV 
        if (data.count > 16) {
            // substitute 16 here with the IV length for the UnifiedProvisioner
            let iv = data.subdata(in: Range(0..<16))
            let dataC = data.subdata(in: Range(16..<(data.count/*-16*/)))
            let decryptedData = decrypt(data: dataC, key: key,iv: iv,
                                        algorithm: algorithm)
            return decryptedData
        }
        return nil
    }

    class func decrypt(data: Foundation.Data, key: Foundation.Data,
            iv: Foundation.Data, algorithm: CCAlgorithm) -> Foundation.Data? {
        
        let encryptedDataLength = data.count
        // Becomes decrypted data
        var result: Foundation.Data = Foundation.Data(count: encryptedDataLength)
        return result.withUnsafeMutableBytes({(resultBytes: UnsafeMutablePointer<UInt8>) in
            return data.withUnsafeBytes({(encryptedData: UnsafePointer<UInt8>) in
                return key.withUnsafeBytes({(keyData: UnsafePointer<UInt8>) in
                    let keyLength = size_t(key.count)
                    return iv.withUnsafeBytes({(ivData: UnsafePointer<UInt8>) in
                        let decryptedDataLength = size_t(result.count)
                        var decryptedLength: size_t = 0
                        let status = CCCrypt(CCOperation(kCCDecrypt), algorithm,
                            CCOptions(kCCOptionPKCS7Padding), keyData, keyLength,
                            ivData, encryptedData, encryptedDataLength, resultBytes,
                            decryptedDataLength, &decryptedLength)
                        if UInt32(status) == UInt32(kCCSuccess) {
                            result.count = Int(decryptedLength)
                            return result
                        } else {
                            return nil
                        }
                    })
                })
            })
        })
    }
}

//------------- end of Crypto Helper --------------------------------
