/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import Alamofire
import SwiftyJSON

public protocol SecureConnection {
    
    // TODO: Why is "unowned" needed ? Causes trustedAssetsManager reference to
    // get lost in the EnterpriseLib
    /*unowned*/ var trustedAssetsManager:TrustedAssetsManager{get set}
    
    var useOnlySharedSecret: Bool {get set}
    
    func get(restApi: String, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws
    
    func post(restApi: String, payload: Foundation.Data?, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws
    
    func put(restApi: String, payload: Foundation.Data, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws
    
    func delete(restApi: String, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws
    
    func patch(restApi: String, payload: Foundation.Data, headers: inout [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws
    
    func getTrustedAssetsManager() -> TrustedAssetsManager
    
    func getEndpointId() -> String?
 
    
}

extension SecureConnection {
    
    public func getTrustedAssetsManager() -> TrustedAssetsManager {
        return trustedAssetsManager
    }
    
    public func getEndpointId() -> String? {
            return trustedAssetsManager.getEndpointId()
    }
    
    // Closeable "override"
    public func close() throws {
        disconnect()
        self.trustedAssetsManager.close()
    }
    
    public func disconnect() {
    }
    
    func usesOnlySharedSecret() -> Bool {
        return useOnlySharedSecret
    }
    
}

class HttpSecureConnection : SecureConnection {

    // TODO: Why is this needed ? Causes trustedAssetsManager reference to
    // get lost in the EnterpriseLib
    /*unowned*/ var trustedAssetsManager:TrustedAssetsManager
    var useOnlySharedSecret: Bool
    
    var accessToken: AccessToken
    var defaultHttpManager:Alamofire.SessionManager?
    // Create a DispatchQueue for executing the Alamofire request
    // completion handler. It must not be on the main thread. Note that
    // since it is not on the main thread it cannot update UI components
    // All requests must set the queue parameter to this queue.
    // See Transport.swift
    var responseQueue:DispatchQueue {
        get {
            return DispatchQueue(label:"HttpSecureConnection response queue",
                          qos: .utility)
        }
    }
    
    init(trustedAssetsManager: TrustedAssetsManager, useOnlySharedSecret: Bool) {
        self.trustedAssetsManager = trustedAssetsManager
        self.useOnlySharedSecret = useOnlySharedSecret
        self.accessToken = AccessToken(expires: 0, tokenType: "", token: "")
        
        let configuration = URLSessionConfiguration.default
        if (!Config.DISABLE_LONG_POLLING) {
            configuration.timeoutIntervalForRequest = Config.LONG_POLLING_TIMEOUT
            configuration.timeoutIntervalForResource = Config.LONG_POLLING_TIMEOUT
        }
        // else {
        //  keep default Alamofire's timeout (60 seconds)
        // }
        
        if (self.trustedAssetsManager.getServerTrustPolicy() != nil) {
            let serverTrustPolicies: [String: ServerTrustPolicy] = [self.trustedAssetsManager.getServerHost(): self.trustedAssetsManager.getServerTrustPolicy()!]
            
            configuration.httpAdditionalHeaders = Alamofire.SessionManager.defaultHTTPHeaders
            
            self.defaultHttpManager = Alamofire.SessionManager(
                configuration: configuration,
                serverTrustPolicyManager: ServerTrustPolicyManager(policies: serverTrustPolicies)
            )
        } else {
            self.defaultHttpManager = Alamofire.SessionManager(configuration: configuration)
        }
        //self.responseQueue = DispatchQueue(label: "HttpSecureConnection response queue",
        //                                   qos: .utility)
        
    }
    
    deinit {
        self.defaultHttpManager = nil
    }
    
    
    func get(restApi: String, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws {
        try invoke(method: "GET", restApi: restApi, payload: Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0), headers: headers, response: { (invokeGet, error) -> Void in
            response(invokeGet, error)
        })
    }

    func post(restApi: String, payload: Foundation.Data?, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws {
        try invoke(method: "POST", restApi: restApi, payload: payload , headers: headers, response: { (invokePost, error) -> Void in
            response(invokePost, error)
        })
    }
    
    func put(restApi: String, payload: Foundation.Data, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws  {
        try invoke(method: "PUT", restApi: restApi, payload: payload, headers: headers, response: { (invokePut, error) -> Void in
            response(invokePut, error)
        })
    }
    
    func delete(restApi: String, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws {
        try invoke(method: "DELETE", restApi: restApi, payload: Foundation.Data(bytes: UnsafePointer<UInt8>([]), count: 0), headers: headers, response: { (invokeDelete, error) -> Void in
            response(invokeDelete, error)
        })
    }

    func patch(restApi: String, payload: Foundation.Data, headers: inout [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws  {
        headers["X-HTTP-Method-Override"] = "PATCH"
        try invoke(method: "POST", restApi: restApi, payload: payload, headers: headers, response: { (invokePatch, error) -> Void in
            response(invokePatch, error)
        })
    }
    
    func invoke(method: String, restApi: String, payload: Foundation.Data?, headers: [String : String], response: @escaping ((HttpResponse?, ClientError?) -> Void)) throws  {
        if self.accessToken.getToken().isEmpty || self.accessToken.hasExpired() {
            do {
                try renewAccessToken(secureHttpConnection: self, callback: { error -> Void in
                    if (error != nil) {
                        response(nil, error)
                    } else {
                        self.callRestMethod(method: method, restApi: restApi, payload: payload, headers: headers, callback: response)
                    }
                })
            } catch let error as ClientError {
                response(nil, error)
            }
        } else {
            self.callRestMethod(method: method, restApi: restApi, payload: payload, headers: headers, callback: response)
        }
    }
    
    private func callRestMethod(method: String, restApi: String, payload: Foundation.Data?, headers: [String : String], callback: @escaping ((HttpResponse?, ClientError?) -> Void)) {
        let serverHost: String = getTrustedAssetsManager().getServerHost()
        let serverPort: String = getTrustedAssetsManager().getServerPort()
        var URL: String = "\(getTrustedAssetsManager().getServerScheme())://\(serverHost):\(serverPort)\(restApi)"
        
        var _headers = headers
        _headers["Authorization"] = self.accessToken.getTokenType() + " " + self.accessToken.getToken()
        if RestApi.V2.isWebApi() == false {
            if !self.getTrustedAssetsManager().getEndpointId().isEmpty {
                _headers["X-EndpointId"] = self.getTrustedAssetsManager().getEndpointId()
            }
        }
        
        URL = URL.addingPercentEncoding(withAllowedCharacters: .urlQueryAllowed)!
        
        let httpClient = HttpClient(url: URL, httpManager: self.defaultHttpManager!)
        
        self.invoke(httpClient: httpClient, method: method, payload: payload, headers: _headers, callback: { (firstResponse, error) -> Void in
            if (error != nil) {
                callback(nil, error)
                return
            } else {
                if (firstResponse != nil) {
                    // Check if credentials have expired (401) or client-secret used (403)
                    if firstResponse!.getStatus() == StatusCode.UNAUTHORIZED.getCode() || firstResponse!.getStatus() == StatusCode.FORBIDDEN.getCode() {
                        do {
                            try self.renewAccessToken(secureHttpConnection: self, callback: { error -> Void in
                                if (error != nil) {
                                    callback(nil, error)
                                } else {
                                    _headers["Authorization"] = self.accessToken.getTokenType() + " " + self.accessToken.getToken()
                                    
                                    self.invoke(httpClient: httpClient, method: method, payload: payload, headers: _headers, callback: { (secondResponse, error) -> Void in
                                        if (error != nil) {
                                            callback(nil, error)
                                            return
                                        } else {
                                            callback(secondResponse, nil)
                                            return
                                        }
                                    })
                                }
                            })
                        } catch let error as ClientError {
                            callback(nil, error)
                        }
                        catch {
                            callback(nil, ClientError.network(NSLocalizedString("\(error.localizedDescription)", comment: "")))
                        }
                    } else {
                        callback(firstResponse, nil)
                        return
                    }
                } else {
                    callback(nil, ClientError.argument(NSLocalizedString("HttpSecureConnection callRestMethod() firstResponse should not be nil", comment: "")))
                }
            }
        })
    }
    
    private func invoke(httpClient: HttpClient, method: String, payload: Foundation.Data?, headers: [String : String], callback: @escaping ((HttpResponse?, ClientError?) -> Void)) -> Void {
        do {
            if method == "GET" {
                try httpClient.get(headers: headers, response: { (invokeResult, error) -> Void in
                    callback(invokeResult, error)
                })
            } else if method == "POST" {
                var ipayload:Foundation.Data? = payload
                if(payload == nil) {
                    ipayload = String("[]").data(using: .utf8)!
                }
                try httpClient.post(data: ipayload!, headers: headers, response: { (invokeResult, error) -> Void in
                    callback(invokeResult, error)
                })
            } else if method == "DELETE" {
                try httpClient.delete(headers: headers, response: { (invokeResult, error) -> Void in
                    callback(invokeResult, error)
                })
            } else { // Must be "PUT"
                try httpClient.put(data: payload!, headers: headers, response: { (invokeResult, error) -> Void in
                    callback(invokeResult, error)
                })
            }
        } catch {
            callback(nil, ClientError.state(NSLocalizedString("Error while executing HTTP method: " + method, comment: "")))
        }
        
    }
    
    func renewAccessToken(secureHttpConnection: HttpSecureConnection, callback: @escaping (ClientError?) -> Void) throws -> Void {
        let trustedAssetsManager = secureHttpConnection.getTrustedAssetsManager()
        let serverHost = trustedAssetsManager.getServerHost()
        let serverPort = trustedAssetsManager.getServerPort()
        
        RestApi.USE_WEBAPI = false
        let restAPI = RestApi(version: "v2")
        
        // build the URL String
        let oauthRequestURL:String = trustedAssetsManager.getServerScheme() + "://" +
            serverHost + ":" + serverPort + restAPI.getReqRoot() + trustedAssetsManager.getOAuthRequestPath()
        
        let headers: [String:String] = ["Content-Type" : "application/x-www-form-urlencoded", "Accept" : "application/json"]
        // create the client Assertion
        let client_asertion:String?
        var useSharedSecret:Bool = true
        
        if(trustedAssetsManager.isActivated() && !trustedAssetsManager.getClientId().isEmpty) {
            useSharedSecret = false
        }
        
        // sha
        client_asertion = try self.buildClientAssertion(tam: trustedAssetsManager, secureHttpConnection: secureHttpConnection,
                                               useOnlySharedSecret: useSharedSecret)
        
        // generate the OAuth2 parameters; parameters must come from the TAM
        let oauthParameters: [String:String] = ["grant_type" : "client_credentials",
                                                "client_assertion_type" : "urn:ietf:params:oauth:client-assertion-type:jwt-bearer",
                                                "client_assertion" : client_asertion!,
                                                "scope" : trustedAssetsManager.isActivated() ? "":"oracle/iot/activation"]
        
        Config.postCounter += 1
        defaultHttpManager!.request(oauthRequestURL,
            method: .post,
            parameters: oauthParameters,
            encoding: URLEncoding.default,
            headers: headers)
            .validate(statusCode: 200..<404)
            .responseJSON(queue: self.responseQueue,
                          completionHandler: { (response) -> Void in
                switch response.result {
                    case .success(let value):
                        if response.response?.statusCode == StatusCode.BAD_REQUEST.getCode() ||
                                response.response?.statusCode == StatusCode.FORBIDDEN.getCode() ||
                                response.response?.statusCode == StatusCode.UNAUTHORIZED.getCode() {
                            self.nextTry += 1
                            let responseTime = HttpSecureConnection.parseResponseTime(responseData: response.data!)
                            TimeManager.setCurrentTimeMillis(serverTime: responseTime)
                            if self.nextTry < 2 {
                                do {
                                    try self.renewAccessToken(secureHttpConnection: secureHttpConnection, callback: callback)
                                } catch let error as ClientError {
                                    self.nextTry = 0
                                    callback(error)
                                } catch {
                                    self.nextTry = 0
                                    callback(ClientError.network(NSLocalizedString("\(error.localizedDescription)", comment: "")))
                                }
                            } else {
                                callback(ClientError.network(String(format: NSLocalizedString(
                                    "Failed to renew access token after %d attempts", comment: ""),
                                    self.nextTry)))
                                self.nextTry = 0
                                return
                            }
                        } else {
                            let json = JSON(value)
                            if ((json["error"] != JSON.null) || (json["error_description"] != JSON.null)) {
                                self.nextTry = 0
                                callback(ClientError.network(NSLocalizedString("Renew token error: " + json["error_description"].string!, comment: "")))
                                return
                            } else if let myAccessToken = json["access_token"].string, let myTokenType = json["token_type"].string, let myExpiresIn = json["expires_in"].int64 {
                                let token: AccessToken = AccessToken(expires: 0, tokenType: "", token: "")
                                token.token = myAccessToken
                                token.tokenType = myTokenType
                                // In the mock server is only 3 seconds multiply by 1000
                                token.expires = myExpiresIn * 1000
                                token.expirationTime += token.expires
                                self.accessToken = token
                                self.nextTry = 0
                                callback(nil)
                                return
                            } else {
                                self.nextTry = 0
                                callback(ClientError.network(NSLocalizedString("Renew token: retrieved JSON error", comment: "")))
                                return
                            }
                        }
                    case .failure(let error):
                        print(error)
                        self.nextTry = 0
                        callback(ClientError.network(NSLocalizedString("Renew token: JSON response validation failed with error " /* +
                            String(describing: error)*/, comment: "")))
                        return
                }
            }
        )
    }
    
    func close() { 
        // TODO: In case we need to wait for tasks to finish then use
        // finishTasksAndInvalidate() but be aware method returns immediately
        // => handle callback from SessionDelegate
        self.defaultHttpManager?.session.invalidateAndCancel()
        self.trustedAssetsManager.close()
        disconnect()
    }
    
    func disconnect() {
        self.accessToken = AccessToken(expires: 0, tokenType: "", token: "")
    }
    
    private static func parseResponseTime(responseData: Foundation.Data) -> Int64 {
        
        // TODO: This 900000 is the ACCESS_TOKEN EXP_DELTA used in JCL
        // Needs to be moved out of here.
        let responseTime:Int64 = Int64(Int64(Date().timeIntervalSince1970 * 1000) + 900000)
        var jsonObject:JSON!
        do {
            jsonObject = try JSON(data: responseData)
        } catch {
            print(String(format:NSLocalizedString(
                "%@:JSON threw exception: %@", comment:""),
                "HttpSecureConnection.parseResponseTime()", "\(error)"))
            return responseTime
        }
        
        if jsonObject["currentTime"] == JSON.null {
            return responseTime
        }
        return jsonObject["currentTime"].int64Value
    }
    
    private var nextTry = 0
    
    func buildClientAssertion(tam:TrustedAssetsManager, secureHttpConnection:
        HttpSecureConnection, useOnlySharedSecret:Bool) throws -> String {
        
        var id:String
        let expTime = TimeManager.currentTimeMillis() / 1000
        
        if (!useOnlySharedSecret && tam.isActivated()) {
            id = tam.getClientId()
        } else {
            id = tam.getEndpointId()
            if id.isEmpty {
                id = tam.getClientId()
            }
        }
        
        let sharedSecret = useOnlySharedSecret || !tam.isActivated()
        var alg:String  = sharedSecret ? "HS256" : "RS256"
        // TODO: Note was not in Enterprise version
        // sha
        if(tam.isActivated()) {
            alg = "RS256"
        }
        // TODO: swift 3 Use Dictionary not NSDictionary
        // set up the Dictionary for the JSon objects
        let header:NSMutableDictionary = NSMutableDictionary()
        header.setValue("JWT", forKey: "typ")
        header.setValue(alg, forKey: "alg")
        
        let headerStr = self.generateJSONString(para: header) // the header JSON string
        
        var claimStr = "{\"iss\":\"";
        claimStr += id
        claimStr += "\",\"sub\":\""
        claimStr += id
        claimStr += "\",\"aud\":\"oracle/iot/oauth2/token\",\"exp\":"
        claimStr += String(expTime)
        claimStr += "}"
        
        let inputToSign = self.encodeStringBase64String(str: headerStr) + "." + self.encodeStringBase64String(str: claimStr)
        
        var signed:String = ""
        // TODO: The if clause in the Enterprise version included "|| tam.getEndpointId().isEmpty == false)"
        // Is it needed here ?
        // SHA
        if(tam.isActivated() == false /*|| tam.getEndpointId().isEmpty == false*/) {
            
            signed = try tam.signWithSharedSecretString(data: inputToSign, algorithm: "HmacSHA256", activationId: nil)
            
            var data = self.dataFromHexadecimalString(str: signed)
            let count = data!.count / MemoryLayout<UInt8>.size
            
            // create an array of Uint8
            var array = [UInt8](repeating: 0, count: count)
            
            // copy bytes into array
            (data! as Foundation.Data).copyBytes(to: &array, count:count * MemoryLayout<UInt8>.size)
            //Convert [UInt8] to Data
            data = Foundation.Data(bytes: UnsafePointer<UInt8>(array), count: array.count)
            
            //Encode to base64
            let base64Data = data!.base64EncodedData(options: .lineLength64Characters)
            
            if let tempStr = String(data: base64Data, encoding: .utf8) {
                signed = tempStr
            } else {
                print("not a valid UTF-8 sequence")
            }
            
        }
        else {
            signed = try self.trustedAssetsManager.signWithPrivateKeyString(str: inputToSign, algorithm: "PKCS1256")!
        }
        
        var result = inputToSign + "." + signed
        result = result.replacingOccurrences(of: "+", with: "-").replacingOccurrences(of: "/", with: "_")
        
        let nextToLast:String.Index = result.index(before: result.endIndex)
        if result[nextToLast] == "=" {
            result = String(result[result.startIndex..<nextToLast])
        }
        
        return result
    }
    
    func generateJSONString(para:NSMutableDictionary) -> String {
        
        JSONSerialization.isValidJSONObject(para)
        let jsonData = try! JSONSerialization.data(withJSONObject: para, options: JSONSerialization.WritingOptions())
        
        let jsonString = String(data: jsonData, encoding: .utf8)!
        return jsonString
    }
    
    /// - returns: String representing encoded String a BASE64 String
    func encodeStringBase64String(str: String) -> String {
        let utf8str = self.encodeStringBase64(str: str)
        let base64Encoded = utf8str.base64EncodedString(
            options: Foundation.Data.Base64EncodingOptions(rawValue: 0))
        return base64Encoded
    }
    
    /// - returns: Data representing the UTF8 byte array converted input String
    func encodeStringBase64(str: String) -> Foundation.Data {
        return str.data(using: .utf8)!
    }
    
    /// - returns: Data represented by this hexadecimal string. Returns nil if string contains characters outside the 0-9 and a-f range.
    
    func dataFromHexadecimalString(str: String) -> Foundation.Data? {
        let trimmedString = str.trimmingCharacters(in: CharacterSet(charactersIn: "<> ")).replacingOccurrences(of: " ", with: "")
        
        // make sure the cleaned up string consists solely of hex digits, and that we have even number of them
        
        let regex = try! NSRegularExpression(pattern: "^[0-9a-f]*$", options: .caseInsensitive)
        
        let found = regex.firstMatch(in: trimmedString, options: [], range: NSMakeRange(0, trimmedString.count))
        if found == nil || found?.range.location == NSNotFound || trimmedString.count % 2 != 0 {
            return nil
        }
        
        // everything ok, so now let's build Data
        
        // TODO: swift 3 Don't use "NSMutableData"
        let data = NSMutableData(capacity: trimmedString.count / 2)
        
        var index = trimmedString.startIndex
        while index < trimmedString.endIndex {
            let range = index..<trimmedString.index(index, offsetBy: 2)
            let byteString = trimmedString[range]
            
            let num = UInt8(byteString.withCString { strtoul($0, nil, 16) })
            data?.append([num] as [UInt8], length: 1)
            
            //index = trimmedString.index(after: trimmedString.index(index, offsetBy: 2))
            index = trimmedString.index(index, offsetBy: 2)
        }
        
        return data as Foundation.Data?
    }
    
    static func createHttpSecureConnection(tam: TrustedAssetsManager, useOnlySharedSecret: Bool) -> HttpSecureConnection {
        return HttpSecureConnection(trustedAssetsManager: tam, useOnlySharedSecret: useOnlySharedSecret)
    }
}
