/*
 * Copyright (c) 2017, Oracle and/or its affiliates.  All rights reserved.
 * 
 * This software is dual-licensed to you under the MIT License (MIT) and
 * the Universal Permissive License (UPL).  See the LICENSE file in the root
 * directory for license terms.  You may choose either license, or both.
 */

import Foundation

//package com.oracle.iot.client.impl

// TODO: Encapsulate the contentQueue so that the synchronization
// is encapsulated by the contentQueue implementation vs. explicitly
// using a dispatch queue to synchronize every reference to the content queue.

/**
 * Handles the content queue for pending uploads and downloads to the SCS.
 */
fileprivate class ContentTransmitter : NSObject {
    
    var impl:StorageDispatcherImpl!
    init(impl: StorageDispatcherImpl) {
        self.impl = impl
        super.init()
    }
    fileprivate func run() {
        //self.impl.group.enter()
        // Serialize requests to upload/download but return immediately after
        // queuing the request
        self.impl.storageQueue.async(execute: {
            // run until close is requested
            // query the contentQueue in a synchronized fashion
            while self.impl.contentQueue.sync(execute: {
                    return !self.impl.requestClose }) {
                self.impl.contentQueued.lock()
                while self.impl.contentQueue.sync(execute: {
                                    return self.impl.queue.isEmpty}) {
                    self.impl.contentQueued.wait()
                }
                // we've been signalled
                self.impl.contentQueued.unlock()
                // TODO: use better name than queuePoll, it doesn't block if
                // there are no entries
                //let storageObject: AdvancedStorageObject? = self.impl.queuePoll()
                let storageObject: StorageObjectDelegate? = self.impl.queuePoll() as? StorageObjectDelegate
                if storageObject != nil {
                    storageObject!.setState(state:ProgressState.inProgress)
                    self.impl.dispatchProgressCallback(storageObject:storageObject!)
                    self.impl.transferAndCallback(storageObject:storageObject!,
                                                  callback: { (error) in })
                }
            }
            /*
            defer {
                self.impl.group.leave()
            }
            */
        })
    }
}

public class ProgressImpl : Progress {

    private var storageObject: AdvancedStorageObject
    private var exception:Error? = nil
    private var bytesTransferred:Int64 = 0
    
    init(storageObject: AdvancedStorageObject) {
        self.storageObject = storageObject
    }
    
    func setFailureCause(e:Error) {
        self.exception = e
    }
    
    func setBytesTransferred(bytesTransferred:Int64) {
        self.bytesTransferred = bytesTransferred
    }

    func setStorageObject(storageObject: AdvancedStorageObject) {
        self.storageObject = storageObject
    }
    
    public func getStorageObject() -> AdvancedStorageObject {
        return storageObject
    }
    
    public func getState() -> ProgressState {
        return (storageObject as! StorageObjectDelegate).getState()
    }
    
    public func getFailureCause() -> Error? {
        return exception
    }
    
    public func getBytesTransferred() -> Int64 {
        return bytesTransferred
    }
}

fileprivate let STORAGE_QUEUE_LABEL:String =
    "com.oracle.storageDispatcher."

/**
 * The StorageDispatcher queues content for automatic upload to or download from
 * the Oracle Storage Cloud Service.
 * There is one StorageDispatcher per StorageConnection.
 */
class StorageDispatcherImpl : StorageDispatcher {

    /**
     * Serial dispatch queues for message dispatching activities
     */
    fileprivate let storageQueue = DispatchQueue(
        label: STORAGE_QUEUE_LABEL + "storageQueue", qos: .utility)
    /**
     * DispatchQueue for progress callbacks
     */
    private let progressQueue = DispatchQueue(label:
        STORAGE_QUEUE_LABEL + "progressQueue", qos: .utility)
    
    /**
     * The `DispatchGroup` that transfers content to/from Oracle Cloud Service
     */
    fileprivate let storageGroup = DispatchGroup()
    /**
     * Called as a transfer progresses
     */
    fileprivate var progressCallback:ProgressCallback?
    // TODO: Make that contentQueued controls access to all properties 
    // and queue operations to ensure only one thread operates on them at
    // a time.
    /**
     * Controls access to queue operations
     */
    fileprivate var contentQueued:NSCondition = NSCondition()
    fileprivate var queue:[AdvancedStorageObject] = [AdvancedStorageObject]()
    
    // Serial queue to synchronize queue operations. execute always called
    // as sync, also used to synchronize close and requestClose
    fileprivate var contentQueue:DispatchQueue =
        DispatchQueue(label: "contentQueue", qos: .utility)
    private var contentTransmitter:ContentTransmitter!
    
    fileprivate var requestClose:Bool = false
    private var closed:Bool = false
    
    private static var dispatcherDcdMap:Dictionary<AdvancedDirectlyConnectedDevice,StorageDispatcher> =
        Dictionary<AdvancedDirectlyConnectedDevice,StorageDispatcher>()

    init() {
        self.closed = false
        self.requestClose = false
        self.contentTransmitter = ContentTransmitter(impl:self)
        self.contentTransmitter.run()
    }

    /**
     * Get the `StorageDispatcher` for the given `AdvancedDirectlyConnectedDevice`.
     * There is one instance of `StorageDispatcher` for a `dcd`.
     *
     * - parameter dcd: the `AdvancedDirectlyConnectedDevice`
     *
     * - returns: a `StorageDispatcher`
     */
    public class func getStorageDispatcher(dcd:AdvancedDirectlyConnectedDevice)
            -> StorageDispatcher {
        
        if let storageDispatcher:StorageDispatcher = dispatcherDcdMap[dcd] {
            return storageDispatcher
        } else {
            let sd = StorageDispatcherImpl()
            dispatcherDcdMap[dcd] = sd
            return sd as StorageDispatcher
        }
    }
    
    /**
     * remove and return the first entry in the queue
     */
    fileprivate func queuePoll() -> AdvancedStorageObject? {
        return contentQueue.sync(execute: {
            contentQueued.lock()
            defer {
                contentQueued.unlock()
            }
            return queue.removeFirst()
        })
    }
    
    /**
     * Add the content storageObject to the queue if it is possible to
     * do so without violating capacity restrictions.
     *
     * - parameter storageObject The content storageObject to be queued
     */
    public func queue(storageObject: AdvancedStorageObject?) throws {
        guard let storageObject = storageObject else {
            throw ClientError.argument(NSLocalizedString(
                "StorageObject cannot be nil.", comment: ""))
        }
        let delegate:StorageObjectDelegate = storageObject as! StorageObjectDelegate
        var quit:Bool = false
        try contentQueue.sync(execute: {
            // TODO: What about locking the state property in storageObject ?
            if delegate.getState() == ProgressState.completed {
                return
            }
            quit = (delegate.getState() == ProgressState.queued ||
                delegate.getState() == ProgressState.inProgress)
            
            //Don't queue something already queued or in progress
            if quit {
                throw ClientError.state(NSLocalizedString(
                    "Transfer is pending", comment: ""))
            }
            contentQueued.lock()
            queue.append(storageObject)
            delegate.setState(state: ProgressState.queued)
            if !quit {
                self.dispatchProgressCallback(storageObject:storageObject)
            }
            contentQueued.signal()
            contentQueued.unlock()
        })

    }
    
    /**
     * Cancel the transfer of content to or from storage. This call has no effect
     * if the transfer is completed, already canceled, has failed, or the
     * storageObject is not queued.
     * - parameter storageObject the content storageObject to be canceled.
     */
    public func cancel(storageObject: AdvancedStorageObject) {
        var canceled:Bool = false
        contentQueue.sync(execute: {
            contentQueued.lock()
            if let i = queue.index(where: { $0 === storageObject }) {
                canceled = true
                (storageObject as! StorageObjectDelegate).setState(state:ProgressState.canceled)
                queue.remove(at: i)
            }
            defer {
                if canceled {
                    self.dispatchProgressCallback(storageObject:storageObject)
                }
                // TODO: Need this ?
                contentQueued.signal()
                contentQueued.unlock()
                
            }
        })
    }
    
    /**
     * Set a callback to be notified as the transfer progresses.
     * - parameter callback callback to invoke, if `nil`,
     *                 the existing callback will be removed
     */
    public func setProgressCallback(callback:ProgressCallback?) {
        self.progressCallback = callback
    }

    public func close() /*throws*/ {
        contentQueue.sync(execute: {
            if !self.closed {
                self.requestClose = true
                self.closed = true
            }
        })
    }
    
    public func isClosed() -> Bool {
        return contentQueue.sync(execute: {
            return self.closed
        })
    }
    
    public func updateProgress(storageObject: AdvancedStorageObject, bytes:Int64) {
            self.dispatchProgressCallback(storageObject:storageObject, bytes: bytes)
    }

    public func transferAndCallback(storageObject: AdvancedStorageObject,
                                    callback: @escaping (ClientError?)->()) {
        
        //print("#### In transferAndCallback ####")
        
        let delegate = storageObject as! StorageObjectDelegate
        var state:ProgressState = ProgressState.failed
        var bytes:Int64? = nil
        do {
            try delegate.sync(callback: { error in
                if error != nil {
                    state = delegate.isCanceled() ?
                        ProgressState.canceled : ProgressState.failed
                } else {
                    state = ProgressState.completed
                    bytes = storageObject.getLength()
                }
                delegate.setState(state:state)
                self.dispatchProgressCallback(storageObject: storageObject,
                                         bytes: bytes, error: error)
                callback(error)
            })
        } catch {
            state = delegate.isCanceled() ?
                ProgressState.canceled : ProgressState.failed
            delegate.setState(state:state)
            dispatchProgressCallback(storageObject:storageObject,
                                     error: error as? ClientError)
            callback(error as? ClientError)
        }
    }

    fileprivate func dispatchProgressCallback(storageObject: AdvancedStorageObject,
            bytes:Int64? = nil, error: ClientError? = nil) {
        if let progressCallback = progressCallback {
            progressQueue.async(execute: {
                let progress = ProgressImpl(storageObject: storageObject)
                if let bytes = bytes {
                    progress.setBytesTransferred(bytesTransferred: bytes)
                }
                if let error = error {
                    progress.setFailureCause(e: error)
                }
                progressCallback.progress(progress:progress)
            })
        }
    }
}
