/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation
import Alamofire

/**
 * Transport interface provides hook into HTTPS connection,
 * or can be stubbed out to allow for unit tests without real server.
 * The post/put/get/delete methods call this interface.
 */
class Transport {
    
    var defaultHttpManager:Alamofire.SessionManager?
    // Create a DispatchQueue for executing the Alamofire request
    // completion handler. It must not be on the main thread. Note that
    // since it is not on the main thread it cannot update UI components
    // All requests must set the queue parameter to this queue.
    // See Transport.swift
    var responseQueue:DispatchQueue {
        get {
            return DispatchQueue(label: "Transport response queue",
                                 qos: .utility)
        }
    }
    
    init(conManager: Alamofire.SessionManager) {
        self.defaultHttpManager = conManager
    }
    
    func invokeMethod(method: String, data: Foundation.Data?, headers: [String : String]?, url: String,
                      httpResponse: @escaping ((HttpResponse?, ClientError?) -> Void)) throws -> Void {
 
        struct CustomEncoding : ParameterEncoding {
            var body: Foundation.Data?
            
            init(body: Foundation.Data?) {
                self.body = body
            }
            // Called by Alamofire ignores first arg label
            func encode(_ convertible: URLRequestConvertible,
                        with parameters: Parameters?) throws -> URLRequest {
                var request = try! URLEncoding().encode(convertible, with: [:])
                request.httpBody = self.body
                return request
            }
        }
        
        switch method {
        case "GET":
            defaultHttpManager!.request(url,
                method: .get,
                parameters: [:],
                encoding: URLEncoding.default,
                headers: headers)
                .responseJSON(queue: responseQueue,
                              completionHandler: { (response) -> Void in
                    switch response.result {
                    case .success:
                        let tmpResponse: HttpResponse = HttpResponse(responseCode: (response.response?.statusCode)!, data: response.data!, headers: [:])
                        httpResponse(tmpResponse, nil)
                        return
                    case .failure(let error):
                        print(error)
                        httpResponse(nil, ClientError.network(NSLocalizedString("HTTP method " + method + " error: " + error.localizedDescription, comment: "")))
                        return
                    }
                })
            break
            
        case "POST" :
            let customEncoding: ParameterEncoding = CustomEncoding(body: data)
            Config.postCounter += 1
            if data == nil {
                print("Payload for POST request must not be nil")
                // TODO: Improvements does not return
                //return
            }
            defaultHttpManager!.request(url,
                method: .post,
                parameters: [:],
                encoding: customEncoding,
                headers: headers)
                .responseJSON(queue: responseQueue,
                              completionHandler: { (response) -> Void in
                    switch response.result {
                    case .success:
                        let tmpResponse: HttpResponse = HttpResponse(responseCode: (response.response?.statusCode)!, data: response.data!, headers: [:])
                        httpResponse(tmpResponse, nil)
                        return
                    case .failure(let error):
                        print(error)
                        httpResponse(nil, ClientError.network(NSLocalizedString("HTTP method " + method + " error: " + error.localizedDescription, comment: "")))
                        return
                    }
                })
            break
        case "DELETE" :
            defaultHttpManager!.request(url,
                method: .delete,
                parameters: [:],
                encoding: URLEncoding.default,
                headers: headers)
                .responseJSON(queue: responseQueue,
                              completionHandler: { (response) -> Void in
                    switch response.result {
                    case .success:
                        let tmpResponse: HttpResponse = HttpResponse(responseCode: (response.response?.statusCode)!, data: response.data!, headers: [:])
                        httpResponse(tmpResponse, nil)
                        return
                    case .failure(let error):
                        print(error)
                        httpResponse(nil, ClientError.network(NSLocalizedString("HTTP method " + method + " error: " + error.localizedDescription, comment: "")))
                        return
                    }
                })
            break
        case "PUT" :
            let customEncoding: ParameterEncoding = CustomEncoding(body: data)
            defaultHttpManager!.request(url,
                method: .put,
                parameters: [:],
                encoding: customEncoding,
                headers: headers)
                .responseJSON(queue: responseQueue,
                              completionHandler: { (response) -> Void in
                    switch response.result {
                    case .success:
                        let tmpResponse: HttpResponse = HttpResponse(responseCode: (response.response?.statusCode)!, data: response.data!, headers: [:])
                        httpResponse(tmpResponse, nil)
                        return
                    case .failure(let error):
                        print(error)
                        httpResponse(nil, ClientError.network(NSLocalizedString("HTTP method " + method + " error: " + error.localizedDescription, comment: "")))
                        return
                    }
                })
            break
        case "PATCH" :
            defaultHttpManager!.request(url,
                method: .patch,
                parameters: [:],
                encoding: URLEncoding.default,
                headers: headers)
                .responseJSON(queue: responseQueue,
                              completionHandler: { (response) -> Void in
                    switch response.result {
                    case .success:
                        let tmpResponse: HttpResponse = HttpResponse(responseCode: (response.response?.statusCode)!, data: response.data!, headers: [:])
                        httpResponse(tmpResponse, nil)
                        return
                    case .failure(let error):
                        print(error)
                        httpResponse(nil, ClientError.network(NSLocalizedString("HTTP method " + method + " error: " + error.localizedDescription, comment: "")))
                        return
                    }
                })
            break
        default:
            httpResponse(nil, ClientError.state(NSLocalizedString("Invalid HTTP method name: " + method, comment: "")))
        }
    }
}
