/*
 *  Copyright © 2016, 2017, Oracle and/or its affiliates. All rights reserved.
 *  This software is dual-licensed to you under the MIT License (MIT) and
 *  the Universal Permissive License (UPL). See the LICENSE file in the root
 *  directory for license terms. You may choose either license, or both.
 */

import Foundation

/**
 * `VirtualDeviceAttribute` is an attribute in a `VirtualDevice` model. The
 * `AbstractVirtualDevice.get(String)` and `AbstractVirtualDevice.set(String,
 * AnyObject)` methods
 * are wrappers around the `VirtualDeviceAttribute` instances of a
 * `VirtualDevice`.
 */
class VirtualDeviceAttributeBase {

    // Optional in enterprise
    var model: DeviceModelAttribute

    var value: AnyObject?
    var lastKnownValue: AnyObject?

    var changeCallback: ((ChangeEvent)->Void)?
    var errorCallback: ((ErrorEvent)->Void)?
    
    init(model: DeviceModelAttribute) {
        self.model = model
    }

    func getDeviceModelAttribute() -> DeviceModelAttribute? {
        return model
    }
    
    /**
     * Get the last&ndash;known value of an attribute.
     * This method returns the last&ndash;known value of the attribute, which
     * is not to be interpreted as the &quot;previous&quot; value.
     * For a device client, the last&ndash;known value is the device value.
     * For an enterprise client, the last&ndash;known value is the last value
     * retrieved from the server for a specific attribute, irrespective of 
     * the local 'set' value.
     *
     * - returns: the `lastKnownValue` attribute value
     */
    func getLastKnown() -> AnyObject? {
        return lastKnownValue
    }
    
    /**
     * Get the value of an attribute.
     * This method returns the current value
     * held by the virtual device model. For an enterprise client, no
     * REST API call is made to the server. For a device client, no call is
     * made to the physical device.
     *
     * - returns: the attribute value
     */
    func get() -> AnyObject? {
        return value ?? getDeviceModelAttribute()?.getDefaultValue()
    }

    // TODO: Impl
    /**
     * Set the value of an attribute.
     * This method is used by the application
     * to synchronize the `VirtualDevice` state with the physical device.
     * This results in a REST API call to the server. For an enterprise
     * client, an endpoint resource REST API call is made to the server. For
     * a device client, an endpoint DATA message REST API call is made to the
     * server.
     * The value is validated according to the constraints in the device model.
     *
     * - parameter value: the value to set
     *
     * - throws: `ClientError.argument` if `value` is an invalid value<br/>
     *           `ClientError.network` if a `MessageDispatcher` cannot be
     *           obtained.
     */
    func set(value: AnyObject) throws {
        fatalError("Subclass need to implement the 'set()' method.")
    }

    // Impl
    /**
     * Check if the attribute can be set. On the device-client, an attribute
     * should always be settable. On the enterprise-client, an attribute is
     * settable if the model says the attribute is writable.
     *
     * - returns: `true` if the attribute can be set
     */
    func isSettable() -> Bool {
        fatalError("Subclass need to implement the 'isSettable()' method.")
    }

    /**
     * Set a callback that is invoked when the value of an attribute changes on
     * the remote virtual device.
     *
     * - parameter callback: A callback to invoke when an attribute value
     *                       changes
     *
     * - see: `AbstractVirtualDevice.setOnChange(event: ChangeEvent)->Void)`
     */
    func setOnChange(callback: @escaping (ChangeEvent)->Void) {
        self.changeCallback = callback
    }
    
    /**
     * Get the `ChangeCallback` for this attribute
     * - returns: `changeCallback`, which may be `nil`
     */
    func getOnChange() -> ((ChangeEvent)->Void)? {
        return changeCallback
    }
    
    /**
     * Set a callback that is invoked when there is an error in setting the
     * value of an attribute. An error may arise because of a network exception.
     * In addition, on an enterprise client, an error indicates that a
     * request timed out or was rejected by the device client.
     *
     * - parameter callback: a callback to invoke when an attribute value 
     *                       changes
     *
     * - see: `AbstractVirtualDevice.setOnError((event: ErrorEvent)->Void)`
     */
    func setOnError(callback: @escaping (ErrorEvent)->Void) {
        self.errorCallback = callback
    }
    
    /**
     * Get the ErrorCallback for this attribute
     * - returns: *ErrorCallback*, which may be nil
     */
    func getOnError() -> ((ErrorEvent)->Void)? {
        return errorCallback
    }
    
    
    /**
     * - returns: only if the value is valid for the attribute 
     *            (otherwise throws)
     *
     * - throws: `ClientError.argument` if `value` is not the correct type or is
     *           is an invalid value for `attribute`.
     */
    func validate(attribute: DeviceModelAttribute, value: AnyObject?) throws {
        
        if value == nil {
            return
        }
        
        let type: DeviceModelAttributeType = attribute.getType()
        
        // block assumes value is not null
        switch (type) {
        case .INTEGER:
            if !(value is Int) {
                throw ClientError.argument(NSLocalizedString(
                    "Device Model: value is not Integer", comment: ""))
            }
            break
        case .NUMBER:
            if !(value is NSNumber) {
                throw ClientError.argument(NSLocalizedString(
                    "Device Model: value is not Number", comment: ""))
            }
            break
        case .STRING:
            if !(value is String) {
                throw ClientError.argument(NSLocalizedString(
                    "Device Model: value is not String", comment: ""))
            }
            break
        case .BOOLEAN:
            if !(value is Bool) {
                throw ClientError.argument(NSLocalizedString(
                    "Device Model: value is not Boolean", comment: ""))
            }
            break
        case .DATETIME:
            let i: Int64? = (value! as? NSNumber)?.int64Value
            if (i == nil) {
                throw ClientError.argument(NSLocalizedString(
                    "Device Model: value is not DateTime", comment: ""))
            }
            break
        case .URI:
            if !(value is ExternalObject) {
                throw ClientError.argument(NSLocalizedString(
                    "value is not an ExternalObject", comment: ""));
            }
            break
        }
        
        // Is Double.nan a valid value ?
        if (type == .INTEGER) || (type == .NUMBER) {
            // Assumption here is that lowerBound <= upperBound
            let val: Double = Double(truncating: value! as! NSNumber)
            if let nuval = attribute.getUpperBound() {
                let upper: Double = Double(truncating: nuval)
                 if val > upper {
                    throw ClientError.argument(NSLocalizedString(
                        "Device Model: value exceeds upper bound",
                        comment: ""))
                 }
            }
            if let nlval = attribute.getLowerBound() {
                let lower = Double(truncating: nlval)
                if val < lower {
                    throw ClientError.argument(NSLocalizedString(
                        "Device Model: value undercuts lower bound",
                        comment: ""))
                }
            }
        }
    }
}
